#include "udm_config.h"

#ifdef HAVE_ZLIB

#include <stdlib.h>
#include <string.h>
#include <zlib.h>

#include "udm_common.h"
#include "udm_contentencoding.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"

size_t UdmInflate(UDM_AGENT * Indexer, char *buffer, size_t length) {

  z_stream zstream;
  int header;

	if ((buffer[0] == (char)0x1f) && (buffer[1] == (char)0x8b)) {
		header = 1;
	} else {
		header = 0;
	}

  zstream.zalloc = Z_NULL;
  zstream.zfree = Z_NULL;
  zstream.opaque = Z_NULL;

  inflateInit2(&zstream, -MAX_WBITS);

  zstream.next_in = (Byte*)UdmXmalloc((size_t)Indexer->Conf->max_doc_size);
  if (header) {
  	bcopy(&buffer[2], zstream.next_in, length - 2);
	zstream.avail_in = length - 6;   /* 2 bytes - header, 4 bytes - trailing CRC */
  } else {
	bcopy(buffer, zstream.next_in, length);
  	zstream.avail_in = length;
  }

  zstream.next_out = buffer;
  zstream.avail_out = (uLong)Indexer->Conf->max_doc_size;

  inflate(&zstream, Z_FINISH);
  inflateEnd(&zstream);

  free(zstream.next_in);
  return zstream.total_out;
}


struct gztrailer {
    int crc32;
    int zlen;
};

size_t UdmUnGzip(UDM_AGENT * Indexer, char *buffer, size_t length) {

  const char gzheader[10] = { 0x1f, 0x8b, Z_DEFLATED, 0, 0, 0, 0, 0, 0, 3 };
  z_stream zstream;
  Byte *buf;

  zstream.zalloc = Z_NULL;
  zstream.zfree = Z_NULL;
  zstream.opaque = Z_NULL;

  inflateInit2(&zstream, -MAX_WBITS);

  buf = zstream.next_in = (Byte*)UdmXmalloc((size_t)Indexer->Conf->max_doc_size);
  bcopy(&buffer[sizeof(gzheader)], zstream.next_in, length - sizeof(gzheader));
  zstream.next_out = buffer;
  zstream.avail_in = length - sizeof(gzheader) - sizeof(struct gztrailer);
  zstream.avail_out = (uLong)Indexer->Conf->max_doc_size;

  inflate(&zstream, Z_FINISH);
  inflateEnd(&zstream);

  UDM_FREE(buf);

  return zstream.total_out;

}

size_t UdmUncompress(UDM_AGENT * Indexer, char *buffer, size_t length) {
  Byte *buf;
  uLong	Len;

  buf = (Byte*)UdmXmalloc((size_t)Indexer->Conf->max_doc_size);
  bcopy(buffer, buf, length);

	uncompress(buffer, &Len, buf, length);

	return (size_t) Len;
}

#endif
