#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_env.h"
#include "udm_parser.h"
#include "udm_robots.h"
#include "udm_host.h"
#include "udm_hrefs.h"
#include "udm_server.h"
#include "udm_filter.h"
#include "udm_url.h"
#include "udm_proto.h"
#include "udm_alias.h"
#include "udm_log.h"
#include "udm_mimetype.h"
#include "udm_cache.h"
#include "udm_stopwords.h"
#include "udm_urldcl.h"
#include "udm_guesser.h"
#include "udm_vars.h"
#include "udm_sdp.h"
#include "udm_synonym.h"
#include "udm_section.h"
#include "udm_doc.h"

/**************************** DBAddr ***********************************/
static int UdmStr2DBMode(const char * str1){
int m;
	m = UDM_DBMODE_SINGLE;
	if(!UDM_STRNCASECMP(str1,"multi-crc"))m=UDM_DBMODE_MULTI_CRC;
	else
	if(!UDM_STRNCASECMP(str1,"crc-multi"))m=UDM_DBMODE_MULTI_CRC;
	else
	if(!UDM_STRNCASECMP(str1,"single"))m=UDM_DBMODE_SINGLE;
	else
	if(!UDM_STRNCASECMP(str1,"crc"))m=UDM_DBMODE_SINGLE_CRC;
	else
	if(!UDM_STRNCASECMP(str1,"multi"))m=UDM_DBMODE_MULTI;
	else
	if(!UDM_STRNCASECMP(str1,"word2url"))m=UDM_DBMODE_WORD2URL;
	else
	if(!UDM_STRNCASECMP(str1,"cache"))m=UDM_DBMODE_CACHE;
	else
	if(!UDM_STRNCASECMP(str1,"blob"))m=UDM_DBMODE_BLOB;
	return(m);
}
__INDLIB__ int UdmEnvSetDBMode(UDM_ENV * Conf,const char *str){
	Conf->DBMode=UdmStr2DBMode(str);
	return(Conf->DBMode);
}

__INDLIB__ UDM_ENV * UdmAllocEnv(void){
	UDM_ENV * Conf;

	Conf=(UDM_ENV *)malloc(sizeof(UDM_ENV));
	bzero(Conf,sizeof(*Conf));
	UdmInitHrefList(&Conf->HrefList);

	snprintf(Conf->vardir,sizeof(Conf->vardir),"%s%s",UDM_VAR_DIR,UDMSLASHSTR);
	Conf->max_doc_size=UDM_MAXDOCSIZE;
	Conf->have_targets=1;
	Conf->use_remote_cont_type=1;
	Conf->DBMode=UDM_DBMODE_SINGLE;
	Conf->DBUseLock=1;

	/* for logger */
	/* Set default facility & loglevel */                                       
	Conf->logFacility=LOG_FACILITY;                                              
	Conf->logLevel=UDM_LOG_INFO;                                                 

	Conf->min_word_len=1;
	Conf->max_word_len=32;
	Conf->flashmode=UDM_FLASH_BEST;
	
	Conf->max_text_size=UDM_PLAINMAXTEXTSIZE;

	return(Conf);
}
__INDLIB__ void UdmFreeEnv(UDM_ENV * Env){
	if(Env){
		UDM_FREE(Env->DBAddr);
		UDM_FREE(Env->csrv);
		UdmSDClientListFree(&Env->sdcl);
		
		if (Env->csrv) {
			if(Env->csrv->regexp)regfree(Env->csrv->regexp);
			UDM_FREE(Env->csrv->url);
			UDM_FREE(Env->csrv->alias);
			UDM_FREE(Env->csrv->charset);
			UDM_FREE(Env->csrv->category);
			UDM_FREE(Env->csrv->tag);
			UDM_FREE(Env->csrv->basic_auth);
			UDM_FREE(Env->csrv->user);
			UDM_FREE(Env->csrv->passwd);
			UDM_FREE(Env->csrv->proxy);
			UDM_FREE(Env->csrv->proxy_basic_auth);
			UDM_FREE(Env->csrv->htdb_list);
			UDM_FREE(Env->csrv->htdb_doc);
			UDM_FREE(Env->csrv->mirror_root);
			UDM_FREE(Env->csrv->mirror_headers);
			UDM_FREE(Env->csrv);
		}
		
		
		UDM_FREE(Env->logd_addr);
		UDM_FREE(Env->urld_addr);
		UDM_FREE(Env->alias_prog);
		UDM_FREE(Env->user_agent);
		UDM_FREE(Env->local_charset);
		UDM_FREE(Env->browser_charset);
		UDM_FREE(Env->listen);
		UDM_FREE(Env->Affix);
		UDM_FREE(Env->Spell);

		UDM_FREE(Env->stored_addr);

		if (Env->vars != NULL) UdmFreeVarList(Env->vars);

		UdmFreeTargets(&Env->Targets);
		UdmFreeParsers(Env);
		UdmFreeStopList(Env);
		UdmFreeRobots(&Env->Robots);
		UdmFreeHost_addr(Env);
		UdmFreeTypes(Env);
		UdmFreeHrefList(&Env->HrefList);
		UdmFreeServers(Env);
		UdmAliasListFree(&Env->AliasList);
		UdmAliasListFree(&Env->ReverseAliasList);
		UdmSynonymListFree(&Env->SynList);
		UdmSectionListFree(&Env->SectionList);
		UdmFreeFilters(Env);
		UdmCloseCache(Env);
		UdmCloseURLD(Env);
		UdmFreeLangMapList(Env);
		free(Env);
	}
}
__INDLIB__ int UdmEnvErrCode(UDM_ENV * Conf){
	return(Conf->errcode);
}
__INDLIB__ char * UdmEnvErrMsg(UDM_ENV * Conf){
	return(Conf->errstr);
}
