#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <regex.h>
#include <signal.h>

#if (WIN32|WINNT)
#include <time.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_LIBUTIL_H
#include <libutil.h>
#endif

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "udm_utils.h"
#include "udm_common.h"
#include "udm_log.h"
#include "udm_conf.h"
#include "udm_indexer.h"
#include "udm_robots.h"
#include "udm_db.h"
#include "udm_url.h"
#include "udm_parser.h"
#include "udm_proto.h"
#include "udm_hrefs.h"
#include "udm_mutex.h"
#include "udm_crc32.h"
#include "udm_xmalloc.h"
#include "udm_http.h"
#include "udm_host.h"
#include "udm_server.h"
#include "udm_filter.h"
#include "udm_alias.h"
#include "udm_word.h"
#include "udm_crossword.h"
#include "udm_parsehtml.h"
#include "udm_cache.h"
#include "udm_spell.h"
#include "udm_execget.h"
#include "udm_agent.h"
#include "udm_mimetype.h"
#include "udm_urldcl.h"
#include "udm_doc.h"
#include "udm_parsedate.h"
#include "udm_unicode.h"
#include "udm_worddb.h"
#include "udm_contentencoding.h"
#include "udm_header.h"
#ifdef HAVE_ZLIB
#include "udm_store.h"
#endif

#define UDM_THREADINFO(ThreadInfo,h,s,m)	if(ThreadInfo)ThreadInfo(h,s,m)


/*********************** 'UrlFile' stuff (for -f option) *******************/

__INDLIB__ int UdmURLFile(UDM_AGENT *Indexer, int action){
	FILE *url_file;
	char str[1024]="";
	char str1[1024]="";
	int result,res;
	UDM_URL myurl;
	UDM_DOCUMENT Href;

	/* Read lines and clear/insert/check URLs                     */
	/* We've already tested in main.c to make sure it can be read */
	/* FIXME !!! Checking should be done here surely              */

	if(!strcmp(Indexer->Conf->url_file_name,"-"))
		url_file=stdin;
	else
		url_file=fopen(Indexer->Conf->url_file_name,"r");

	while(fgets(str1,sizeof(str1),url_file)){
		char *end;
		if(!str1[0])continue;
		end=str1+strlen(str1)-1;
		while((end>=str1)&&(*end=='\r'||*end=='\n')){
			*end=0;if(end>str1)end--;
		}
		if(!str1[0])continue;
		if(str1[0]=='#')continue;

		if(*end=='\\'){
			*end=0;strcat(str,str1);
			continue;
		}
		strcat(str,str1);
		strcpy(str1,"");

		switch(action){
		case UDM_URL_FILE_REINDEX:
			UdmAddURLLimit(Indexer->Conf,str);
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
			Indexer->Conf->have_targets=1;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
			result=UdmMarkForReindex(Indexer);
			if(result!=IND_OK)return(result);
			UdmClearURLLimit(Indexer->Conf);
			break;
		case UDM_URL_FILE_CLEAR:
			UdmAddURLLimit(Indexer->Conf,str);
			result=UdmClearDB(Indexer);
			if(result!=IND_OK)return(IND_ERROR);
			UdmClearURLLimit(Indexer->Conf);
			break;
		case UDM_URL_FILE_INSERT:
			Href.url=str;
			Href.hops=0;
			Href.referrer=0;
			Href.stored=0;
			Href.tag=NULL;
			Href.category=NULL;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
			if(UdmAddHref(&Indexer->Conf->HrefList,&Href))
				Indexer->Conf->have_targets=1;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);

			break;
		case UDM_URL_FILE_PARSE:
			res=UdmParseURL(&myurl,str);
			if((!res)&&(!myurl.schema[0]))
				res=UDM_URL_BAD;
			if(res){
				switch(res){
				case UDM_URL_LONG:
					UdmLog(Indexer,UDM_LOG_ERROR,"URL too long: '%s'",str);
					break;
				case UDM_URL_BAD:
				default:
					UdmLog(Indexer,UDM_LOG_ERROR,"Error in URL: '%s'",str);
				}
				return(IND_ERROR);
			}
			break;
		}
		str[0]=0;
	}
	if(url_file!=stdin)
		fclose(url_file);
	return(IND_OK);
}


/******* Main indexer functions StartUp & IndexNextURL ********/

/* Init section for indexing */

static int StartUp(UDM_AGENT * Indexer, int index_flags){
	char *tok,*lt;
	char tables[1024];

	if(Indexer->Conf->DBMode==UDM_DBMODE_CACHE){
		if(IND_OK!=UdmOpenCache(Indexer->Conf,UdmDBErrorMsg(Indexer->db))){
			return(1);
		}
	}
	Indexer->Conf->have_targets=1;
	if(index_flags&UDM_FLAG_MARK){
		if(Indexer->Conf->url_file_name){
			UdmURLFile(Indexer,UDM_URL_FILE_REINDEX);
		}else{
			UdmMarkForReindex(Indexer);
		}
		if(UdmDBErrorCode(Indexer->db)){
			return(1);
		}
	}
	if(index_flags&UDM_FLAG_SKIP_LOCKING)
		Indexer->Conf->DBUseLock=0;
	else
		Indexer->Conf->DBUseLock=1;

	if((Indexer->Conf->url_file_name) && (index_flags&UDM_FLAG_INSERT)) {
		UdmURLFile(Indexer,UDM_URL_FILE_INSERT);
		if(UdmDBErrorCode(Indexer->db)){
			return(1);
		}
	}

	/* Now load all server tables from database            */
	/* Their names are in Indexer->Conf->srv_tables string */

	strcpy(tables,Indexer->Conf->srv_tables);
	tok=UdmGetToken(tables," \r\n\t",&lt);
	while(tok){
		UdmLog(Indexer,UDM_LOG_DEBUG,"Load server table '%s'",tok);
		if(UdmLoadServerTable(Indexer,tok,index_flags)!=IND_OK)
			return(1);
		tok=UdmGetToken(NULL," \r\n\t",&lt);
	}
	return(0);
}

__INDLIB__ UDM_DOCUMENT * UdmGetTarget(UDM_AGENT * Indexer,int * result, int index_flags){
	UDM_DOCUMENT * Doc=NULL;

	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
	/* Store URLs from cache into database */
	if(UdmStoreHrefs(Indexer))Indexer->Conf->have_targets=1;

	if(UdmDBErrorCode(Indexer->db)){
		*result=IND_ERROR;
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
		return(NULL);
	}
	if(!Indexer->Conf->have_targets){
		*result=IND_NO_TARGET;
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
		return(NULL);
	}

	/* Get Next URL to be indexed from the database */
	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Selecting","");
		
	if(Indexer->Conf->urld_fd){
		Doc=UdmURLDSelect(Indexer->Conf);
		if(!Doc){
			Indexer->Conf->have_targets=0;
			*result=IND_NO_TARGET;
		}else{	
		}
	}else{
		Doc=UdmGetDocInfo(Indexer,index_flags);
		if(!Doc){
			if(UdmDBErrorCode(Indexer->db)){
				*result=IND_ERROR;
			}else{
				Indexer->Conf->have_targets=0;
				*result=IND_NO_TARGET;
			}
		}
	}
	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
	return(Doc);
}

void UdmParseHTTPResponse(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){			
	char *token, *lt;

	Doc->content=NULL;
	Doc->content_type=NULL;
	Doc->last_mod_time=0;
	Doc->headers=NULL;
	Doc->status=1;

	/* Cut HTTP response header first        */
	for(token=Indexer->buf;*token;token++){
		if(!strncmp(token,"\r\n\r\n",4)){
			*token='\0';
			Doc->content=token+4;
			break;
		}else
		if(!strncmp(token,"\n\n",2)){
			*token='\0';
			Doc->content=token+2;
			break;
		}
	}

	/* Bad response, return */
	if(!Doc->content)return;

	/* Copy headers for mirrors */
	Doc->headers = strdup(Indexer->buf);
	
	/* Now lets parse response header lines */
	token=UdmGetToken(Indexer->buf,"\r\n",&lt);
	
	if(!token)return;
	
	if(!strncmp(token,"HTTP/",5)){
		Doc->status=atoi(token+8);
	}else{
		return;
	}
	token=UdmGetToken(NULL,"\r\n",&lt);
	
	while(token){
		UDM_HTTPHEADER header;

		UdmLog(Indexer,UDM_LOG_DEBUG,"%s",token);

		header.name=token;
		header.value=strchr(token,':');
		if(!header.value){
			UdmLog(Indexer,UDM_LOG_DEBUG,"Incorrect header");
			token=UdmGetToken(NULL,"\r\n",&lt);
			continue;
		}
		header.value[0]='\0';
		header.value++;
		header.value=UdmTrim(header.value," \t:");

		UdmHeaderListAdd(&Doc->HeaderList,&header);

		if(!strcasecmp(header.name,"Content-Type")){
			UDM_FREE(Doc->content_type);
			if (!Indexer->Conf->use_remote_cont_type) {
				char * ct=UdmContentType(Indexer->Conf,Doc->url);
				Doc->content_type=ct?strdup(ct):NULL;
			}
			if (!Doc->content_type) {
				Doc->content_type=strdup(header.value);
			}
		}else
		if ((!strcasecmp(header.name,"Server"))){
			if(Indexer->Conf->force1251){
				if (!UDM_STRNCASECMP(header.value,"Microsoft")||!UDM_STRNCASECMP(header.value,"IIS")){
					UDM_FREE(Doc->charset);
					Doc->charset=strdup(UdmCharsetCanonicalName("windows-1251"));
				}
			}
		}else
		if(!strcasecmp(header.name,"Content-Language")){
			UDM_FREE(Doc->lang);
			Doc->lang=strdup(header.value);
		}else
		if(!strcasecmp(header.name,"Location")){
			UDM_URL newURL;
			switch(UdmParseURL(&newURL,header.value)){
				case UDM_URL_OK:
					if(newURL.schema[0]){
						UDM_DOCUMENT Href;
						Href.url=header.value;
						Href.hops=Doc->hops+1;
						Href.referrer=Doc->referrer;
						Href.stored=0;
						Href.tag=NULL;
						Href.category=NULL;
						UdmAddHref(&Doc->HrefList,&Href);
					}
					break;
				case UDM_URL_LONG:
					UdmLog(Indexer,UDM_LOG_ERROR,"Redirect URL too long: '%s'",header.value);
					break;
				case UDM_URL_BAD:
				default:
					UdmLog(Indexer,UDM_LOG_ERROR,"Error in redirect URL: '%s'",header.value);
			}
		}else
		if(!strcasecmp(header.name,"Content-Length")){
			/*         */
			/*   core          */
			/* Doc->size=atoi(token+16); */
		}else
		if(!strcasecmp(header.name,"Subject")){
			UDM_FREE(Doc->subject);
			Doc->subject=(char*)malloc(strlen(header.value)+100);
			udm_rfc1522_decode(Doc->subject,header.value);
			Doc->subject[UDM_MAXSUBJSIZE]='\0';
		}else
		if(!strcasecmp(header.name,"Message-ID")){
			UDM_FREE(Doc->message_id);
			header.value=UdmTrim(header.value,"<>");
			Doc->message_id=strdup(header.value);
		}else
		if(!strcasecmp(header.name,"From")){
			Doc->from=(char*)malloc(strlen(header.name)+100);
			udm_rfc1522_decode(Doc->from,header.name);
			Doc->from[UDM_MAXFROMSIZE]='\0';
		}else
		if(!strcasecmp(header.name,"Newsgroups")){
			Doc->newsgroups=strdup(header.value);
		}else
		if(!strcasecmp(header.name,"Date")){
			if(!Doc->last_mod_time)
				Doc->last_mod_time=UdmHttpDate2Time_t(header.value);
		}else		
		if(!strcasecmp(header.name,"References")){
			Doc->references=strdup(header.value);
		}else
		if(!strcasecmp(header.name,"Last-Modified")){
			Doc->last_mod_time=UdmHttpDate2Time_t(header.value);
		}else
		if (!strcasecmp(header.name, "Content-Encoding")) { 
			if (!strcasecmp(header.value, "gzip")) {
				Doc->content_encoding = UDM_ENCODING_GZIP;
			} else if (!strcasecmp(header.value, "x-gzip")) {
				Doc->content_encoding = UDM_ENCODING_GZIP;
			} else if (!strcasecmp(header.value, "deflate")) {
				Doc->content_encoding = UDM_ENCODING_DEFLATE;
			} else if (!strcasecmp(header.value, "identity")) {
				Doc->content_encoding = UDM_ENCODING_IDENTITY;
			} else if (!strcasecmp(header.value, "x-compress")) {
				Doc->content_encoding = UDM_ENCODING_COMPRESS;
			} else if (!strcasecmp(header.value, "compress")) {
				Doc->content_encoding = UDM_ENCODING_COMPRESS;
			} else {
				Doc->content_encoding = UDM_ENCODING_UNKNOWN;
			}
		}
		token=UdmGetToken(NULL,"\r\n",&lt);
	}
	if(Doc->content_type){
		char * p;
		/* Cat charset part from Content-Type */
		if((p=strstr(Doc->content_type,"charset="))){
			const char *cs = UdmCharsetCanonicalName(p+8);
			UDM_FREE(Doc->charset);
			Doc->charset=strdup(cs?cs:(p+8));
			*p = '\0';
		}
		Doc->content_type=UdmRTrim(Doc->content_type, "; ");
	}
}


static void UdmBuildHTTPRequest(UDM_ENV * Conf,UDM_SERVER * CurSrv,UDM_DOCUMENT * Doc,UDM_URL * realURL){
	const char * method;
	method=(Doc->method==UDM_HEAD)?"HEAD":"GET";

	if(CurSrv->proxy){
		Doc->proxy_host=CurSrv->proxy?strdup(CurSrv->proxy):NULL;
		Doc->proxy_port=CurSrv->proxy_port;
		sprintf(Doc->request,"%s %s HTTP/1.0\r\n",method,Doc->url);
	}else{
		Doc->proxy_host=NULL;
		sprintf(Doc->request,"%s %s%s HTTP/1.0\r\n",method,realURL->path,realURL->filename);
	}

	/* Add If-Modified-Since header */
	if((Doc->status)&&(Doc->last_mod_time)){
		char buf[128]="";
		UdmTime_t2HttpStr(Doc->last_mod_time, buf);
		sprintf(UDM_STREND(Doc->request),"If-Modified-Since: %s\r\n",buf);
	}

	/* HTTP and FTP specific stuff */
	if((!strcasecmp(realURL->schema,"http"))||(!strcasecmp(realURL->schema,"https"))||(!strcasecmp(realURL->schema,"ftp"))||(!strcasecmp(realURL->schema,"https"))){

		/* User agent */
		if(Conf->user_agent)
			sprintf(UDM_STREND(Doc->request),"User-Agent: %s\r\n",Conf->user_agent);

		/* If LocalCharset specified */
		if( Conf->local_charset )
			sprintf(UDM_STREND(Doc->request),"Accept-charset: %s;q=1.0, *;q=0.9, utf-8;q=0.8\r\n",
				Conf->local_charset);

		/* Host Name for virtual hosts */
		if(realURL->port){
			sprintf(UDM_STREND(Doc->request),"Host: %s:%d\r\n",realURL->hostname,realURL->port);
		}else{
			sprintf(UDM_STREND(Doc->request),"Host: %s\r\n",realURL->hostname);
		}

		/* Auth if required */
		if(CurSrv->basic_auth)
			sprintf(UDM_STREND(Doc->request),"Authorization: Basic %s\r\n",
				CurSrv->basic_auth);
		
		/* ProxyAuth if required */
		if(CurSrv->proxy_basic_auth)
			sprintf(UDM_STREND(Doc->request),"Proxy-Authorization: Basic %s\r\n",
				CurSrv->proxy_basic_auth);

#ifdef HAVE_ZLIB
		sprintf(UDM_STREND(Doc->request), "Accept-Encoding: gzip,deflate,compress\r\n");
#endif

	}
	/* Add user defined headers */
	strcat(Doc->request,Conf->extra_headers);

	/* Empty line is the end of HTTP header */
	strcat(Doc->request,"\r\n");
#ifdef DEBUG_REQUEST
	fprintf(stderr,"Request:'%s'\n",request);
#endif
}

static int UdmSUpdateURL(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	int result;
	if(Indexer->Conf->urld_fd){
		UdmURLDSUpdate(Indexer->Conf,Doc);
		result=IND_OK;
	}else{
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_SUPDATE);
	}
	return result;
}

static int UdmLUpdateURL(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	int result;
	if(Indexer->Conf->urld_fd){
		UdmURLDLUpdate(Indexer->Conf,Doc);
		result=IND_OK;
	}else{
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_LUPDATE);
	}
	return result;
}

static int UdmDocStoreHrefs(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
	int i;

	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
	for(i=0;i<Doc->HrefList.nhrefs;i++){
		UDM_DOCUMENT Href;
		UDM_ALIAS  * Alias;
		char alias[UDM_URLSIZE+10]="";
		size_t aliassize=sizeof(alias);
		size_t nparts=10;
		UDM_MATCH_PART Parts[10];

		bzero(&Href,sizeof(Href));
		Href.url=Doc->HrefList.Href[i].url;
		Href.referrer=Doc->HrefList.Href[i].referrer;
		Href.hops=Doc->HrefList.Href[i].hops;
		Href.tag=Doc->HrefList.Href[i].tag;
		Href.category=Doc->HrefList.Href[i].category;
		
		UdmLog(Indexer,UDM_LOG_DEBUG,"Link %s",Href.url);
		if((Alias=UdmAliasFind(&Indexer->Conf->ReverseAliasList,Href.url,nparts,Parts))){
			UdmAliasApply(alias,aliassize,Href.url,Alias,nparts,Parts);
			if(alias[0]){
				Href.url=alias;
				UdmLog(Indexer,UDM_LOG_DEBUG,"ReverseAlias: '%s'",Href.url);
			}
		}

		if(Indexer->Conf->urld_fd){
			UdmURLDInsert(Indexer->Conf,&Href);
		}else{
			UDM_SERVER * Srv;
			
			char reason[1024]="";
			

			if(!(Srv=UdmFindServer(Indexer->Conf,Href.url,NULL))){
				UdmLog(Indexer,UDM_LOG_DEBUG,"no Server, skip it",Href.url);
				continue;
			}
			
			/* Check Allow/Disallow/CheckOnly stuff */
			Doc->method=UdmFindFilter(Indexer->Conf,Href.url,reason);
			if(Doc->method==UDM_DISALLOW){
				UdmLog(Indexer,UDM_LOG_DEBUG,"%s, skip it",reason);
				continue;
			}else{
				UdmLog(Indexer,UDM_LOG_DEBUG,"%s",reason);
			}
			
			/* FIXME: add MaxHops, Robots */
			if(UdmAddHref(&Indexer->Conf->HrefList,&Href)){
				Indexer->Conf->have_targets=1;
			}
		}
	}
	if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
	return IND_OK;
}

static int UdmCheckURL(UDM_AGENT *Indexer,UDM_SERVER *CurSrv,UDM_DOCUMENT *Doc,UDM_URL *RealURL,char *aliastr){
	int result=IND_UNKNOWN;
	char reason[1024]="";
	
	switch(UDM_SRV_TYPE(CurSrv->match_type)){
		case UDM_SERVER_STRING:
			UdmLog(Indexer,UDM_LOG_DEBUG,"Realm string '%s'",CurSrv->url);
			break;
		case UDM_SERVER_REGEX:
			UdmLog(Indexer,UDM_LOG_DEBUG,"Realm regex '%s'",CurSrv->url);
			break;
		case UDM_SERVER_SUBNET:
			UdmLog(Indexer,UDM_LOG_DEBUG,"Subnet '%s'",CurSrv->url);
			break;
		case UDM_SERVER_SUBSTR:
		default:
			UdmLog(Indexer,UDM_LOG_DEBUG,"Server '%s'",CurSrv->url);
			break;
	}
	Indexer->doc_timeout=CurSrv->doc_timeout;
	Indexer->read_timeout=CurSrv->read_timeout;

	/* Check Allow/Disallow/CheckOnly stuff */
	Doc->method=UdmFindFilter(Indexer->Conf,Doc->url,reason);
	UdmLog(Indexer,UDM_LOG_DEBUG,"%s",reason);

	if(Doc->method==UDM_DISALLOW){
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
		return(result);
	}

	/* Check that hops is less than MaxHops */
	if(Doc->hops>CurSrv->maxhops){
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
		UdmLog(Indexer,UDM_LOG_WARN,"Too many hops (%d)... deleted.",Doc->hops);
		return(result);
	}

	UdmParseURL(RealURL,Doc->url);
	/* Check for too many errors on this server */
	if((UdmGetServerErr(Indexer->Conf,RealURL->hostname)>=CurSrv->max_net_errors)&&(CurSrv->max_net_errors)){
		UdmLog(Indexer,UDM_LOG_WARN,"Too many network errors for this server, skipped");
		Doc->next_index_time=time(NULL)+CurSrv->net_error_delay_time;
		Doc->status=504;
		result=UdmSUpdateURL(Indexer,Doc);
	}else 
	/* Check whether URL is disallowed by robots.txt */
	if(CurSrv->use_robots){
		UDM_ROBOT_RULE * rule;

		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
		rule=UdmFindRobotRule(&Indexer->Conf->Robots,RealURL);
		if(rule){
			UdmLog(Indexer,UDM_LOG_WARN,"robots.txt: '%s %s'",(rule->cmd==UDM_ROBOT_ALLOW_CMD)?"Allow":"Disallow",rule->path);
			if(rule->cmd==UDM_ROBOT_DISALLOW_CMD)
				result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
		}
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
	}
	if(result) return(result);
	return(0);
}

static int UdmAliasURL(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_URL *RealURL,char *aliastr,size_t aliasize){
	UDM_ALIAS	*Alias;
	UDM_MATCH_PART	Parts[10];
	size_t nparts=10;

#ifdef UDM_WITH_ALIASPROG
	/* Start AliasProg if no alias in "Server" command */
	if((Indexer->Conf->alias_prog)&&(!aliastr[0])){
		FILE * aprog;
		char cmd[UDM_URLSIZE*3]="";
		char arg[UDM_URLSIZE*3]="";
		char * ares=NULL;
		char * args[1];
		char * a, *c;
 
		/* Escape command argument */
		for(a=Doc->url,c=arg;*a;a++,c++){
			switch(*a){
				case '\\':
				case '\'':
				case '\"':
					(*c)='\\';
					c++;
					break;
				default:
					break;
			}
			*c=*a;
		}
		*c='\0';
		args[0]=arg;

		UdmBuildParamStr(cmd,sizeof(cmd),Indexer->Conf->alias_prog,args,1);
		aprog=popen(cmd,"r");
		UdmLog(Indexer,UDM_LOG_EXTRA,"Starting AliasProg: '%s'",cmd);
		if(aprog){
			ares=fgets(aliastr,sizeof(aliastr),aprog);
			pclose(aprog);
			if(!ares){
				UdmLog(Indexer,UDM_LOG_ERROR,"AliasProg didn't return result: '%s'",cmd);
				return(IND_ERROR);
			}
		}else{
			UdmLog(Indexer,UDM_LOG_ERROR,"Can't start AliasProg: '%s'",cmd);
			return(IND_ERROR);
		}
		if(ares[0]){
			ares+=strlen(ares);
			ares--;
			while((ares>=aliastr)&&strchr(" \r\n\t",*ares)){
				*ares='\0';
				ares--;
			}
		}
		UdmLog(Indexer,UDM_LOG_EXTRA,"AliasProg result: '%s'",aliastr);
	}
#endif

	/* Find alias when aliastr is empty, i.e.     */
	/* when there is no alias in "Server" command */
	/* and no AliasProg                           */
	if((!aliastr[0])&&(Alias=UdmAliasFind(&Indexer->Conf->AliasList,Doc->url,nparts,Parts))){
		UdmAliasApply(aliastr,aliasize,Doc->url,Alias,nparts,Parts);
	}

	/* Parse aliased URL */
	if(aliastr[0]){
		if(UdmParseURL(RealURL,aliastr)){
			UdmLog(Indexer,UDM_LOG_ERROR,"Error in aliased URL: '%s'",aliastr);
			return(IND_OK);
		}else{
			UdmLog(Indexer,UDM_LOG_EXTRA,"Alias: '%s'",aliastr);
		}
	}
	return(0);
}

static int UdmIndexURL(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc,int index_flags){
	char aliastr[UDM_URLSIZE+10]="";
	int realsize;
	int changed=1;
	int reindex;
	int found_in_mirror = 0;
	int result=IND_UNKNOWN;
	char *real_content_type=NULL;

	UDM_URL		CurURL;
	UDM_URL		RealURL;
	UDM_SERVER	*CurSrv=0;

#ifdef USE_PARSER
	char *mime;
	int mimeno;
#endif

	/* Alloc buffer for document */
	if(!Indexer->buf){
		Indexer->buf=(char*)UdmXmalloc((size_t)Indexer->Conf->max_doc_size);
	}
	UdmLog(Indexer,UDM_LOG_INFO,"URL: %s",Doc->url);
	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Found",Doc->url);
	reindex=index_flags&UDM_FLAG_REINDEX; 
	if(reindex)Doc->last_mod_time=0;

#ifdef HAVE_SETPROCTITLE
	/* To see the URL being indexed in "ps" output on FreeBSD */
	/* Do it if single thread version */
	if(!(Indexer->handle)) setproctitle("%s",Doc->url);
#endif
	/* Check that URL has valid syntax */
	if(UdmParseURL(&CurURL,Doc->url)){
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
		UdmLog(Indexer,UDM_LOG_WARN,"Invalid URL: %s ... deleting",Doc->url);
		return(result);
	}

	/* Find correspondent Server record from indexer.conf */
	if(!(CurSrv=UdmFindServer(Indexer->Conf,Doc->url,aliastr))){
		UdmLog(Indexer,UDM_LOG_WARN,"No 'Server' command for url... deleted.");
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
		return(result);
	}
	
	if(!strcmp(CurURL.filename,"robots.txt")&&!CurSrv->use_robots&&!strncmp(CurURL.schema,"http",4)){
		UdmLog(Indexer,UDM_LOG_WARN,"robots.txt support is disallowed for '%s'",CurURL.hostinfo);
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
		return(UdmParseRobot(&Indexer->Conf->Robots,NULL,CurURL.hostinfo));
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
	}

	if((result=UdmCheckURL(Indexer,CurSrv,Doc,&RealURL,aliastr))) 
		return(result);
	if((result=UdmAliasURL(Indexer,Doc,&RealURL,aliastr,sizeof(aliastr)))) 
		return(result);

	/* Compose HTTP/1.0 request header */
	Doc->request=(char*)malloc(1024);

	if(CurSrv->check_mp3_tag)Doc->method=UDM_CHECKMP3;

	Doc->size=UDM_NET_UNKNOWN;
	UdmBuildHTTPRequest(Indexer->Conf,CurSrv,Doc,&RealURL);

	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Getting",Doc->url);

	/* If mirroring is enabled */
	if (CurSrv->use_mirror >= 0) {
		char errstr[1024]="";
		/* on u_m==0 it returned by mtime from mirrorget */
		/* but we knew that it should be put in mirror  */

		Doc->size = UdmMirrorGET(Indexer,
				CurSrv->mirror_root,CurSrv->mirror_headers,
				RealURL.schema, RealURL.hostname,
				RealURL.path, RealURL.filename,
				Indexer->buf,CurSrv->use_mirror,errstr);
		if(Doc->size>0){
			UdmLog(Indexer,UDM_LOG_DEBUG,"%s has been taken from mirror",Doc->url);
			found_in_mirror=1;
		}else{
			UdmLog(Indexer,UDM_LOG_DEBUG,"%s",errstr);
		}
	}

	/* Get it from the source  */
	/* if no mirror copy found */
	if(!found_in_mirror){
		int is_htdb=0;
#ifdef USE_HTDB
		if(!strcasecmp(RealURL.schema,"htdb")){
			is_htdb=1;
			Doc->size = UdmHTDBGet(Indexer,RealURL.path,RealURL.filename,CurSrv->htdb_list,CurSrv->htdb_doc);
			if(UdmDBErrorCode(Indexer->db))
				return(IND_ERROR);
		}
#endif
		if(!is_htdb){
			UdmGetURL(Indexer,Doc,&RealURL);
		}
	}

	result=IND_OK;
	switch(Doc->size){
		case UDM_NET_UNKNOWN:
			UdmLog(Indexer,UDM_LOG_WARN,"Protocol not supported");
			Doc->status=UDM_HTTP_STATUS_NOT_SUPPORTED;
			break;
		case UDM_NET_ERROR:
			UdmLog(Indexer,UDM_LOG_WARN,"Network error");
			Doc->status=UDM_HTTP_STATUS_TIMEOUT;
			break;
		case UDM_NET_TIMEOUT:
			UdmLog(Indexer,UDM_LOG_WARN,"Download timeout");
			Doc->status=UDM_HTTP_STATUS_TIMEOUT;
			break;
		case UDM_NET_CANT_CONNECT:
			UdmLog(Indexer,UDM_LOG_WARN,"Can't connect to host %s:%d",
			CurSrv->proxy?CurSrv->proxy:RealURL.hostname,
			CurSrv->proxy?CurSrv->proxy_port:(RealURL.port?RealURL.port:RealURL.default_port));
			Doc->status=UDM_HTTP_STATUS_UNAVAIL;
			break;
		case UDM_NET_CANT_RESOLVE:
			UdmLog(Indexer,UDM_LOG_WARN,"Unknown %shost '%s'",
				CurSrv->proxy?"proxy ":"",
				CurSrv->proxy?CurSrv->proxy:RealURL.hostname);
			Doc->status=UDM_HTTP_STATUS_UNAVAIL;
			break;
		default:
			if(Doc->size<0){	/* No connection */
				UdmLog(Indexer,UDM_LOG_WARN,"Can't connect to host %s:%d",
					CurSrv->proxy?CurSrv->proxy:RealURL.hostname,
					CurSrv->proxy?CurSrv->proxy_port:(RealURL.port?RealURL.port:RealURL.default_port));
				Doc->status=UDM_HTTP_STATUS_UNAVAIL;
			}else{
				/* Document has been successfully loaded */
				/* Add NULL termination */
				Indexer->buf[Doc->size]=0;
				/* Then parse headers */
				UdmParseHTTPResponse(Indexer,Doc);
				if(!Doc->content){
					UdmLog(Indexer,UDM_LOG_ERROR,"Illegal HTTP headers in response");
					Doc->status=UDM_HTTP_STATUS_UNAVAIL;
				}else{
#ifdef HAVE_ZLIB
				        if (Doc->content_encoding == UDM_ENCODING_UNKNOWN) {
#else
				        if (Doc->content_encoding != UDM_ENCODING_IDENTITY) {
#endif
					  UdmLog(Indexer,UDM_LOG_ERROR,"Unsupported Content-Encoding");
					  Doc->status=UDM_HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE;
					} else {
					  /* Continue parsing */
					  result=IND_UNKNOWN;
					}
				}
			}
	}

	/* Update If everything already done         */
	/* i.e. there was an error while downloading */
	if(result==IND_OK){
		UdmAddServerErr(Indexer->Conf,RealURL.hostname);
		Doc->next_index_time=time(NULL)+CurSrv->net_error_delay_time;
		result=UdmSUpdateURL(Indexer,Doc);
		return(result);
	}

	UdmLog(Indexer,UDM_LOG_EXTRA,"Status: %d %s, type: %s, Size: %d", Doc->status, UdmHTTPErrMsg(Doc->status),
	       Doc->content_type, Doc->size);

	realsize=Doc->size;
	Doc->size-=(Doc->content-Indexer->buf);     /* Can be modified by Content-Length   */
	realsize-=(Doc->content-Indexer->buf); /* Will be safe for crc32, Parsers etc */


	switch(Doc->content_encoding) {
#ifdef HAVE_ZLIB
	case UDM_ENCODING_GZIP:
	  realsize = Doc->size = UdmUnGzip(Indexer, Doc->content, (size_t)Doc->size);
	  break;
	case UDM_ENCODING_DEFLATE:
	  realsize = Doc->size = UdmInflate(Indexer, Doc->content, (size_t)Doc->size);
	  break;
	case UDM_ENCODING_COMPRESS:
	  realsize = Doc->size = UdmUncompress(Indexer, Doc->content, (size_t)Doc->size);
	  break;
#endif
	default:
	}
	Doc->content[Doc->size]=0;


	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Parsing",Doc->url);

	/*
		FIXME!!! What is it?
	if((SQLDate=UdmParseDate(HeaderDate))){
		UdmDBEscStr(Indexer->Conf->DBType,HeaderDateEsc,SQLDate);
		free(SQLDate);
	}
	*/

	switch(UdmHTTPResponseType(Doc->status)){
	case 1: /* No HTTP code */
		UdmAddServerErr(Indexer->Conf,RealURL.hostname);
		UdmLog(Indexer,UDM_LOG_ERROR,"No HTTP response status");
		Doc->next_index_time=time(NULL)+CurSrv->net_error_delay_time;
		result=UdmSUpdateURL(Indexer,Doc);
		break;

	case UDM_HTTP_STATUS_OK:
	case UDM_HTTP_STATUS_PARTIAL_OK:
		if(!Doc->content_type){
			UdmLog(Indexer,UDM_LOG_ERROR,"No Content-type in '%s'!",Doc->url);
			UdmAddServerErr(Indexer->Conf,RealURL.hostname);
			Doc->next_index_time=time(NULL)+CurSrv->net_error_delay_time;
			result=UdmSUpdateURL(Indexer,Doc);
		}
		break;

	case UDM_HTTP_STATUS_REDIRECT:	   /* Moved to another location  */
	case UDM_HTTP_STATUS_NOT_MODIFIED: /* Not Modified,nothing to do */
		Doc->next_index_time=time(NULL)+CurSrv->period;
		result=UdmSUpdateURL(Indexer,Doc);
		break;

	case UDM_HTTP_STATUS_DELETE:
		if ((Doc->status==404) && 
		    (!strcmp(CurURL.path, "/")) &&
		    (!strcmp(CurURL.filename,"robots.txt"))){
		    
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
			UdmParseRobot(&Indexer->Conf->Robots,NULL,CurURL.hostinfo);
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
		}else{
			/* delete it if not robots.txt */
			UdmLog(Indexer,UDM_LOG_EXTRA,"Deleting URL");
			if(CurSrv->deletebad){
				result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
			}else{
				Doc->next_index_time=time(NULL)+CurSrv->period;
				result=UdmSUpdateURL(Indexer,Doc);
			}
		}
		break;

	case UDM_HTTP_STATUS_RETRY: /* We'll retry later, maybe host is down */
		UdmAddServerErr(Indexer->Conf,RealURL.hostname);
		UdmLog(Indexer,UDM_LOG_EXTRA,"Could not read URL");
		Doc->next_index_time=time(NULL)+CurSrv->net_error_delay_time;
		result=UdmSUpdateURL(Indexer,Doc);
		break;

	default: /* Unknown status, retry later */
		UdmLog(Indexer,UDM_LOG_WARN,"HTTP %d We don't yet know how to handle it, skipped",Doc->status);
		Doc->next_index_time=time(NULL)+CurSrv->period;
		result=UdmSUpdateURL(Indexer,Doc);
	}

	/* Return if Content parsing is not required */
	if(result){
		/* From Location header */
		if(result==IND_OK)result=UdmDocStoreHrefs(Indexer,Doc);
		return(result);
	}

	/* Now we have HTTP_OK and know Content-Type */
	/* Lets start parse the body                 */

	/* Put into mirror if required */
	if ((CurSrv->use_mirror>=0)&&(!found_in_mirror)){
		char errstr[1024]="";
		/* I tried to get it from mirror, but nothing was there  */
		/* or it was expired. Now we will store it on the mirror */
		if(UdmMirrorPUT(Indexer,
				CurSrv->mirror_root,CurSrv->mirror_headers,
				RealURL.schema, RealURL.hostname,
				RealURL.path, RealURL.filename,
				Doc->headers, Doc->content, realsize,errstr)){
			UdmLog(Indexer,UDM_LOG_DEBUG,"%s",errstr);
		}
	}

#ifdef USE_PARSER
	/* Let's try to start external parser for this Content-Type */
	if((mime=UdmExecParser(Indexer,Doc->content_type,&mimeno,Doc->content,(size_t)realsize,(size_t)(Indexer->Conf->max_doc_size-(Doc->content-Indexer->buf)),Doc->url))){
		char * to_charset;
		Doc->content=mime;
		real_content_type=Indexer->Conf->parsers[mimeno].to_mime;
		UdmLog(Indexer,UDM_LOG_DEBUG,"Parser-Content-Type: %s",real_content_type?real_content_type:"unknown");
		if((to_charset=strstr(Indexer->Conf->parsers[mimeno].to_mime,"charset="))){
			UDM_FREE(Doc->charset);
			Doc->charset=strdup(to_charset+8);
#ifdef DEBUG_PARSER
			fprintf(stderr,"to_charset='%s'\n",to_charset+8);
#endif
		}
#ifdef DEBUG_PARSER
		fprintf(stderr,"content='%s'\n",Doc->content);
#endif
	};
#endif

	if(!real_content_type)real_content_type=Doc->content_type;

#ifdef HAVE_ZLIB
	if ((Indexer->Conf->stored_addr) && (Doc->url_id > 0)) {
	  UdmStoreDoc(Indexer->Conf, Doc->content, (size_t)Doc->size, (unsigned int)UdmStrCRC32(Doc->url));
	}
#endif

	/* plain TEXT or the same */
	if(!UDM_STRNCASECMP(real_content_type,"text/plain")||
		!UDM_STRNCASECMP(real_content_type,"text/tab-separated-values")||
		!UDM_STRNCASECMP(real_content_type,"text/css"))
	{
		if(Doc->method==UDM_GET){
			if(strcmp(CurURL.filename,"robots.txt")){
				int crc32;

				crc32=Doc->crc32;
				UdmParseText(Indexer,Doc,CurSrv,&CurURL);
				changed=(crc32!=Doc->crc32);
				if(CurSrv->use_clones){
					UdmURLAction(Indexer,Doc,UDM_URL_ACTION_FINDORIG);
				}
			}else{
				reindex=changed=0; /* To avoid LUpdate() */
				if(CurSrv->use_robots){
					if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
					UdmParseRobot(&Indexer->Conf->Robots,Doc->content,CurURL.hostinfo);
					if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
				}
			}
		}
	}else

	/* HTML text */
	if(!UDM_STRNCASECMP(real_content_type,"text/html")){
		if(Doc->method==UDM_GET){
			int crc32=0;
			crc32=Doc->crc32;
			UdmParseHtml(Indexer,Doc,CurSrv,&CurURL);
			changed=(crc32!=Doc->crc32);
			if(CurSrv->use_clones){
				UdmURLAction(Indexer,Doc,UDM_URL_ACTION_FINDORIG);
			}
		}
	}else{
		/* Unknown Content-Type */
		if(Doc->method==UDM_GET){
			int crc32=0;

			crc32=UdmCRC32(Doc->content, (size_t)realsize);
			changed=(crc32!=Doc->crc32);
			Doc->crc32=crc32;
			if(CurSrv->use_clones){
				UdmURLAction(Indexer,Doc,UDM_URL_ACTION_FINDORIG);
			}
		}
	}


	UDM_THREADINFO(Indexer->Conf->ThreadInfo,Indexer->handle,"Updating",Doc->url);
	if((Doc->origin_id)&&(Doc->origin_id!=Doc->url_id)){
		UdmLog(Indexer,UDM_LOG_EXTRA,"Duplicate Document with #%d",Doc->origin_id);
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELWORDS);
		if((Indexer->Conf->use_crossword)&&(Indexer->Conf->DBMode!=UDM_DBMODE_CACHE)){
			if(result==IND_OK)result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELCWORDS);
		}
		Doc->next_index_time=time(NULL)+CurSrv->period;
		if(result==IND_OK)result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_UPDCLONE);
	}else
	if((!changed)&&(!reindex)){
		Doc->next_index_time=time(NULL)+CurSrv->period;
		result=UdmSUpdateURL(Indexer,Doc);
	}else{
		char site_id_str[UDM_URLSIZE];

		sprintf(site_id_str,"%s://%s/",CurURL.schema,CurURL.hostinfo);
		Doc->site_id=UdmStrCRC32(site_id_str);
#ifdef DEBUG_SITE_ID
		fprintf(stderr,"SIDE_ID '%s' -> %08X\n",site_id_str,site_id);
#endif
		if(result!=IND_ERROR){
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
			if(UdmStoreHrefs(Indexer))
				Indexer->Conf->have_targets=1;
			if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
			if(UdmDBErrorCode(Indexer->db))
				result=IND_ERROR;
			else
				result=IND_OK;
		}
		if(result!=IND_ERROR){
			if((!CurSrv->check_only_mp3_tag)||(Doc->is_mp3)){
				if(Indexer->Conf->DBMode!=UDM_DBMODE_BLOB){
					if(result!=IND_ERROR)result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_INSWORDS);
				}else{
					if(result!=IND_ERROR)result=UdmBlobWriteStamp(Indexer,Doc);
				}					

				/* Store URLs from cache into database */
				/* This is for StoreCrossWords here    */
				if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
				if(UdmStoreHrefs(Indexer))Indexer->Conf->have_targets=1;
				if(UdmDBErrorCode(Indexer->db))result=IND_ERROR;
				if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
				if(Indexer->Conf->use_crossword&&Indexer->Conf->DBMode!=UDM_DBMODE_CACHE){
					if(result!=IND_ERROR)result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_INSCWORDS);
				}
			}
		}
		if(result!=IND_ERROR){
			
			/* get rec_id from my parent out of db (if I have one...) */
			if(Indexer->Conf->use_newsext){
				char *parent=NULL;
				
				Doc->parent_id=0;
				if(Doc->references){
					/* References contains all message IDs of my */
					/* predecessors, space separated             */
					/* my direct parent is the last in the list  */
					if((parent = strrchr(Doc->references,' '))){
						/* parent now points to the space */
						/* character skip it              */
						++parent;
					}else{
						/* there is only one entry in */
						/* references, so this is my parent */
						parent=Doc->references;
					}	
				}
				
				/* get parent from database */
				if(parent && strlen(parent) && strchr(parent,'@')){
					char * pend, *pbeg;
					UDM_DOCUMENT Msg;
					
					/* Parent now points to the string in */
					/* format    <blabla@blabla.bla>      */
					/* Lets remove < > characters         */
					
					pbeg=parent=strdup(parent);
					if((pend=strrchr(parent,'>')))*pend=0;
					if(pbeg[0]=='<')pbeg++;
					
					UdmInitDoc(&Msg);
					Msg.message_id=pbeg;
					result = UdmURLAction(Indexer,&Msg,UDM_URL_ACTION_FINDBYMSG);
					Doc->parent_id = Msg.url_id;
					free(parent);
				}

				/* now register me with my parent  */
				if(Doc->parent_id)result = UdmURLAction(Indexer,Doc,UDM_URL_ACTION_REGCHILD);
			}
			Doc->next_index_time=time(NULL)+CurSrv->period;
			Doc->tag=CurSrv->tag?strdup(CurSrv->tag):NULL;
			Doc->category=CurSrv->category?strdup(CurSrv->category):NULL;
			if(!Doc->lang){
				if(CurSrv->lang){
					Doc->lang=strdup(CurSrv->lang);
				}
			}
			if(result==IND_OK)result=UdmLUpdateURL(Indexer,Doc);
		}
	}
	if(result==IND_OK)result=UdmDocStoreHrefs(Indexer,Doc);
	
	return(result);
}

__INDLIB__ int UdmIndexNextURL(UDM_AGENT *Indexer,int index_flags){
	int result=IND_UNKNOWN;
	UDM_DOCUMENT	*Doc=NULL;
	UDM_URL		CurURL;

	/* Do init stuff if required */
	if(index_flags&UDM_FLAG_INIT){
		if(StartUp(Indexer, index_flags))
			return(IND_ERROR);
		else
			return(IND_OK);
	}

	if(!(Doc=UdmGetTarget(Indexer,&result,index_flags))){
		if(Indexer->Conf->DBMode==UDM_DBMODE_BLOB){
			int res;
			Indexer->Conf->flashmode=UDM_FLASH_ALL;
			res=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_INSWORDS);
			Indexer->Conf->flashmode=UDM_FLASH_BEST;
			if(res!=IND_OK)return(res);
		}
		/* No targets or error */
		return(result);
	}

	/* Check that URL has valid syntax */
	if(UdmParseURL(&CurURL,Doc->url)){ /* FIXME twice */
		result=UdmURLAction(Indexer,Doc,UDM_URL_ACTION_DELETE);
		UdmLog(Indexer,UDM_LOG_WARN,"Invalid URL: %s ... deleting",Doc->url);
		UdmFreeDoc(Doc);
		return(result);
	}

	if(!strcmp(CurURL.schema,"http")||!strcmp(CurURL.schema,"https")){
		int has_robots=0;
		
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_ROBOTS);
		has_robots=(NULL!=UdmGetRobot(&Indexer->Conf->Robots,CurURL.hostinfo));
		if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_ROBOTS);
		
		if(!has_robots){
			char rurl[UDM_URLSIZE];
			UDM_DOCUMENT *rDoc = UdmAllocDoc();

			if (rDoc != NULL) {
			  snprintf(rurl, UDM_URLSIZE, "%s://%s/robots.txt", CurURL.schema, CurURL.hostinfo);
			  rDoc->url = strdup(rurl);
			  result = UdmIndexURL(Indexer, rDoc, index_flags);
			  UdmFreeDoc(rDoc);
  			}
		}
	}

	result=UdmIndexURL(Indexer,Doc,index_flags);

	UdmFreeDoc(Doc);

	return(result);
}


/************ Misc functions *****************************/



__INDLIB__ int UdmClearDatabase(UDM_AGENT * Indexer){
	return(UdmClearDB(Indexer));
}

