#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_config.h"
#include "udmsearch.h"
#include "udm_crc32.h"
#include "udm_xmalloc.h"

/*
	TODO:
		- categories
		- pass tag,cat,ul arguments
*/

#ifndef UDM_CONF_DIR
#define UDM_CONF_DIR "/usr/local/mnogosearch/etc/"
#endif


/******************** Misc functions *********************************/

static char * BuildPageURL(UDM_VARLIST * vars,char * dst){
	int i,nargs=0;
	char * end=dst;

	strcpy(dst,UdmFindStrVar(vars,"self",""));
	end=dst+strlen(dst);

	for(i=0;i<vars->nvars;i++){
		if(vars->var[i].scope==UDM_VARSRC_QSTRING){
			strcpy(end,nargs?"&":"?");
			end++;
			UdmEscapeURL(end,vars->var[i].name);
			end=end+strlen(end);
			strcpy(end,"=");end++;
			UdmEscapeURL(end,vars->var[i].val);
			end=end+strlen(end);
			nargs++;
		}
	}
	return NULL;
}


/*****************************************************************/

int main(int argc, char ** argv){
	const char * env, *bcharset, *lcharset;
	char template_name[1024]="";
	char query_string[1024]="";
	const char * qw=NULL;
	char self[1024]="";
	TMPL tmpl[T_MAX];
	int res;
	int httpd=0;
	int to_recode = 0;
	UDM_ENV    * Env;
	UDM_AGENT  * Agent;
	UDM_RESULT * Res;
	UDM_CHARSET *bcs = NULL, *lcs = NULL;
	
	/* Output Content-type if under HTTPD	 */
	/* Some servers do not pass QUERY_STRING */
	/* if the query was empty, so check	 */
	/* REQUEST_METHOD too     to be safe     */
	
	if(getenv("QUERY_STRING")||getenv("REQUEST_METHOD")){
		httpd=1;
	}
	
	UdmInit();
	UdmTemplateInit(tmpl);
	Env=UdmAllocEnv();
	Env->vars=UdmAllocVarList();

	/* Detect self and template name */
	if((env=getenv("UDMSEARCH_TEMPLATE")))
		strncpy(template_name,env,sizeof(template_name)-1);

	if((env=getenv("UDMSEARCH_SELF")))
		strncpy(self,env,sizeof(self)-1);

	if((env=getenv("QUERY_STRING"))){
		strncpy(query_string,env,sizeof(query_string)-1);
		if((env=getenv("REDIRECT_STATUS"))){

			/* Check Apache internal redirect  */
			/* via   "AddHandler" and "Action" */
			if(!self[0]){
				strncpy(self,(env=getenv("REDIRECT_URL"))?env:"search.cgi",sizeof(self)-1);
			}
			if(!template_name[0]){
				strncpy(template_name,(env=getenv("PATH_TRANSLATED"))?env:"",sizeof(template_name)-1);
			}
		}else{
			/* CGI executed without Apache internal redirect */

			/* Detect $Self variable with OS independant SLASHES */
			if(!self[0]){
				strncpy(self,(env=getenv("SCRIPT_NAME"))?env:"search.cgi",sizeof(self)-1);
			}

			if(!template_name[0]){
				char *s,*e;
				
				/*This is with OS specific SLASHES */
				env=((env=getenv("SCRIPT_FILENAME"))?env:"search.cgi");

				if(strcmp(UDM_CONF_DIR,".")){
					/* Take from the config directory */
					snprintf(template_name,sizeof(template_name)-1,"%s/%s", UDM_CONF_DIR,(s=strrchr(env,UDMSLASH))?(s+1):(self));
				}else{
					/* Take from the current directory */
					strncpy(template_name,env,sizeof(template_name)-1);
				}

				/* Find right slash if it presents */
				s=((s=strrchr(template_name,UDMSLASH))?s:template_name);

				/* Find .cgi substring */
				if((e=strstr(s,".cgi"))){
					/* Replace ".cgi" with ".htm" */
					e[1]='h';e[2]='t';e[3]='m';
				}else{
					strcpy(template_name,"search.htm");
				}
			}
		}
	}else{
		/* Executed from command line     */
		/* or under server which does not */
		/* pass an empty QUERY_STRING var */
		if(argv[1])snprintf(query_string,sizeof(query_string)-1,"q=%s",argv[1]);
		if(!template_name[0])snprintf(template_name,sizeof(template_name),"%s/%s", UDM_CONF_DIR,"search.htm");
	}

	bzero(tmpl,sizeof(tmpl));
	res=UdmTemplateLoad(Env,Env->vars,template_name,tmpl);
	if(res){
		if(httpd)printf("Content-Type: text/plain\r\n\r\n");
		printf("%s\n",Env->errstr);
		UdmTemplateFree(tmpl);
		UdmFreeEnv(Env);
		return(0);
	}
	
	Env->DBAddr=strdup(UdmFindStrVar(Env->vars,"DBAddr","mysql://foo:bar@localhost/mnogosearch/"));
	UdmEnvSetDBMode(Env,UdmFindStrVar(Env->vars,"DBMode","single"));

	Agent=UdmAllocAgent(Env,0,UDM_OPEN_MODE_READ);
	{
		size_t i;
		for(i=0;i<Env->sdcl.nclients;i++){
			UdmSDCLientListAdd(&Agent->sdcl,Env->sdcl.Clients[i].addr);
		}
	}
	
	UdmAddStrVar(Env->vars,"QUERY_STRING",query_string,UDM_VARSRC_GLOBAL);
	UdmAddStrVar(Env->vars,"self",self,UDM_VARSRC_GLOBAL);
	
	UdmParseQueryString(Agent,Env->vars,query_string);
	
	
	bcharset=UdmFindStrVar(Env->vars,"BrowserCharset","utf-8");
	bcs=UdmGetCharSet(bcharset);
	lcharset=UdmFindStrVar(Env->vars,"LocalCharset","utf-8");
	lcs=UdmGetCharSet(lcharset);

	if(httpd){
		if(!bcs){
			printf("Content-Type: text/plain\r\n\r\n");
			printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
			exit(0);
		}else if(!lcs){
			printf("Content-Type: text/plain\r\n\r\n");
			printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
			exit(0);
		}else{
			printf("Content-type: text/html; charset=%s\r\n\r\n",bcharset);
		}
	}else{
		if(!bcs){
			printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
			exit(0);
		}
		if(!lcs){
			printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
			exit(0);
		}
	}

	/* These parameters taken from "variable section of template"*/
	if((env=UdmFindStrVar(Env->vars,"VarDir",NULL))){
		snprintf(Agent->Conf->vardir,sizeof(Agent->Conf->vardir)-1,"%s%s",env,UDMSLASHSTR);
	}
	

	Agent->page_size=UdmFindIntVar(Env->vars,"ps",20);
	Agent->page_number=UdmFindIntVar(Env->vars,"np",0);
	UdmAddIntVar(Env->vars,"pn",(int)Agent->page_number*Agent->page_size,UDM_VARSRC_GLOBAL);



	/* Now start displaying template*/
	UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_TOP);

	qw=UdmFindStrVar(Env->vars,"q","");
	if ((to_recode = strcmp(lcharset, bcharset)) != 0) { /* recode query to localcharset, if need */
	  int *ustr;
	  size_t len;
	  ustr = (int*)UdmXmalloc((len = strlen(qw) + 1) * 8);
	  bcs->to_uni(ustr, qw, len); UdmUniStrToLower(ustr);
	  qw = (const char*)UdmXmalloc(len * 8);
	  lcs->from_uni(qw, ustr, len);
	  free(ustr);
	}

	if(!qw[0]){
		UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_NOQUERY);
	}else{
		if((Res=UdmFind(Agent,qw))){
			char search_time[100];
			UdmAddIntVar(Env->vars,"first",(int)Res->first,UDM_VARSRC_GLOBAL);
			UdmAddIntVar(Env->vars,"last",(int)Res->last,UDM_VARSRC_GLOBAL);
			UdmAddIntVar(Env->vars,"total",(int)Res->total_found,UDM_VARSRC_GLOBAL);
			UdmAddStrVar(Env->vars,"W",Res->wordinfo,UDM_VARSRC_GLOBAL);
			sprintf(search_time,"%.3f",((double)Res->work_time)/1000);
			UdmAddStrVar(Env->vars,"SearchTime",search_time,UDM_VARSRC_GLOBAL);
			
			if(Res->num_rows){
				int i,page1,page2,npages;
				int ppp=10;
				char nav[4096]=""; /* FIXME */
				char url[4096]=""; /* FIXME */
				char *searchwords = NULL;
				size_t swlen = 0;
				char *storedstr = NULL;
				
				for (i = 0; i < Agent->WWList.nwords; i++) {
				  swlen += strlen(Agent->WWList.Words[i].word + 2);
				}
				if ((searchwords = UdmXmalloc(swlen)) != NULL) {
				  int z=0;
				  for (i = 0; i < Agent->WWList.nwords; i++) {
				    if (Agent->WWList.Words[i].count > 0) {
				      sprintf(UDM_STREND(searchwords), (z)?" %s":"%s", Agent->WWList.Words[i].word);
				      z++;
				    }
				  }
				}
				storedstr = UdmXmalloc(3*(swlen + UDM_URLSIZE));

				npages=(Res->total_found/Agent->page_size)+1;
				
				page1=Agent->page_number-ppp/2;
				page2=Agent->page_number+ppp/2;
				if(page1<0){
					page2-=page1;
					page1=0;
				}else
				if(page2>npages){
					page1-=(page2-npages);
					page2=npages;
				}else{
				}
				if(page1<0)page1=page1=0;
				if(page2>npages)page2=npages;
				
				for(i=page1;i<page2;i++){
					UdmReplaceIntVar(Env->vars,"np",i,UDM_VARSRC_QSTRING);
					BuildPageURL(Env->vars,url);
					UdmReplaceStrVar(Env->vars,"NH",url,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"NP",i+1,UDM_VARSRC_GLOBAL);
					UdmTemplatePrint(Agent,NULL,UDM_STREND(nav),Env->vars,tmpl,(i==Agent->page_number)?T_NAVBAR0:T_NAVBAR1);
				}
				UdmAddStrVar(Env->vars,"NB",nav,UDM_VARSRC_GLOBAL);

				UdmReplaceIntVar(Env->vars,"np",Agent->page_number-1,UDM_VARSRC_QSTRING);
				BuildPageURL(Env->vars,url);
				UdmReplaceStrVar(Env->vars,"NH",url,UDM_VARSRC_GLOBAL);
				
				if(Res->first==1){/* First page */
					UdmTemplatePrint(Agent,NULL,nav,Env->vars,tmpl,T_NAVLEFT_NOP);
					UdmReplaceStrVar(Env->vars,"NL",nav,UDM_VARSRC_GLOBAL);
				}else{
					UdmTemplatePrint(Agent,NULL,nav,Env->vars,tmpl,T_NAVLEFT);
					UdmReplaceStrVar(Env->vars,"NL",nav,UDM_VARSRC_GLOBAL);
				}

				UdmReplaceIntVar(Env->vars,"np",Agent->page_number+1,UDM_VARSRC_QSTRING);
				BuildPageURL(Env->vars,url);
				UdmReplaceStrVar(Env->vars,"NH",url,UDM_VARSRC_GLOBAL);

				if(Res->last==Res->total_found){/* Last page */
					UdmTemplatePrint(Agent,NULL,nav,Env->vars,tmpl,T_NAVRIGHT_NOP);
					UdmReplaceStrVar(Env->vars,"NR",nav,UDM_VARSRC_GLOBAL);
				}else{
					UdmTemplatePrint(Agent,NULL,nav,Env->vars,tmpl,T_NAVRIGHT);
					UdmReplaceStrVar(Env->vars,"NR",nav,UDM_VARSRC_GLOBAL);
				}

				UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_RESTOP);
				for(i=0;i<Res->num_rows;i++){
					UDM_RESULT   * ClRes;
					UDM_DOCUMENT * Doc=&Res->Doc[i];
					UDM_ALIAS    * Alias;
					char clist[2048]=""; /* FIXME */
					char last_mod[128];
					char aliastr[UDM_URLSIZE];
					const char *u, *dm;
					char *eu, *edm;

					/* Build clone list */
					if((ClRes=UdmCloneList(Agent,Doc))){
						int c;
						
						for(c=0;c<ClRes->num_rows;c++){
							UDM_DOCUMENT * Clone=&ClRes->Doc[c];
							if(Res->Doc[i].url_id!=Clone->url_id){
								UdmReplaceIntVar(Env->vars,"DI",Clone->url_id,UDM_VARSRC_GLOBAL);
								UdmReplaceStrVar(Env->vars,"DU",Clone->url,UDM_VARSRC_GLOBAL);
								UdmReplaceStrVar(Env->vars,"DC",Clone->content_type,UDM_VARSRC_GLOBAL);
								UdmReplaceIntVar(Env->vars,"DS",Clone->size,UDM_VARSRC_GLOBAL);
								UdmTime_t2HttpStr(Clone->last_mod_time,last_mod);
								UdmReplaceStrVar(Env->vars,"DM",last_mod,UDM_VARSRC_GLOBAL);
								UdmTemplatePrint(Agent,NULL,clist,Env->vars,tmpl,T_CLONE);
								
								/*sprintf(UDM_STREND(clist),"%s",Clone->url);*/
							}
						}
						UdmFreeResult(ClRes);
					}
					UdmReplaceStrVar(Env->vars,"CL",clist,UDM_VARSRC_GLOBAL);

					UdmReplaceIntVar(Env->vars,"DI",Doc->url_id,UDM_VARSRC_GLOBAL);
					if((Doc->url)&&(Alias=UdmAliasFind(&Agent->Conf->AliasList,Doc->url,0,NULL))){
						snprintf(aliastr,sizeof(aliastr),"%s%s",Alias->replace,Doc->url+strlen(Alias->match.pattern));
					}else{
						snprintf(aliastr,sizeof(aliastr),"%s",Doc->url?Doc->url:"");
					}

					UdmReplaceStrVar(Env->vars,"DU",aliastr,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DT",Doc->title,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DX",Doc->text,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DC",Doc->content_type,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DD",Doc->description,UDM_VARSRC_GLOBAL);
					UdmReplaceStrVar(Env->vars,"DK",Doc->keywords,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"DR",Doc->rating,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"DS",Doc->size,UDM_VARSRC_GLOBAL);
					UdmReplaceIntVar(Env->vars,"DN",Doc->order,UDM_VARSRC_GLOBAL);
					UdmTime_t2HttpStr(Doc->last_mod_time,last_mod);
					UdmReplaceStrVar(Env->vars,"DM",last_mod,UDM_VARSRC_GLOBAL);
					
					/* DE is DD or DT */
					if(Doc->description&&Doc->description[0])
						UdmReplaceStrVar(Env->vars,"DE",Doc->description,UDM_VARSRC_GLOBAL);
					else
						UdmReplaceStrVar(Env->vars,"DE",Doc->text,UDM_VARSRC_GLOBAL);

					u =  UdmFindStrVar(Env->vars, "DU", "");
					eu = (char*)malloc(strlen(u)*3);
					UdmEscapeURL(eu, u);
					dm = UdmFindStrVar(Env->vars, "DM", last_mod);
					edm = (char*)malloc(strlen(dm)*3);
					UdmEscapeURL(edm, dm);
					snprintf(storedstr, 3*(swlen + UDM_URLSIZE), "%s?rec_id=%u&DM=%s&DS=%d&L=%s&CS=%s&DU=%s&q=%s",
						 UdmFindStrVar(Env->vars, "StoredocURL", "/cgi-bin/storedoc.cgi"),
						 UdmStrCRC32(Doc->url),
						 edm,
						 UdmFindIntVar(Env->vars, "DS", Doc->size),
						 (Doc->lang)?Doc->lang:"",
						 (Doc->charset)?Doc->charset:"",
						 eu,
						 searchwords
						 );
					free(eu);
					free(edm);
					UdmReplaceStrVar(Env->vars, "stored", storedstr, UDM_VARSRC_GLOBAL);

					UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_RESULT);
				}
				UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_RESBOT);
				UDM_FREE(searchwords);
				UDM_FREE(storedstr);
			}else{
				UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_NOTFOUND);
			}
			UdmFreeResult(Res);
		}else{
			if(UdmDBErrorCode(Agent->db)){
				UdmAddStrVar(Env->vars,"E",UdmDBErrorMsg(Agent->db),UDM_VARSRC_GLOBAL);
				UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_ERROR);
			}
		}
	}

	UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_BOTTOM);

	UdmTemplateFree(tmpl);
	UdmFreeAgent(Agent);
	UdmFreeEnv(Env);

	return(0);
} 
