#include "udm_config.h"

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <sys/stat.h>
#include <stdio.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_unicode.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"
#include "udm_xmalloc.h"
#include "udm_spell.h"
#include "udm_stopwords.h"
#include "udm_unicode.h"
#include "udm_word.h"
#include "udm_vars.h"
#include "udm_cache.h"
#include "udm_db.h"
#include "udm_url.h"
#include "udm_crc32.h"

static const size_t search_cache_size=1000;

static int nbits[256] = {
  0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
  4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8,
};

static size_t bit_count(unsigned int v)
{
  return (nbits[v&0xFF]+nbits[(v>>8)&0xFF]+nbits[(v>>16)&0xFF]+nbits[(v>>24)&0xFF]);
}


/********** QSORT functions *******************************/

static int cmpword(UDM_URL_CRD *s1,UDM_URL_CRD *s2){
int res;
	if(!(res=s2->coord-s1->coord))
		if(!(res=s1->url_id-s2->url_id));
	return(res);
}

static int cmpurlid(UDM_URL_CRD *s1,UDM_URL_CRD *s2){
	int res;
	if(!(res=(s1->url_id-s2->url_id)))
		if(!(res=(UDM_WRDPOS(s1->coord)-UDM_WRDPOS(s2->coord))));
	return(res);
}

/***************** Results cache functions **************/

int UdmStoreToCache(UDM_AGENT * query,char * fname,UDM_URL_CRD *wrd, size_t nwrd){
	int fd;
	char info[1024]="";

	if(query->page_number<FAST_PRESORT_PAGES)
		qsort((void*)wrd,nwrd,sizeof(*wrd),(qsort_cmp)cmpword);
#ifdef DEBUG_CACHE
		fprintf(stderr,"write to %s\n",fname);
#endif
		if((fd=open(fname,O_WRONLY|O_CREAT|O_TRUNC|UDM_BINARY,UDM_IWRITE))>=0){
#ifdef DEBUG_CACHE
			fprintf(stderr,"found:%d\n",nwrd);	
#endif
			write(fd,&nwrd,4);
			strncpy(info,query->wordinfo,sizeof(info));
			write(fd,info,sizeof(info));
			if((search_cache_size==0)||(nwrd<search_cache_size)){
				write(fd,wrd,nwrd*sizeof(*wrd));
			}else{
				write(fd,wrd,search_cache_size*sizeof(*wrd));
			}
			close(fd);
		}else{
#ifdef DEBUG_CACHE
			fprintf(stderr,"%s\n",strerror(errno));
#endif
		}
	UDM_FREE(wrd);
	return(0);
}

UDM_URL_CRD* UdmFindInCache(UDM_AGENT * Agent,char * query_id, size_t * nwrd){
	char fname[1024];
	char info[1024]="HZ";
	int fd;
	UDM_URL_CRD *wrd=NULL;
	int bytes;

	*nwrd=0;
	if((search_cache_size!=0)&&((Agent->page_number+1)*Agent->page_size>search_cache_size))return(NULL);

	snprintf(fname,sizeof(fname)-1,"%s",query_id);
	if((fd=open(fname,O_RDONLY|UDM_BINARY))<0)return(NULL);
	if(-1==read(fd,&Agent->total_found,4)){
		close(fd);
		return(NULL);
	}
	if(-1==read(fd,info,sizeof(info))){
		close(fd);
		return(NULL);
	}
	strcpy(Agent->wordinfo,info);
#ifdef DEBUG_CACHE
	fprintf(stderr,"found: %d wordinfo: %s\n",*found,info);
#endif
	wrd=(UDM_URL_CRD*)UdmXmalloc(Agent->page_size*sizeof(*wrd));
	if(-1==lseek(fd,(off_t)(Agent->page_number*Agent->page_size*sizeof(*wrd)),SEEK_CUR)){
		close(fd);
		return(NULL);
	}
	if(-1==(bytes=read(fd,wrd,Agent->page_size*sizeof(*wrd)))){
		close(fd);
		return(NULL);
	}
	close(fd);
	*nwrd=bytes/sizeof(*wrd);

	return(wrd);
}
/****************************************************/

void UdmSortSearchWordsByWeight(UDM_URL_CRD *wrd,size_t num){
	qsort((void*)wrd,num,sizeof(*wrd),(qsort_cmp)cmpword);
	return;
}

void UdmSortSearchWordsByURL(UDM_URL_CRD *wrd,size_t num){
	qsort((void*)wrd,num,sizeof(*wrd),(qsort_cmp)cmpurlid);
	return;
}

/* Find topcount best results */
void UdmWrdTopSort(UDM_URL_CRD *wrd, size_t nwrd,size_t topcount){
	size_t j;
	UDM_URL_CRD w;

#ifdef DEBUG_TOP_SORT
	fprintf(stderr,"top+1=%d nwrd=%d\n",topcount+1,nwrd);
#endif

	qsort((void*)wrd,topcount+1,sizeof(*wrd),(qsort_cmp)cmpword);
	for(j=topcount;j<nwrd;j++){
		register int res;
		if(!(res=(wrd[j].coord-wrd[topcount].coord)))
		if(!(res=(wrd[topcount].url_id-wrd[j].url_id)));
#ifdef DEBUG_TOP_SORT
fprintf(stderr,"(%d,%d) %d (%d,%d) %d\n",
	wrd[topcount].coord,wrd[topcount].url_id,topcount,
	wrd[j].cooord,wrd[j].url_id,j);
#endif
		if(res>0){
			size_t l,c,r;
			
			l=0;r=topcount;
			while(l<r){
				c=(l+r)/2;
				if(!(res=(wrd[c].coord-wrd[j].coord)))
				if(!(res=(wrd[j].url_id-wrd[c].url_id)));
				if(res>0){
					l=c+1;
				}else{
					r=c;
				}
			}
			w=wrd[topcount];
			memmove(&wrd[r+1],&wrd[r],(topcount-r)*sizeof(*wrd));
			wrd[r]=wrd[j];
			wrd[j]=w;
		}
	}
}

int UdmPrepare(UDM_AGENT * query,const char * txt){
	UDM_CHARSET * browser_cs, * local_cs;
	int ures;
	int * ustr, * lt, * lex;
	size_t ulen;
	
	/* FIXME: add free words */
	query->wordinfo[0]='\0';
	query->total_found=0;
	query->nitems=0;

	if(!(browser_cs=UdmGetCharSet(query->Conf->browser_charset?query->Conf->browser_charset:"utf-8"))){
		browser_cs=UdmGetCharSet("utf-8");
	}
	if(!(local_cs=UdmGetCharSet(query->Conf->local_charset?query->Conf->local_charset:"utf-8"))){
		local_cs=UdmGetCharSet("utf-8");
	}
	ulen=strlen(txt);
	ustr=(int*)(malloc((sizeof(int))*(ulen+1)));
	ures=browser_cs->to_uni(ustr,txt,ulen+1);
	UdmUniStrToLower(ustr);
	
	/* Parse query and build boolean search stack*/

	lex=UdmUniGetToken(ustr,&lt);
	while(lex){
		if((lex[0]=='&')||(lex[0]=='|')||(lex[0]=='~')){
			/* Boolean language operator */
			query->items[query->nitems].cmd=lex[0];
			query->items[query->nitems].arg=0;
			query->nitems++;
		}else{
			int addwrd=1;
			size_t wlen;
			char wrd[UDM_MAXWORDSIZE*4+1];
			int uwrd[UDM_MAXWORDSIZE+1];

			query->items[query->nitems].cmd=UDM_STACK_WORD;
			query->items[query->nitems].arg=1<<(query->WWList.nuniq);
			query->nitems++;

			wlen=lt-lex;
			memcpy(uwrd,lex,wlen*sizeof(int));
			uwrd[wlen]=0;
			ures=local_cs->from_uni(wrd,uwrd,sizeof(wrd));

			if(query->word_match==UDM_MATCH_FULL){
				/* Check stopword only when full word         */
				/* Substring searches should not exclude them */
				if(UdmIsStopWord(query->Conf,wrd)||(query->Conf->min_word_len>wlen)||(query->Conf->max_word_len<wlen)){
					if(query->wordinfo[0])strcat(query->wordinfo,", ");
					/* FIXME UNI: convert to browser_cs must be here */
					sprintf(UDM_STREND(query->wordinfo)," %s :stopword",wrd);
					addwrd=0;
				}
			}
			if(query->WWList.nwords>=UDM_MAXWORDPERQUERY-1){
				addwrd=0;
			}
			if(addwrd){
				UDM_WIDEWORD OWord;
				UDM_WIDEWORDLIST * forms;
				
				OWord.len=strlen(wrd);
				OWord.order=query->WWList.nuniq;
				OWord.count=0;
				OWord.crcword=UdmStrCRC32(wrd);
				OWord.word=wrd;
				OWord.uword=uwrd;
				
				if((forms=UdmAllForms(query,&OWord))){
					UDM_WIDEWORD FWord;
					int frm;
					for(frm=0;frm<forms->nwords;frm++){
						local_cs->from_uni(wrd,forms->Words[frm].uword,sizeof(wrd));
						FWord.len=strlen(wrd);
						FWord.order=query->WWList.nuniq;
						FWord.count=0;
						FWord.crcword=UdmStrCRC32(wrd);
						FWord.word=wrd;
						FWord.uword=forms->Words[frm].uword;
						
						UdmWideWordListAdd(&query->WWList,&FWord);
					}
					UdmWideWordListFree(forms);
					free(forms);
				}
				query->WWList.nuniq++;
			}
		}
		lex=UdmUniGetToken(NULL,&lt);
	}
	free(ustr);
		
	return(0);
}

static size_t UdmCalcPhraseWeight(int url_id,size_t weight,size_t uniq_words,size_t *phr,size_t words_in_query,int search_mode){
	size_t res=0;
	if(search_mode==UDM_MODE_PHRASE){
		res=phr[words_in_query-1];
	}else{
		size_t i;

		for(i=0;i<words_in_query-1;i++){
			res+=phr[i]*(i+1);
		}	
		if(res>0xFFFF)res=0xFFFF;
		if(phr[words_in_query-1]>0xFF)phr[words_in_query-1]=0xFF;
		res=(size_t)(phr[words_in_query-1]<<24)|(size_t)(uniq_words<<16)|res;
	}
	return res;
}

void UdmGroupByURL(UDM_AGENT *query,UDM_URL_CRD *wrd){
	size_t	i,j=0;
	size_t	Doc_weight=UDM_WRDSEC(wrd[0].coord);
/* BOOL stuff */
	size_t	item;
	UDM_STACK_ITEM temp_items[UDM_MAXSTACK];
	uint4	count=UDM_WRDMASK(wrd[0].coord);
/* PHRASE stuff */
	size_t	*Doc_phr;
	size_t	phr_size=query->WWList.nwords*sizeof(size_t);
	uint4	phr_beg=UDM_WRDPOS(wrd[0].coord);
	uint4	phr_mask=UDM_WRDMASK(wrd[0].coord);
	uint4	phr_weight=UDM_WRDSEC(wrd[0].coord);

	if(query->search_mode==UDM_MODE_BOOL){
		memcpy(temp_items,query->items,query->nitems*sizeof(UDM_STACK_ITEM));
	}
	Doc_phr=(size_t*)malloc(phr_size);
	bzero(Doc_phr, phr_size);

	for(i=1;i<query->total_found;i++){
		/* Group by url_id */
		if(wrd[j].url_id==wrd[i].url_id){
		/* Same document */
			//??wrd[j].count|=wrd[i].count;
			count|=UDM_WRDMASK(wrd[i].coord);
			
			Doc_weight+=UDM_WRDSEC(wrd[i].coord);
			/* Calculate phrase (subphrase) summary weight */

			if((UDM_WRDPOS(wrd[i].coord)-phr_beg)<=((bit_count(phr_mask)+1))){
				phr_mask|=UDM_WRDMASK(wrd[i].coord);
				phr_weight+=UDM_WRDSEC(wrd[i].coord);
			}else{
				Doc_phr[bit_count(phr_mask)-1]+=phr_weight;
				phr_beg=UDM_WRDPOS(wrd[i].coord);
				phr_mask=UDM_WRDMASK(wrd[i].coord);
				phr_weight=UDM_WRDSEC(wrd[i].coord);
			}
		}else{
			/* Next document */
			Doc_phr[bit_count(phr_mask)-1]+=phr_weight;

			if(query->search_mode==UDM_MODE_BOOL){
				/* Create a copy of lexem array       */
				/* and change word masks by 1 or 0    */
				/* depending on whether word presents */
				/* in the search query being executed */
				for(item=0;item<query->nitems;item++){
					if(temp_items[item].cmd==UDM_STACK_WORD){
						/* Change word mask by 1 or 0 */
						temp_items[item].arg=((query->items[item].arg)&(count))?1:0;
					}
				}
				if(UdmCalcBoolItems(temp_items,query->nitems)){
					wrd[j].coord=UdmCalcPhraseWeight(wrd[j].url_id,Doc_weight,bit_count(count),Doc_phr,query->WWList.nuniq,query->search_mode);
					j++;
				}else{
					/* Skip this result */
				}
			}else
			if((query->search_mode==UDM_MODE_ALL)&&(bit_count(count)<query->WWList.nuniq)){
				/* Skip this result */
			}else
			if((query->search_mode==UDM_MODE_PHRASE)&&(Doc_phr[query->WWList.nwords-1]==0)){
				/* Skip this result */
			}else{
				wrd[j].coord=UdmCalcPhraseWeight(wrd[j].url_id,Doc_weight,bit_count(count),Doc_phr,query->WWList.nuniq,query->search_mode);
				j++;
			}
			Doc_weight=UDM_WRDSEC(wrd[i].coord);
			phr_mask=UDM_WRDMASK(wrd[i].coord);
			phr_beg=UDM_WRDPOS(wrd[i].coord);
			phr_weight=UDM_WRDSEC(wrd[i].coord);
			bzero(Doc_phr, phr_size);
			count=UDM_WRDMASK(wrd[i].coord);
			wrd[j]=wrd[i];
		}
	}

	/* Check last word */
	Doc_phr[bit_count(phr_mask)-1]+=phr_weight;
	
	switch(query->search_mode){
		case UDM_MODE_BOOL:
			for(item=0;item<query->nitems;item++){
				if(temp_items[item].cmd==UDM_STACK_WORD){
					/* Change word mask by 1 or 0 */
					temp_items[item].arg=((query->items[item].arg)&(count))?1:0;
				}
			}
			query->total_found=(UdmCalcBoolItems(temp_items,query->nitems))?j+1:j;
			break;
		case UDM_MODE_PHRASE:
			query->total_found=(Doc_phr[query->WWList.nuniq-1]>0)?j+1:j;
			break;
		case UDM_MODE_ALL:
			query->total_found=(bit_count(count)>=query->WWList.nuniq)?j+1:j;
			break;
		case UDM_MODE_ANY:
		default:
			query->total_found=j+1;
			break;
	}
	wrd[j].coord=UdmCalcPhraseWeight(wrd[j].url_id,Doc_weight,count,Doc_phr,query->WWList.nuniq,query->search_mode);
	free(Doc_phr);
	return;
}

typedef struct{
	//char	phr_len;
	uint4	count;
	uint4	weight;
} UDM_PHR_PAR;

void UdmGroupByURLBool(UDM_AGENT *query,UDM_URL_CRD *wrd,UDM_WRD_PHR *wrd_par){//,UDM_PHR_PAR *phr_par){
	size_t	i,j=0;
	size_t	Doc_weight=UDM_WRDSEC(wrd[0].coord);
/* BOOL stuff */
	size_t	item;
	UDM_STACK_ITEM temp_items[UDM_MAXSTACK];
/* PHRASE stuff */
	uint4	phr_num=wrd_par[0].phr_num;
	uint4	phr_mask=wrd_par[0].mask;
	uint4	phr_beg=UDM_WRDPOS(wrd[0].coord);
	uint4	phr_weight=UDM_WRDSEC(wrd[0].coord);
	UDM_PHR_PAR *phr_par=malloc(0000);

	bzero(phr_par,0000);
	memcpy(temp_items,query->items,query->nitems*sizeof(UDM_STACK_ITEM));

//printf("%d: %d&nbsp;",wrd[0].url_id,wrd[0].pos);
	for(i=1;i<query->total_found;i++){
		/* Group by url_id */
		if(wrd[j].url_id==wrd[i].url_id){
		/* The same document */
			Doc_weight+=UDM_WRDSEC(wrd[i].coord);

			/* Calculate phrase (subphrase) summary weight */
			if(((UDM_WRDPOS(wrd[i].coord)-phr_beg)<=((bit_count(phr_mask)+1)))
				&&(wrd_par[i].phr_num==phr_num)){
				phr_mask|=wrd_par[i].mask;
				phr_weight+=UDM_WRDSEC(wrd[i].coord);
			}else{
				if(wrd_par[i-1].phr_len==bit_count(phr_mask)){
					phr_par[(int)wrd_par[i-1].phr_num].count++;
					phr_par[(int)wrd_par[i-1].phr_num].weight+=phr_weight;
				}
				phr_num=wrd_par[i].phr_num;
				phr_mask=wrd_par[i].mask;
				phr_beg=UDM_WRDPOS(wrd[i].coord);
				phr_weight=UDM_WRDSEC(wrd[i].coord);
			}
		}else{
//printf("<BR>%d: ",wrd[i].url_id);
			/* Next document */

			/* Create a copy of lexem array       */
			/* and change word masks by 1 or 0    */
			/* depending on whether word presents */
			/* in the search query being executed */
			for(item=0;item<query->nitems;item++){
				if(temp_items[item].cmd==UDM_STACK_WORD){
					/* Change word mask by 1 or 0 */
					temp_items[item].arg=(phr_par[item].count)?1:0;
				}
			}
			if(UdmCalcBoolItems(temp_items,query->nitems)){
				wrd[j].coord=Doc_weight;
				j++;
			}else{
				/* Skip this result */
			}
			Doc_weight=UDM_WRDSEC(wrd[i].coord);
			phr_num=wrd_par[i].phr_num;
			phr_mask=wrd_par[i].mask;
			phr_beg=UDM_WRDPOS(wrd[i].coord);
			phr_weight=UDM_WRDSEC(wrd[i].coord);
			wrd[j]=wrd[i];
		}
//printf("%d&nbsp;",wrd[i].pos);
	}
//printf("<BR>");
	/* Check last word */
	for(item=0;item<query->nitems;item++){
		if(temp_items[item].cmd==UDM_STACK_WORD){
			/* Change word mask by 1 or 0 */
			temp_items[item].arg=(phr_par[item].count)?1:0;
		}
	}
	query->total_found=(UdmCalcBoolItems(temp_items,query->nitems))?j+1:j;

	wrd[j].coord=Doc_weight;
	return;
}

static void AddLimits(UDM_AGENT * Agent,UDM_VARLIST * vars){
	int i;
	
	for(i=0;i<vars->nvars;i++){
		if(!strcmp(vars->var[i].name,"ul")){
			if((vars->var[i].val)&&(vars->var[i].val[0])){
				UDM_URL url;
				char * wild;
				
				wild=(char*)malloc(strlen(vars->var[i].val)+3);
				UdmParseURL(&url,vars->var[i].val);
				if(url.schema[0]){
					sprintf(wild,"%s%%",vars->var[i].val);
				}else{
					sprintf(wild,"%%%s%%",vars->var[i].val);
				}
				UdmAddURLLimit(Agent->Conf,wild);
				free(wild);
			}
		}
		if(!strcmp(vars->var[i].name,"tag")){
			if((vars->var[i].val)&&(vars->var[i].val[0]))
				UdmAddTagLimit(Agent->Conf,vars->var[i].val);
		}
		if(!strcmp(vars->var[i].name,"cat")){
			if((vars->var[i].val)&&(vars->var[i].val[0]))
				UdmAddCatLimit(Agent->Conf,vars->var[i].val);
		}
		if(!strcmp(vars->var[i].name,"lang")){
			if((vars->var[i].val)&&(vars->var[i].val[0]))
				UdmAddLangLimit(Agent->Conf,vars->var[i].val);
		}
	}
}


static int SearchMode(const char * mode){
	if(!mode)return(UDM_MODE_ALL);
	if(!strcmp(mode,"all"))return(UDM_MODE_ALL);
	if(!strcmp(mode,"any"))return(UDM_MODE_ANY);
	if(!strcmp(mode,"bool"))return(UDM_MODE_BOOL);
	if(!strcmp(mode,"phrase"))return(UDM_MODE_PHRASE);
	return(UDM_MODE_ALL);
}


static int MatchMode(const char * mode){
	if(!mode)return(UDM_MATCH_FULL);
	if(!strcmp(mode,"wrd"))return(UDM_MATCH_FULL);
	if(!strcmp(mode,"full"))return(UDM_MATCH_FULL);
	if(!strcmp(mode,"beg"))return(UDM_MATCH_BEGIN);
	if(!strcmp(mode,"end"))return(UDM_MATCH_END);
	if(!strcmp(mode,"sub"))return(UDM_MATCH_SUBSTR);
	return(UDM_MATCH_FULL);
}



int UdmParseQueryString(UDM_AGENT * Agent,UDM_VARLIST * vars,char * query_string){
	char * tok, *lt;
	const char * wf;
	char str[1024];
	
	tok=UdmGetToken(query_string,"&",&lt);
	while(tok){
		char empty[]="";
		char * val;
		const char * lim;
		
		if((val=strchr(tok,'='))){
			*val='\0';
			val++;
		}else{
			val=empty;
		}
		UdmUnescapeCGIQuery(str,val);
		UdmReplaceStrVar(vars,tok,str,UDM_VARSRC_QSTRING);
		sprintf(str,"Limit-%s",tok);
		if((lim=UdmFindStrVar(vars,str,NULL))){
			int ltype=0;
			char * type, * fname, * llt;
			strncpy(str,lim,sizeof(str)-1);
			
			if((type=UdmGetToken(str,":",&llt))){
				if(!strcasecmp(type,"nested"))ltype=0;
				if(!strcasecmp(type,"time"))ltype=1;
				if(!strcasecmp(type,"linear"))ltype=2;
				if(!strcasecmp(type,"host"))ltype=3;
				if((fname=UdmGetToken(NULL,":",&llt))){
					UdmAddSearchLimit(Agent,ltype,fname,val);
				}
			}
		}
		tok=UdmGetToken(NULL,"&",&lt);
	}
	AddLimits(Agent,vars);
	
	Agent->Conf->local_charset=strdup(UdmFindStrVar(Agent->Conf->vars,"LocalCharset","utf-8"));
	Agent->Conf->browser_charset=strdup(UdmFindStrVar(Agent->Conf->vars,"BrowserCharset",Agent->Conf->local_charset));
	Agent->search_mode=SearchMode(UdmFindStrVar(Agent->Conf->vars,"m",NULL));
	Agent->word_match=MatchMode(UdmFindStrVar(Agent->Conf->vars,"wm",NULL));

	/* Now set sections weight factors */
	if((wf=UdmFindStrVar(Agent->Conf->vars,"wf",NULL))){
		size_t len;
		
		len=strlen(wf);
		if((len>0)&&(len<256)){
			const char *sec;
			int sn;
			
			for(sn=0;sn<256;sn++){
				Agent->wf[sn]=0;
			}
			
			for(sec=wf+len-1;sec>=wf;sec--){
				Agent->wf[len-(sec-wf)]=UdmHex2Int(*sec);
			}
		}
	}
	return 0;
}


char * UdmHlConvert(UDM_AGENT * query,const char * src, UDM_CHARSET * lcs, UDM_CHARSET * bcs) {
	int * tok, * lt, ures;
	char * hpart, * htxt;
	size_t len;
	int * uni;

	if(!src)return NULL;
	
	len=strlen(src);
	hpart=(char*)malloc(len+1);
	htxt=(char*)malloc(len*10);
	htxt[0]='\0';

	/* Convert to unicode */
	uni=(int *)malloc((len+10)*sizeof(int));
	ures=lcs->to_uni(uni,src,len+1);

	/* Parse unicode string */
	tok=UdmUniGetSepToken(uni,&lt);
	while(tok){
		int found=0;
		size_t slen,flen;
		int euchar;
		size_t uw;

		flen=lt-tok;

		/* Convert token to BrowserCharset */
		euchar=tok[flen];
		tok[flen]=0;
		hpart[0]='\0';
		ures=bcs->from_uni(hpart,tok,len+1);
		tok[flen]=euchar;

		/* Check that it is word to be marked */
		for(uw=0;uw<query->WWList.nwords;uw++){
			slen=UdmUniLen(query->WWList.Words[uw].uword);
			if((flen==slen)&&(!UdmUniStrNCaseCmp(tok,query->WWList.Words[uw].uword,slen))){
				found=1;
				break;
			}
		}
		if(found)strcat(htxt,"\2");
		strcat(htxt,hpart);
		if(found)strcat(htxt,"\3");
		
		tok=UdmUniGetSepToken(NULL,&lt);
	}
	free(hpart);
	free(uni);
	/*
	fprintf(stderr,"otxt='%s'\n",src);
	fprintf(stderr,"htxt='%s'\n",htxt);
	*/
	return(htxt);
}
