#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <regex.h>
#include <string.h>
#include <ctype.h>
#ifdef   HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif


#include "udm_utils.h"
#include "udm_common.h"
#include "udm_unicode.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_xmalloc.h"
#include "udm_word.h"
#include "udm_synonym.h"
#include "udm_crc32.h"
#include "udm_vars.h"

#define MAXNORMLEN 56
#ifndef MSG_WAITALL
#define MSG_WAITALL 0
#endif



/* Unicode regex lite BEGIN */

static const int * UdmUniRegTok(const int * s,const int **  last){
	if(s == NULL && (s=*last) == NULL)
		return NULL;

	switch(*s){
		case 0:
			return(NULL);
			break;
		case '[':
			for(*last=s+1;(**last)&&(**last!=']');(*last)++);
			if(**last==']')(*last)++;
			break;
		case '$':
		case '^':
			*last=s+1;
			break;
		default:
			for(*last=s+1;(**last)&&(**last!=']')&&(**last!='[')&&(**last!='^')&&(**last!='$');(*last)++);
			break;
	}
	return s;
}

static int UdmUniRegComp(UDM_UNIREG_EXP *reg, const int *pattern){
	const int * tok, * lt;

	reg->ntokens=0;
	reg->Tokens=NULL;

	tok=UdmUniRegTok(pattern,&lt);
	while(tok){
		size_t len;
		reg->Tokens=(UDM_UNIREG_TOK*)realloc(reg->Tokens,sizeof(*reg->Tokens)*(reg->ntokens+1));
		len=lt-tok;
		reg->Tokens[reg->ntokens].str=(int*)malloc((len+1)*sizeof(int));
		memcpy(reg->Tokens[reg->ntokens].str,tok,len*sizeof(int));
		reg->Tokens[reg->ntokens].str[len]=0;
		
		reg->ntokens++;
		tok=UdmUniRegTok(NULL,&lt);
	}
	return 0;
}


static int UdmUniRegExec(const UDM_UNIREG_EXP *reg, const int *string){
	const int * start=string;
	int match=0;
	
	for(start=string;*start;start++){
		const int * tstart=start;
		size_t i;
		
		for(i=0;i<reg->ntokens;i++){
			const int * s;
			int inc=UDM_UNIREG_INC;
#ifdef DEBUG_UNIREG
			char sstr[1024];
			char rstr[1024];

			unitostr(sstr,tstart);
			unitostr(rstr,reg->Tokens[i].str);
			printf("tstart='%s'\ttok='%s'\n",sstr,rstr);
#endif
			switch(reg->Tokens[i].str[0]){
				case '^':
					if(string!=tstart){
						match=0;
					}else{	
						match=1;
					}
					break;
				case '[':
					match=0;
					for(s=reg->Tokens[i].str+1;*s;s++){
						if(*s==']'){
						}else
						if(*s=='^'){
							inc=UDM_UNIREG_EXC;
							match=1;
						}else{
							if((*tstart==*s)&&(inc==UDM_UNIREG_EXC)){
								match=0;
								break;
							}
							if((*tstart==*s)&&(inc==UDM_UNIREG_INC)){
								match=1;
								break;
							}
						}
					}
					tstart++;
					break;
				case '$':
					if(*tstart!=0){
						match=0;
					}else{
						match=1;
					}
					break;
				default:
					match=1;
					for(s=reg->Tokens[i].str;(*s)&&(*tstart);s++,tstart++){
						if(*s=='.'){
							/* Any char */
						}else
						if((*s)!=(*tstart)){
							match=0;
							break;
						}
					}
					if((*s)&&(!*tstart))match=0;
					break;
			}
#ifdef DEBUG_UNIREG
			printf("match=%d\n",match);
#endif
			if(!match)break;
		}
		if(match)break;
	}

	return match;

}


static void UdmUniRegFree(UDM_UNIREG_EXP *reg){
	size_t i;
	
	for(i=0;i<reg->ntokens;i++)
		if(reg->Tokens[i].str)
			free(reg->Tokens[i].str);
}


/* Unicode regex lite END */




static int cmpspell(const void *s1,const void *s2){
  int lc;
  lc = strcmp(((const UDM_SPELL*)s1)->lang,((const UDM_SPELL*)s2)->lang);
  if (lc == 0) {
    int u1[BUFSIZ], u2[BUFSIZ];
    UdmUniStrCpy(u1,((const UDM_SPELL*)s1)->word); *u1 &= 255;
    UdmUniStrCpy(u2,((const UDM_SPELL*)s2)->word); *u2 &= 255;
      return(UdmUniStrCmp(u1, u2));
  }
  return lc;
}

static int cmpaffix(const void *s1,const void *s2){
  int lc;
  if (((const UDM_AFFIX*)s1)->type < ((const UDM_AFFIX*)s2)->type) return -1;
  if (((const UDM_AFFIX*)s1)->type > ((const UDM_AFFIX*)s2)->type) return 1;
  lc = strcmp(((const UDM_AFFIX*)s1)->lang,((const UDM_AFFIX*)s2)->lang);
  if (lc == 0) {
    if ( (((const UDM_AFFIX*)s1)->replen == 0) && (((const UDM_AFFIX*)s2)->replen == 0) ) return 0;
    if (((const UDM_AFFIX*)s1)->replen == 0) return -1;
    if (((const UDM_AFFIX*)s2)->replen == 0) return 1;
    {
      int u1[BUFSIZ], u2[BUFSIZ];
      UdmUniStrCpy(u1,((const UDM_AFFIX*)s1)->repl); 
      UdmUniStrCpy(u2,((const UDM_AFFIX*)s2)->repl); 
      if (((const UDM_AFFIX*)s1)->type == 'p') {
	*u1 &= 255; *u2 &= 255;
	return(UdmUniStrCmp(u1, u2));
      } else {
	u1[((const UDM_AFFIX*)s1)->replen - 1] &= 255; u2[((const UDM_AFFIX*)s2)->replen -1] &= 255;
	return(UdmUniStrBCmp(u1, u2));
      }
    }
  }
  return lc;
}

int UdmAddSpell(UDM_ENV * Conf,const int * word,const char *flag,const char *lang){
	if(Conf->nspell>=Conf->mspell){
		if(Conf->mspell){
			Conf->mspell+=1024*20;
			Conf->Spell=(UDM_SPELL *)UdmXrealloc(Conf->Spell,Conf->mspell*sizeof(UDM_SPELL));
		}else{
			Conf->mspell=1024*20;
			Conf->Spell=(UDM_SPELL *)UdmXmalloc(Conf->mspell*sizeof(UDM_SPELL));
		}
	}
	Conf->Spell[Conf->nspell].word = UdmUniDup(word);
	strncpy(Conf->Spell[Conf->nspell].flag,flag,10);
	strncpy(Conf->Spell[Conf->nspell].lang,lang,32);
	Conf->Spell[Conf->nspell].lang[32] = Conf->Spell[Conf->nspell].flag[10] = 0;
	Conf->nspell++;
	return(0);
}


__INDLIB__ int UdmImportDictionary(UDM_ENV * Conf, const char *lang, const char *charset,
				   const char *filename, int skip_noflag, const char *first_letters){
	unsigned char str[BUFSIZ];	
	int ustr[BUFSIZ];	
	FILE *dict;
	UDM_CHARSET *dict_charset = NULL;
	size_t len;

	dict_charset = UdmGetCharSet(charset);
	if (dict_charset == NULL) return 1;

	if(!(dict=fopen(filename,"r")))return(1);
	while(fgets(str,sizeof(str),dict)){
		unsigned char *s;
		const unsigned char *flag;

	        flag = NULL;
		if((s=strchr(str,'/'))){
			*s=0;
			s++;flag=s;
			while(*s){
				if(((*s>='A')&&(*s<='Z'))||((*s>='a')&&(*s<='z')))s++;
				else{
					*s=0;
					break;
				}
			}
		}else{
			if(skip_noflag)	continue;
			flag="";
		}
		dict_charset->to_uni(ustr, str, len = strlen(str) + 1); UdmUniStrToLower(ustr);
		dict_charset->from_uni(str, ustr, len);

		/* Dont load words if first letter is not required */
		/* It allows to optimize loading at  search time   */
		if(*first_letters)
			if(!strchr(first_letters,str[0]))
				continue;
		s=str;
		while(*s){
			if(*s=='\r')*s=0;
			if(*s=='\n')*s=0;
			s++;
		}
		UdmAddSpell(Conf,ustr,flag,lang);
	}
	fclose(dict);
	return(0);
}




static UDM_SPELL * UdmFindWord(UDM_AGENT * Indexer, const int *word, int affixflag) {
  int l,c,r,resc,resl,resr, nlang = Indexer->spellang, li, li_from, li_to, i;

        if (nlang == -1) {
	  li_from = 0; li_to = Indexer->Conf->nLang;
	} else {
	  li_from = nlang; li_to = li_from + 1;
	}
	if (Indexer->Conf->nspell) {
	  i = (int)(*word) & 255;
	  for(li = li_from; li < li_to; li++) {
	    l = Indexer->Conf->SpellTree[li].Left[i];
	    r = Indexer->Conf->SpellTree[li].Right[i];
	    if (l == -1) continue;
	    while(l<=r){
		    c = (l + r) >> 1;
		    resc = UdmUniStrCmp(Indexer->Conf->Spell[c].word, word);
		    if( (resc == 0) && 
			((affixflag == 0) || (strchr(Indexer->Conf->Spell[c].flag, affixflag) != NULL)) ) {
		      return(&Indexer->Conf->Spell[c]);
		    }
		    resl = UdmUniStrCmp(Indexer->Conf->Spell[l].word, word);
		    if( (resl == 0) && 
			((affixflag == 0) || (strchr(Indexer->Conf->Spell[l].flag, affixflag) != NULL)) ) {
		      return(&Indexer->Conf->Spell[l]);
		    }
		    resr = UdmUniStrCmp(Indexer->Conf->Spell[r].word, word);
		    if( (resr == 0) && 
			((affixflag == 0) || (strchr(Indexer->Conf->Spell[r].flag, affixflag) != NULL)) ) {
		      return(&Indexer->Conf->Spell[r]);
		    }
		    if(resc < 0){
			    l = c + 1;
			    r--;
		    } else if(resc > 0){
			    r = c - 1;
			    l++;
		    } else {
		      l++;
		      r--;
		    }
	    }
	  }
	}
	return(NULL);
}


int UdmAddAffix(UDM_ENV * Conf,int flag, const char * lang, const int *mask,const int *find, const int *repl, int type) {
#define ERRSTRSIZE 100

	if(Conf->naffixes>=Conf->maffixes){
		if(Conf->maffixes){
			Conf->maffixes+=16;
			Conf->Affix = UdmXrealloc(Conf->Affix,Conf->maffixes*sizeof(UDM_AFFIX));
		}else{
			Conf->maffixes=16;
			Conf->Affix = UdmXmalloc(Conf->maffixes*sizeof(UDM_AFFIX));
		}
	}

	Conf->Affix[Conf->naffixes].compile = 1;
	Conf->Affix[Conf->naffixes].flag=flag;
	Conf->Affix[Conf->naffixes].type=type;
	strncpy(Conf->Affix[Conf->naffixes].lang, lang, 32);
	Conf->Affix[Conf->naffixes].lang[32] = 0;
	
	UdmUniStrNCpy(Conf->Affix[Conf->naffixes].mask, mask, 40);
	UdmUniStrNCpy(Conf->Affix[Conf->naffixes].find,find, 15);
	UdmUniStrNCpy(Conf->Affix[Conf->naffixes].repl,repl, 15);
	Conf->Affix[Conf->naffixes].replen = UdmUniLen(repl);
	Conf->Affix[Conf->naffixes].findlen = UdmUniLen(find);
	Conf->naffixes++;
	return(0);
}

static char * remove_spaces(char *dist,char *src){
char *d,*s;
	d=dist;
	s=src;
	while(*s){
		if((*s != ' ')&& (*s != '-') && (*s != '\t')){
			*d=*s;
			d++;
		}
		s++;
	}
	*d=0;
	return(dist);
}


__INDLIB__ int UdmImportAffixes(UDM_ENV * Conf,const char *lang, const char*charset, 
				const char *filename) {
unsigned char str[BUFSIZ];
unsigned char flag=0;
unsigned char mask[BUFSIZ]="";
unsigned char find[BUFSIZ]="";
unsigned char repl[BUFSIZ]="";
unsigned char *s;
int i;
int suffixes=0;
int prefixes=0;
int IspellUsePrefixes;
/*int imp=0;
int rej=0;*/
FILE *affix;
 UDM_CHARSET *affix_charset = NULL;
 int umask[2*BUFSIZ];
 int ufind[2*BUFSIZ];
 int urepl[2*BUFSIZ];
 size_t len;

	    if(!(affix=fopen(filename,"r")))
		    return(1);

	    affix_charset = UdmGetCharSet(charset);
	    if (affix_charset == NULL) return 1;

	    IspellUsePrefixes = UdmFindIntVar(Conf->vars, "IspellUsePrefixes", 1);

	    while(fgets(str,sizeof(str),affix)){
		    if(!UDM_STRNCASECMP(str,"suffixes")){
		    	    suffixes=1;
			    prefixes=0;
			    continue;
		    }
		    if(!UDM_STRNCASECMP(str,"prefixes")){
			    suffixes=0;
			    prefixes=1;
			    continue;
		    }
		    if(!UDM_STRNCASECMP(str,"flag ")){
			    s=str+5;
			    while(strchr("* ",*s))s++;
			    flag=*s;
			    continue;
		    }
		    if((!suffixes)&&(!prefixes))continue;
		    if((prefixes)&&(!IspellUsePrefixes)) continue;
		
		    if((s=strchr(str,'#')))*s=0;
		    if(!*str)continue;

		    strcpy(mask,"");
		    strcpy(find,"");
		    strcpy(repl,"");

		    i=sscanf(str,"%[^>\n]>%[^,\n],%[^\n]",mask,find,repl);

		    remove_spaces(str,repl);strcpy(repl,str);
		    remove_spaces(str,find);strcpy(find,str);
		    remove_spaces(str,mask);strcpy(mask,str);

		    switch(i){
		    case 3:break;
		    case 2:
			    if(*find != '\0'){
				    strcpy(repl,find);
				    strcpy(find,"");
			    }
			    break;
		    default:
			    continue;
		    }

		    affix_charset->to_uni(urepl, repl, len = strlen(repl) + 1); UdmUniStrToLower(urepl);
		    affix_charset->from_uni(repl, urepl, len);

		    affix_charset->to_uni(ufind, find, len = strlen(find) + 1); UdmUniStrToLower(ufind);
		    affix_charset->from_uni(find, ufind, len);

		    if (suffixes) {
		      sprintf(str, "%s$", mask);
		    } else {
		      sprintf(str, "^%s", mask);
		    }

		    affix_charset->to_uni(umask, str, len = strlen(str) + 1); UdmUniStrToLower(umask);
		    affix_charset->from_uni(mask, umask, len);

		    UdmAddAffix(Conf,(int)flag,lang,umask,ufind,urepl,suffixes?'s':'p');
	    }
	    fclose(affix);
	    
	    return(0);
}

__INDLIB__ void UdmSortDictionary(UDM_ENV * Conf){
  int  j, CurLet = -1, Let;size_t i;
  char *CurLang = NULL;

#ifdef HAVE_MERGESORT
        mergesort((void*)Conf->Spell,Conf->nspell,sizeof(UDM_SPELL),cmpspell);
#else
        qsort((void*)Conf->Spell,Conf->nspell,sizeof(UDM_SPELL),cmpspell);
#endif
	for(i = 0; i < Conf->nspell; i++) {
	  if (CurLang == NULL || strncmp(CurLang, Conf->Spell[i].lang, 2) != 0) {
	    CurLang = Conf->Spell[i].lang;
	    strncpy(Conf->SpellTree[Conf->nLang].lang, CurLang, 2);
	    Conf->SpellTree[Conf->nLang].lang[3] = 0;
	    for(j = 0; j < 256; j++)
	      Conf->SpellTree[Conf->nLang].Left[j] =
		Conf->SpellTree[Conf->nLang].Right[j] = -1;
	    if (Conf->nLang > 0) {
	      CurLet = -1;
	    }
	    Conf->nLang++;
	  }
	  Let = (int)(*(Conf->Spell[i].word)) & 255;
	  if (CurLet != Let) {
	    Conf->SpellTree[Conf->nLang-1].Left[Let] = i;
	    CurLet = Let;
	  }
	  Conf->SpellTree[Conf->nLang-1].Right[Let] = i;
	}
}

__INDLIB__ void UdmSortAffixes(UDM_ENV * Conf) {
  int  j, CurLetP = -1, CurLetS = -1, Let, cl = -1;
  char *CurLangP = NULL, *CurLangS = NULL;
  UDM_AFFIX *Affix; size_t i;
  
  if (Conf->naffixes > 1)
    qsort((void*)Conf->Affix,Conf->naffixes,sizeof(UDM_AFFIX),cmpaffix);
  for(i = 0; i < Conf->nLang; i++)
    for(j = 0; j < 256; j++) {
      Conf->PrefixTree[i].Left[j] = Conf->PrefixTree[i].Right[j] = -1;
      Conf->SuffixTree[i].Left[j] = Conf->SuffixTree[i].Right[j] = -1;
    }

  for(i = 0; i < Conf->naffixes; i++) {
    Affix = &(((UDM_AFFIX*)Conf->Affix)[i]);
    if(Affix->type == 'p') {
      if (CurLangP == NULL || strcmp(CurLangP, Affix->lang) != 0) {
	cl = -1;
	for (j = 0; j < Conf->nLang; j++) {
	  if (strncmp(Conf->SpellTree[j].lang, Affix->lang, 2) == 0) {
	    cl = j;
	    break;
	  }
	}
	CurLangP = Affix->lang;
	strcpy(Conf->PrefixTree[cl].lang, CurLangP);
	CurLetP = -1;
      }
      if (cl < 0) continue; /* we have affixes without spell for this lang */
      Let = (int)(*(Affix->repl)) & 255;
      if (CurLetP != Let) {
	Conf->PrefixTree[cl].Left[Let] = i;
	CurLetP = Let;
      }
      Conf->PrefixTree[cl].Right[Let] = i;
    } else {
      if (CurLangS == NULL || strcmp(CurLangS, Affix->lang) != 0) {
	cl = -1;
	for (j = 0; j < Conf->nLang; j++) {
	  if (strcmp(Conf->SpellTree[j].lang, Affix->lang) == 0) {
	    cl = j;
	    break;
	  }
	}
	CurLangS = Affix->lang;
	strcpy(Conf->SuffixTree[cl].lang, CurLangS);
	CurLetS = -1;
      }
      if (cl < 0) continue; /* we have affixes without spell for this lang */
      Let = (Affix->replen) ? (int)(Affix->repl[Affix->replen-1]) & 255 : 0;
      if (CurLetS != Let) {
	Conf->SuffixTree[cl].Left[Let] = i;
	CurLetS = Let;
      }
      Conf->SuffixTree[cl].Right[Let] = i;
    }
  }
}

static UDM_SPELL * CheckSuffix(const int *word, size_t len, UDM_AFFIX *Affix, int *res, UDM_AGENT *Indexer) {
  int newword[2*MAXNORMLEN] = {0};
  int err;
/*3.1  int curlang, curspellang;*/
  
  *res = UdmUniStrBNCmp(word, Affix->repl, Affix->replen);
  if (*res < 0) {
    return NULL;
  }
  if (*res > 0) {
    return NULL;
  }
  UdmUniStrCpy(newword, word);
  UdmUniStrCpy(newword+len-Affix->replen, Affix->find);

  if (Affix->compile) {
    err = UdmUniRegComp(&(Affix->reg), Affix->mask);
    if(err){
      UdmUniRegFree(&(Affix->reg));
      return(NULL);
    }
    Affix->compile = 0;
  }
  if((err=UdmUniRegExec(&(Affix->reg),newword))){
    UDM_SPELL * curspell;

    if((curspell=UdmFindWord(Indexer, newword, Affix->flag))){

/*3.1 FIXME: language statistics collection while normalizing       
      curlang = Indexer->curlang;
      curspellang = Indexer->spellang;
      UdmSelectLang(Indexer, curspell->lang);
      Indexer->lang[Indexer->curlang].count++;
      Indexer->curlang = curlang;
      Indexer->spellang = curspellang;
*/
      return curspell;
    }
  }
  return NULL;
}

#define MAX_NORM 512
static int CheckPrefix(const int *word, size_t len, UDM_AFFIX *Affix, UDM_AGENT *Indexer, int li, int pi,
		UDM_SPELL **forms, UDM_SPELL ***cur ) {
  int newword[2*MAXNORMLEN] = {0};
  int err, ls, rs, lres,rres, res;
  size_t newlen;
  UDM_AFFIX *CAffix = Indexer->Conf->Affix;
/*3.1  int curlang, curspellang;*/
  
  res = UdmUniStrNCaseCmp(word, Affix->repl, Affix->replen); /* UdmUniStrNCmp */
  if (res != 0) {
    return res;
  }
  UdmUniStrCpy(newword, Affix->find);
  UdmUniStrCat(newword, word+Affix->replen);

  if (Affix->compile) {
    err = UdmUniRegComp(&(Affix->reg),Affix->mask);
    if(err){
      /*regerror(err, &(Affix->reg), regerrstr, ERRSTRSIZE);*/
      UdmUniRegFree(&(Affix->reg));
      return (0);
    }
    Affix->compile = 0;
  }
  if((err=UdmUniRegExec(&(Affix->reg),newword))){
    UDM_SPELL * curspell;

    if((curspell=UdmFindWord(Indexer, newword, Affix->flag))){
/* FIXME: language stat collection while normalizing 
      curlang = Indexer->curlang;
      curspellang = Indexer->spellang;
      UdmSelectLang(Indexer, curspell->lang);
      Indexer->lang[Indexer->curlang].count++;
      Indexer->curlang = curlang;
      Indexer->spellang = curspellang;
*/
      if ((*cur - forms) < (MAX_NORM-1)) {
	**cur =  curspell;
	(*cur)++; **cur = NULL;
      }
    } 
    newlen = UdmUniLen(newword);
    ls = Indexer->Conf->SuffixTree[li].Left[pi];
    rs = Indexer->Conf->SuffixTree[li].Right[pi];
    while (ls >= 0 && ls <= rs) {
      if (  ((*cur - forms) < (MAX_NORM-1)) ) {
	**cur = CheckSuffix(newword, newlen, &CAffix[ls], &lres, Indexer);
	if (**cur) {
	  (*cur)++; **cur = NULL;
	}
      }
      if ( (rs > ls) && ((*cur - forms) < (MAX_NORM-1)) ) {
	**cur = CheckSuffix(newword, newlen, &CAffix[rs], &rres, Indexer);
	if (**cur) {
	  (*cur)++; **cur = NULL;
	}
      }
      ls++;
      rs--;
    }
  }
  return 0;
}


__INDLIB__ UDM_SPELL ** UdmNormalizeWord(UDM_AGENT * Indexer, UDM_WIDEWORD *wword){
size_t len;
UDM_SPELL **forms;
UDM_SPELL **cur;
UDM_AFFIX * Affix;
int ri, pi, ipi, lp, rp, cp, ls, rs, nlang = Indexer->spellang;
int li, li_from, li_to, lres, rres, cres = 0;
/*3.1 int curlang, curspellang;*/
  UDM_SPELL *spell;
  int *uword = wword->uword;
/*  char *word = wword->word;
  UDM_CHARSET *local_charset = NULL;

	    local_charset = UdmGetCharSet(Indexer->Conf->local_charset);
	    if (local_charset == NULL) return NULL;
*/
	len=UdmUniLen(uword);
	if (len < Indexer->Conf->min_word_len 
		|| len > MAXNORMLEN
		|| len > Indexer->Conf->max_word_len
		)
		return(NULL);

	forms=(UDM_SPELL **) UdmXmalloc(MAX_NORM*sizeof(UDM_SPELL **));
	cur=forms;*cur=NULL;

	ri = (int)(*uword) & 255;
	pi = (int)(uword[UdmUniLen(uword)-1]) & 255;
	if (nlang == -1) {
	  li_from = 0; li_to = Indexer->Conf->nLang;
	} else {
	  li_from  = nlang;
	  li_to = nlang + 1;
	}
	Affix=(UDM_AFFIX*)Indexer->Conf->Affix;

	/* Check that the word itself is normal form */
	if((spell = UdmFindWord(Indexer, uword, 0))){
		*cur = spell;
		cur++; *cur=NULL;
/* FIXME: lanuage stat collection while normalizing
		curlang = Indexer->curlang;
		curspellang = Indexer->spellang;
		UdmSelectLang(Indexer, spell->lang);
		Indexer->lang[Indexer->curlang].count++;
		Indexer->curlang = curlang;
		Indexer->spellang = curspellang;
*/
	}

	/* Find all other NORMAL forms of the 'word' */

	for (ipi = 0; ipi <= pi; ipi += pi) {

	  for (li = li_from; li < li_to; li++) {
	    /* check prefix */
	    lp = Indexer->Conf->PrefixTree[li].Left[ri];
	    rp = Indexer->Conf->PrefixTree[li].Right[ri];
	    while (lp >= 0 && lp <= rp) {
	      cp = (lp + rp) >> 1;
	      cres = 0;
	      if ((cur - forms) < (MAX_NORM-1)) {
		cres = CheckPrefix(uword, len, &Affix[cp], Indexer, li, ipi, forms, &cur);
	      }
	      if ((lp < cp) && ((cur - forms) < (MAX_NORM-1)) ) {
		lres = CheckPrefix(uword, len, &Affix[lp], Indexer, li, ipi, forms, &cur);
	      }
	      if ( (rp > cp) && ((cur - forms) < (MAX_NORM-1)) ) {
		rres = CheckPrefix(uword, len, &Affix[rp], Indexer, li, ipi, forms, &cur);
	      }
	      if (cres < 0) {
		rp = cp - 1;
		lp++;
	      } else if (cres > 0) {
		lp = cp + 1;
		rp--;
	      } else {
		lp++;
		rp--;
	      }
	    }

	    /* check suffix */
	    ls = Indexer->Conf->SuffixTree[li].Left[ipi];
	    rs = Indexer->Conf->SuffixTree[li].Right[ipi];
	    while (ls >= 0 && ls <= rs) {
	      if (  ((cur - forms) < (MAX_NORM-1)) ) {
		*cur = CheckSuffix(uword, len, &Affix[ls], &lres, Indexer);
		if (*cur) {
		  cur++; *cur = NULL;
		}
	      }
	      if ( (rs > ls) && ((cur - forms) < (MAX_NORM-1)) ) {
		*cur = CheckSuffix(uword, len, &Affix[rs], &rres, Indexer);
		if (*cur) {
		  cur++; *cur = NULL;
		}
	      }
	      ls++;
	      rs--;
	    } /* end while */
	  
	  } /* for li */
	} /* for ipi */

	if(cur==forms){
		free(forms);
		return(NULL);
	}

	return(forms);
}

int UdmSelectSpellLang(UDM_ENV *Conf, char *lang) {
  int i;
  for(i=0; i < Conf->nLang; i++) {
    if(strcmp(lang, Conf->SpellTree[i].lang) == 0) {
      return i;
    }
  }
  return -1;
}



void UdmSelectLang(UDM_AGENT * Indexer, char *lang) {
/*	size_t i;
	char *s;
*/
	/* Convert to lower case */
/* FIXME
	for(s=lang;*s;*s=tolower(*s),s++);
	Indexer->spellang = UdmSelectSpellLang(Indexer->Conf, lang);
	for (i = 0; i < Indexer->nlangs; i++) {
		if (!UDM_STRNCMP(Indexer->lang[i].lang, lang)) {
			Indexer->curlang = i;
			return;
		}
	}
	if (i < UDM_LANGPERDOC) {
		strncpy(Indexer->lang[i].lang, lang, 3);
		Indexer->lang[i].count = 0;
		Indexer->curlang = i;
		Indexer->nlangs = i + 1;
	}
*/
}

__INDLIB__ void UdmFreeIspell (UDM_ENV *Conf) {
  int i;
  UDM_AFFIX *Affix = (UDM_AFFIX *)Conf->Affix;

  for (i = 0; i < Conf->naffixes; i++) {
    if (Affix[i].compile == 0) {
      UdmUniRegFree(&(Affix[i].reg));
    }
  }
  UDM_FREE(Conf->Affix);
  UDM_FREE(Conf->Spell);
  Conf->Affix = NULL;
  Conf->Spell = NULL;
  Conf->nspell = Conf->naffixes = Conf->nLang = 0;
  return;
}

void UdmAllFormsWord (UDM_AGENT *Indexer, UDM_SPELL *word, UDM_WIDEWORDLIST *result, size_t order);

void UdmAllFormsWord (UDM_AGENT *Indexer, UDM_SPELL *word, UDM_WIDEWORDLIST *result, size_t order) {
  UDM_WIDEWORD w;
  size_t i;
  size_t naffixes = Indexer->Conf->naffixes;
  UDM_AFFIX *Affix = (UDM_AFFIX *)Indexer->Conf->Affix;
  int err;
  UDM_CHARSET *local_charset = NULL;

  local_charset = UdmGetCharSet(Indexer->Conf->local_charset);
  if (local_charset == NULL) return;
  w.word = NULL;
  w.uword = NULL;

    for (i = 0; i < naffixes; i++) {

      if ( (word->flag != NULL)
	  && (strchr(word->flag, Affix[i].flag) != NULL)
	   && (strcmp(word->lang, Affix[i].lang) == 0 )
	   ) {
	if (Affix[i].compile) {
	  err = UdmUniRegComp(&(Affix[i].reg), Affix[i].mask);
	  if(err){
	    UdmUniRegFree(&(Affix[i].reg));
	    return;
	  }
	  Affix[i].compile = 0;
	}
      
	if( ((err=UdmUniRegExec(&(Affix[i].reg),word->word)) ) 
	    && (
		(Affix[i].type == 'p') ? (UdmUniStrNCaseCmp(word->word, Affix[i].find, Affix[i].findlen) == 0) :
		(UdmUniStrBNCmp(word->word, Affix[i].find, Affix[i].findlen) == 0)
		)
	    ){
	  
	  w.len = UdmUniLen(word->word) - Affix[i].findlen + Affix[i].replen;
	  if ( ( (w.word = realloc(w.word, w.len + 1)) == NULL) ||
	       ( (w.uword = realloc(w.uword, (w.len + 1) * sizeof(int))) == NULL)) 
	    return;

	  bzero(w.uword, (w.len + 1) * sizeof(int));

	  if (Affix[i].type == 'p') {
	    UdmUniStrCpy(w.uword, Affix[i].repl);
	    UdmUniStrCat(w.uword, &(word->word[Affix[i].findlen]));
	  } else {
	    UdmUniStrNCpy(w.uword, word->word, UdmUniLen(word->word) - Affix[i].findlen);
	    UdmUniStrCat(w.uword, Affix[i].repl);
	  }
	  local_charset->from_uni(w.word, w.uword, w.len + 1);
	  w.crcword = UdmStrCRC32(w.word);
	  w.order = order;
	  w.count = 0;
	  UdmWideWordListAdd(result, &w);
	}

      }
    }
}


__INDLIB__ UDM_WIDEWORDLIST * UdmAllForms (UDM_AGENT *Indexer, UDM_WIDEWORD *wword) {
  UDM_SPELL **norm, **cur, *nword;
  UDM_WIDEWORDLIST *result, *syn;
  UDM_WIDEWORD w;
  size_t i;
  UDM_CHARSET *local_charset = NULL;

  local_charset = UdmGetCharSet(Indexer->Conf->local_charset);
  if (local_charset == NULL) return NULL;

  if ((result = UdmXmalloc(sizeof(UDM_WIDEWORDLIST))) == NULL) {
    return NULL;
  }
  w.word = NULL;
  w.uword = NULL;

  UdmWideWordListInit(result);
  cur = norm = UdmNormalizeWord(Indexer, wword);

  if (cur != NULL) {
    while (*cur != NULL) {

      w.len = UdmUniLen((*cur)->word);
      if ( ( (w.word = realloc(w.word, w.len + 1)) == NULL) ||
	   ( (w.uword = realloc(w.uword, (w.len + 1) * sizeof(int))) == NULL)) 
	return NULL;
      UdmUniStrCpy(w.uword,(*cur)->word); 
      local_charset->from_uni(w.word, w.uword, w.len + 1);
      w.crcword = UdmStrCRC32(w.word);
      w.order = wword->order;
      w.count = 0;
      UdmWideWordListAdd(result, &w);

      syn = UdmSynonymListFind(&(Indexer->Conf->SynList), &w);

      if (syn != NULL)
	for(i = 0; i < syn->nwords; i++) {
	  UdmWideWordListAdd(result, &(syn->Words[i]));
	}
    
      UdmAllFormsWord(Indexer, *cur, result, wword->order);
      if (syn != NULL)
	for(i = 0; i < syn->nwords; i++) {
	  nword = UdmFindWord(Indexer, syn->Words[i].uword, 0);
	  if (nword != NULL) 
	    UdmAllFormsWord(Indexer, nword, result, wword->order);
	}

      cur++;
    }
  } else {
    UdmWideWordListAdd(result, wword);
    syn = UdmSynonymListFind(&(Indexer->Conf->SynList), wword);

    if (syn != NULL)
      for(i = 0; i < syn->nwords; i++) {
	UdmWideWordListAdd(result, &(syn->Words[i]));
      }
    
    if (syn != NULL)
      for(i = 0; i < syn->nwords; i++) {
	nword = UdmFindWord(Indexer, syn->Words[i].uword, 0);
	if (nword != NULL) 
	  UdmAllFormsWord(Indexer, nword, result, wword->order);
      }
  }
  UDM_FREE(norm);

  return result;
}
