#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>


#include "udm_config.h"
#include "udmsearch.h"
#include "udm_store.h"
#include "udm_xmalloc.h"

#ifndef UDM_CONF_DIR
#define UDM_CONF_DIR "/usr/local/mnogosearch/etc/"
#endif

int main(int argc, char ** argv){
	char template_name[1024] = "";
	char query_string[1024] = "";
	char self[1024] = "";
	const char * env, * lcharset, *bcharset;
	UDM_ENV    * Env;
	UDM_AGENT  * Agent;
	UDM_CHARSET * lcs = NULL, *bcs = NULL;
	TMPL tmpl[T_MAX];
	int httpd = 0, res;
	char *Doc, *HDoc;
	size_t DocSize = 0;
	unsigned rec_id;
	UDM_HTMLTOK tag;
	const char *htok, *last;

	/* Output Content-type if under HTTPD	 */
	/* Some servers do not pass QUERY_STRING */
	/* if the query was empty, so check	 */
	/* REQUEST_METHOD too     to be safe     */
	
	if(getenv("QUERY_STRING")||getenv("REQUEST_METHOD")){
		httpd=1;
	}
	
	UdmInit();
	UdmTemplateInit(tmpl);
	Env = UdmAllocEnv();
	Env->vars=UdmAllocVarList();

	/* Detect self and template name */
	if((env=getenv("UDMSEARCH_DOCTEMPLATE")))
		strncpy(template_name,env,sizeof(template_name)-1);

	if((env=getenv("QUERY_STRING"))){
		strncpy(query_string,env,sizeof(query_string)-1);
		if((env=getenv("REDIRECT_STATUS"))){

			/* Check Apache internal redirect  */
			/* via   "AddHandler" and "Action" */

			strncpy(self,(env=getenv("REDIRECT_URL"))?env:"storedoc.cgi",sizeof(self)-1);
			if(!template_name[0]){
				strncpy(template_name,(env=getenv("PATH_TRANSLATED"))?env:"",sizeof(template_name)-1);
			}
		}else{
			/* CGI executed without Apache internal redirect */

			/* Detect $Self variable with OS independant SLASHES */
			strncpy(self,(env=getenv("SCRIPT_NAME"))?env:"storedoc.cgi",sizeof(self)-1);

			if(!template_name[0]){
				char *s,*e;
				
				/*This is with OS specific SLASHES */
				env=((env=getenv("SCRIPT_FILENAME"))?env:"storedoc.cgi");

				if(strcmp(UDM_CONF_DIR,".")){
					/* Take from the config directory */
					snprintf(template_name,sizeof(template_name)-1,"%s/%s", UDM_CONF_DIR,(s=strrchr(env,UDMSLASH))?(s+1):(self));
				}else{
					/* Take from the current directory */
					strncpy(template_name,env,sizeof(template_name)-1);
				}

				/* Find right slash if it presents */
				s=((s=strrchr(template_name,UDMSLASH))?s:template_name);

				/* Find .cgi substring */
				if((e=strstr(s,".cgi"))){
					/* Replace ".cgi" with ".htm" */
					e[1]='h';e[2]='t';e[3]='m';
				}else{
					strcpy(template_name,"storedoc.htm");
				}
			}
		}
	}else{
		/* Executed from command line     */
		/* or under server which does not */
		/* pass an empty QUERY_STRING var */
		if(argv[1])snprintf(query_string,sizeof(query_string)-1,"q=%s",argv[1]);
		if(!template_name[0])snprintf(template_name,sizeof(template_name),"%s/%s", UDM_CONF_DIR,"storedoc.htm");
	}

	bzero(tmpl,sizeof(tmpl));
	res = UdmTemplateLoad(Env, Env->vars, template_name, tmpl);
	if(res){
		if(httpd)printf("Content-Type: text/plain\r\n\r\n");
		printf("%s\n",Env->errstr);
		UdmTemplateFree(tmpl);
		UdmFreeEnv(Env);
		return(0);
	}
	
	Env->DBAddr=strdup(UdmFindStrVar(Env->vars,"DBAddr","mysql://foo:bar@localhost/mnogosearch/"));
	UdmEnvSetDBMode(Env,UdmFindStrVar(Env->vars,"DBMode","single"));

	Env->stored_addr = strdup(UdmFindStrVar(Env->vars, "StoredAddr", "localhost"));

	Agent=UdmAllocAgent(Env,0,UDM_OPEN_MODE_READ);
	{
		size_t i;
		for(i=0;i<Env->sdcl.nclients;i++){
			UdmSDCLientListAdd(&Agent->sdcl,Env->sdcl.Clients[i].addr);
		}
	}
	
	UdmAddStrVar(Env->vars,"QUERY_STRING",query_string,UDM_VARSRC_GLOBAL);
	UdmAddStrVar(Env->vars,"self",self,UDM_VARSRC_GLOBAL);
	
	UdmParseQueryString(Agent,Env->vars,query_string);
	
	lcharset = UdmFindStrVar(Env->vars, "CS", "");
	if (lcharset == NULL || (!strcmp(lcharset, ""))) {
	  lcharset=UdmFindStrVar(Env->vars,"LocalCharset","utf-8");
	}
	lcs=UdmGetCharSet(lcharset);
	bcharset=UdmFindStrVar(Env->vars,"BrowserCharset","utf-8");
	bcs=UdmGetCharSet(bcharset);

	if(httpd){
		if(!bcs){
			printf("Content-Type: text/plain\r\n\r\n");
			printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
			exit(0);
		} else if (!lcs) {
			printf("Content-Type: text/plain\r\n\r\n");
			printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
			exit(0);
		}else{
			printf("Content-type: text/html; charset=%s\r\n\r\n",bcharset);
		}
	}else{
		if(!bcs){
			printf("Unknown BrowserCharset '%s' in template '%s'\n",bcharset,template_name);
			exit(0);
		}
		if(!lcs){
			printf("Unknown LocalCharset '%s' in template '%s'\n",lcharset,template_name);
			exit(0);
		}
	}

	/* These parameters taken fro "variable section of template"*/
	if((env=UdmFindStrVar(Env->vars,"VarDir",NULL))){
		snprintf(Agent->Conf->vardir,sizeof(Agent->Conf->vardir)-1,"%s%s",env,UDMSLASHSTR);
	}

	/* Now start displaying template*/
	UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_TOP);

/* UnStore Doc, Highlight and Display */
	sscanf(UdmFindStrVar(Env->vars, "rec_id", "0"), "%u", &rec_id);
	UdmPrepare(Agent,UdmFindStrVar(Env->vars, "q", ""));

	Doc = UdmUnStoreDoc(Env, rec_id, &DocSize);
	HDoc = UdmXmalloc(UDM_MAXDOCSIZE);

	htok = UdmHTMLToken(Doc, &last, &tag);
	while(htok) {
	  char *r = last;
	  char b = *r; 
	  *r = '\0';
	  switch(tag.type) {
	  case UDM_HTML_COM:
	  case UDM_HTML_TAG:
	    sprintf(UDM_STREND(HDoc), "%s", htok);
	    break;
	  case UDM_HTML_TXT:
	    sprintf(UDM_STREND(HDoc), "%s", UdmHlConvert(Agent, htok, lcs, bcs));
	    break;
	  }
	  *r = b;
	  htok = UdmHTMLToken(NULL, &last, &tag);
	}

/*****************/

/*	HDoc = UdmHlConvert(Agent, Doc, lcs, bcs);*/

	UdmAddStrVar(Env->vars, "document", HDoc, UDM_VARSRC_GLOBAL);

	UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_RESULT);

	UdmTemplatePrint(Agent,stdout,NULL,Env->vars,tmpl,T_BOTTOM);

	UdmTemplateFree(tmpl);
	UdmFreeAgent(Agent);
	UdmFreeEnv(Env);
	UDM_FREE(HDoc);

	return(0);
}
