#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_config.h"
#include "udmsearch.h"


const TPART part[] = {
	{T_TOP,		"<!--top-->",		"<!--/top-->"},
	{T_BOTTOM,	"<!--bottom-->",	"<!--/bottom-->"},
	{T_RESTOP,	"<!--restop-->",	"<!--/restop-->"},
	{T_RESBOT,	"<!--resbot-->",	"<!--/resbot-->"},
	{T_RESULT,	"<!--res-->",		"<!--/res-->"},
	{T_NOTFOUND,	"<!--notfound-->",	"<!--/notfound-->"},
	{T_ERROR,	"<!--error-->",		"<!--/error-->"},
	{T_CLONE,	"<!--clone-->",		"<!--/clone-->"},
	{T_NAVLEFT,	"<!--navleft-->",	"<!--/navleft-->"},
	{T_NAVBAR1,	"<!--navbar1-->",	"<!--/navbar1-->"},
	{T_NAVBAR0,	"<!--navbar0-->",	"<!--/navbar0-->"},
	{T_NAVRIGHT,	"<!--navright-->",	"<!--/navright-->"},
	{T_NAVLEFT_NOP,	"<!--navleft_nop-->",	"<!--/navleft_nop-->"},
	{T_NAVRIGHT_NOP,"<!--navright_nop-->",	"<!--/navright_nop-->"},
	{T_NOQUERY,	"<!--noquery-->",	"<!--/-->"},
	{T_VARIABLES,	"<!--variables",	"/variables-->"},
	{0,}
};

/******************* Template functions ********************/

void UdmTemplateInit(TMPL * t){
	int i,j;
	
	for(i=0;i<T_MAX;i++){
		t[i].nformats=0;
		for(j=0;j<MAX_FORMAT;j++)
			t[i].template[j]=NULL;
	}
}

static size_t out_string(FILE * stream,char * dst,const char * src){
	if(src){
		if(stream)fputs(src,stream);
		if(dst){
			strcat(dst,src);
			return(strlen(src));
		}
	}
	return(0);
}

static char * HiLightDup(const char * src,const char * beg, const char * end){
	size_t len=1;
	size_t blen=strlen(beg);
	size_t elen=strlen(end);
	const char * s;
	char  * res, *d;
	
	for(s=src;*s;s++){
		switch(*s){
			case '\2':
				len+=blen;
				break;
			case '\3':
				len+=elen;
				break;
			default:
				len++;
		}
	}
	res=(char*)malloc(len);
	for(s=src,d=res;*s;s++){
		switch(*s){
			case '\2':
				strcpy(d,beg);
				d+=blen;
				break;
			case '\3':
				strcpy(d,end);
				d+=elen;
				break;
			default:
				*d=*s;
				d++;
		}
	}
	*d='\0';
	return(res);
}

static size_t PrintTextTemplate(FILE * stream,char * dst, UDM_VARLIST * vars, const char * templ){
	const char * s;
	size_t dlen=0;
	const char * HlBeg=UdmFindStrVar(vars,"HlBeg","");
	const char * HlEnd=UdmFindStrVar(vars,"HlEnd","");

	for(s=templ;*s;s++){
		int type=0;
		char *value=NULL, *eval=NULL, *eval1=NULL;
		char empty[]="";

		if(*s=='$'){
			const char * vbeg=NULL, * vend;
			
			if(!strncmp(s,"$(",2)){vbeg=s+2;type='(';}
			else	if(!strncmp(s,"$%(",3)){vbeg=s+3;type='%';}
			else	if(!strncmp(s,"$&(",3)){vbeg=s+3;type='&';}
			else	if(!strncmp(s,"$^(",3)){vbeg=s+3;type='^';}

			if((type)&&(vend=strchr(s,')'))){
				UDM_VAR * var;
				size_t len;
				char name[100]="";
				
				len=(vend-vbeg);
				if(len>=sizeof(name))len=sizeof(name);
				strncpy(name,vbeg,len);name[len]='\0';
				if((var=UdmFindVar(vars,name))){
					value=var->val;
					if(!value)value=empty;
				}else{	
					value=empty;
				}
				
				s=vend;
			}else{
				type=0;
			}
		}
		switch(type){
			case '(': 
				dlen+=out_string(stream,dst+dlen,value);
				break;
			case '&':
				eval=UdmHtmlSpecialChars(value);
				eval1=HiLightDup(eval,HlBeg,HlEnd);
				dlen+=out_string(stream,dst+dlen,eval1);
				free(eval);
				free(eval1);
				break;
			case '^':
				eval=HiLightDup(value,HlBeg,HlEnd);
				dlen+=out_string(stream,dst+dlen,eval);
				free(eval);
				break;
			case '%':
				eval=(char*)malloc(strlen(value)*3);
				UdmEscapeURL(eval,value);
				dlen+=out_string(stream,dst+dlen,eval);
				free(eval);break;
			default:	/* One character */
				if((stream)&&(*s))fputc(*s,stream);
				if(dst){
					dst[dlen]=*s;
					dlen++;
					dst[dlen]='\0';
				}
		}
		/*if(dst)fprintf(stderr,"dst='%s'\n",dst);*/
	}
	return dlen;
}

#define T_OPT	1
#define T_INP	2

static char * GetHtmlTok(const char * src,const char ** lt){
	char * res;
	size_t len;
	if((!src)&&!(src=*lt))return(NULL);
	if(*src=='<'){
		/* Find ">" and skip "<" */
		if((*lt=strchr(src,'>')))(*lt)++;
	}else{
		/* Find tag beginning */
		*lt=strchr(src,'<');
	}
	if(!(*lt)){
		/* Last token */
		res=strdup(src);
	}else{
		/* Token in the middle */
		len=(*lt)-src;
		res=(char*)malloc(len+2);
		strncpy(res,src,len);
		res[len]='\0';
	}
	return(res);
}

static void PrintHtmlTemplate(UDM_AGENT * Agent,FILE * stream,char * dst,UDM_VARLIST * vars, const char * template){
	const char *lt;
	char  *tok;
	size_t dlen=0;
	
	tok=GetHtmlTok(template,&lt);
	while(tok){
		int type=0;
		if(!(UDM_STRNCASECMP(tok,"<OPTION"))){
			type=T_OPT;	
		}else
		if(!(UDM_STRNCASECMP(tok,"<INPUT"))){
			type=T_INP;
		}else
		if(!UDM_STRNCASECMP(tok,"<!INCLUDE")){
			if(Agent){
				int size;
				UDM_TAG tag;
				UDM_URL url;
				UDM_DOCUMENT * Inc=UdmAllocDoc();

				if(!Agent->buf)Agent->buf=malloc(Agent->Conf->max_doc_size);

				UdmParseTag(&tag,tok);
				if(tag.content){
					UdmParseURL(&url,tag.content);
					Inc->request=malloc(strlen(tag.content)+512);
					sprintf(Inc->request,"GET %s%s HTTP/1.0\r\nHost: %s\r\n\r\n",url.path,url.filename,url.hostname);
					size=UdmGetURL(Agent,Inc,&url);
					if(size>0)Agent->buf[size]='\0';
					UdmParseHTTPResponse(Agent,Inc);
					if(Inc->content){
						if(stream){
							fprintf(stream,"%s",Inc->content);
						}else{
							/* FIXME */
						}
					}
				}
			}
		}else{
			dlen+=PrintTextTemplate(stream,dst+dlen,vars,tok);
		}
		if(type){
			char * opt;
			UDM_TAG  tag;
			UDM_VAR * var=NULL;
			char * vname=NULL;
			
			opt=(char*)malloc(strlen(tok)+200);
			strcpy(opt,tok);
			UdmParseTag(&tag,opt);
			if((vname=tag.selected)){
				vname=UdmTrim(vname,"$()");
				var=UdmFindVarWithValue(vars,vname,tag.value?tag.value:"");
			}
			if(type==T_OPT){
				sprintf(opt,"<OPTION VALUE=\"%s\"%s>",tag.value?tag.value:"",var?" SELECTED":"");
			}else{
				sprintf(opt,"<INPUT TYPE=\"%s\" NAME=\"%s\" VALUE=\"%s\"%s>",tag.type?tag.type:"",tag.name?tag.name:"",tag.value?tag.value:"",var?" CHECKED":"");
			}
			dlen+=PrintTextTemplate(stream,dst+dlen,vars,opt);
			UdmFreeTag(&tag);
			free(opt);
		}
		
		
		free(tok);
		tok=GetHtmlTok(NULL,&lt);
	}
}

void UdmTemplatePrint(UDM_AGENT * Agent,FILE * stream,char * dst,UDM_VARLIST * vars, TMPL * t, int where){ 	
	if(dst)*dst='\0';
	if(t[where].nformats){
		char * tmpl=NULL;
		int format;
		format=UdmFindIntVar(vars,"o",0);
		if(format>=t[where].nformats)format=t[where].nformats-1;
		if(format<0)format=0;
		tmpl=t[where].template[format];
		PrintHtmlTemplate(Agent,stream,dst,vars,tmpl);
	}
}


/* Load template  */
int UdmTemplateLoad(UDM_ENV * Env,UDM_VARLIST * vars,const char * tname, TMPL * tmpl){
	FILE * file;
	char str[1024];
	int where=0;

	if(!(file=fopen(tname,"r"))){
		snprintf(Env->errstr,sizeof(Env->errstr)-1,"Unable to open template '%s': %s",tname,strerror(errno));
		return(1);
	}

	while(fgets(str,sizeof(str),file)){
		int i;
		int done=0;
		
		for(i=0;part[i].type;i++){
			if(!UDM_STRNCASECMP(str,part[i].begin)){
				where=part[i].type;
				done=1;
				break;
			}
			if(!UDM_STRNCASECMP(str,part[i].end)){
				tmpl[where].nformats++;
				where=0;
				done=1;
				break;
			}
		}
		if(done)continue;
		if(where==T_VARIABLES){
			char *tok,*s,*lt;
			if(!UDM_STRNCMP(str,"-->"))continue;
			s=UdmTrim(str," \r\n");
			if(!*s)continue;
			if(*s=='#')continue;

			if((tok=UdmGetToken(str," \t\r\n",&lt))){
				char * arg=NULL;
					
				if(!strcasecmp(str,"Affix")){
					char aname[1024];
					char * args[5];
					size_t narg=0;
					
					while((tok)&&(narg<5)){
						args[narg++]=tok;
						tok=UdmGetToken(NULL," \t",&lt);
					}
					if(narg!=4){
						Env->errcode=1;
						sprintf(Env->errstr,"Bad Affix command");
						return(1);
					}
					if(args[3][0]=='/')strncpy(aname,args[3],sizeof(aname)-1);
					else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,args[3]);
					if(UdmImportAffixes(Env,args[1],args[2],aname)){
						Env->errcode=1;
						snprintf(Env->errstr,sizeof(Env->errstr)-1,"Can't load affix :%s",aname);
						return(1);
					}
				}else
				if(!strcasecmp(str,"Spell")){
					char aname[1024];
					char * args[5];
					size_t narg=0;

					while((tok)&&(narg<5)){
						args[narg++]=tok;
						tok=UdmGetToken(NULL," \t",&lt);
					}
					if(narg!=4){
						Env->errcode=1;
						sprintf(Env->errstr,"Bad Spell command");
						return(1);
					}
					if(args[3][0]=='/')strncpy(aname,args[3],sizeof(aname)-1);
					else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,args[3]);
					if(UdmImportDictionary(Env,args[1],args[2],aname,0,"")){
						Env->errcode=1;
						snprintf(Env->errstr,sizeof(Env->errstr)-1,"Can't load dictionary :%s",aname);
						return(1);
					}
				}else
				if(!strcasecmp(str, "IspellUsePrefixes")) {
				  char sel[8];
				  int val = 1;
				  sscanf(str+17, "%4s", sel);
				  if (!UDM_STRNCASECMP(sel, "no")) {
				    val = 0;
				  }
				  UdmAddIntVar(Env->vars, "IspellUsePrefixes", val, UDM_VARSRC_GLOBAL);
				}else
				if(!strcasecmp(str,"StopwordFile")){
					char aname[1024];
					
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg){
						if(arg[0]=='/')strncpy(aname,arg,sizeof(aname)-1);
						else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
						if(UdmFileLoadStopList(Env,aname)){
							Env->errcode=1;
							return(1);
						}
					}else{
						Env->errcode=1;
						sprintf(Env->errstr,"Bad StopwordFile command");
						return(1);
					}
				}else
				if(!strcasecmp(str,"Synonym")){
					char aname[1024];
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg){
						if(arg[0]=='/')strncpy(aname,arg,sizeof(aname)-1);
						else	snprintf(aname,sizeof(aname)-1,"%s/%s",UDM_CONF_DIR,arg);
						if(UdmSynonymListLoad(Env,aname)){
							Env->errcode=1;
							return(1);
						}
					}else{
						Env->errcode=1;
						sprintf(Env->errstr,"Bad Synonym command");
						return(1);
					}
				}else
				if(!strcasecmp(str,"SearchdAddr")){
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg)UdmSDCLientListAdd(&Env->sdcl,arg);
				}else
				if((str[0]=='R'||str[0]=='r')&&(str[1]>='0')&&(str[1]<='9')){
					float r;
					int ir;
					arg=UdmGetToken(NULL," =\t\r\n",&lt); 
					if(arg){
						r=atoi(arg);
						srand((unsigned)time(0));
						r=r*rand()/RAND_MAX; ir=r;
						UdmAddIntVar(vars,str,ir,UDM_VARSRC_GLOBAL);
					}
				}else
				if(!strcasecmp(str,"HlBeg")){
					UdmAddStrVar(vars,"HlBeg",lt,UDM_VARSRC_GLOBAL);
				}else
				if(!strcasecmp(str,"HlEnd")){
					UdmAddStrVar(vars,"HlEnd",lt,UDM_VARSRC_GLOBAL);
				}else
				if(!strcasecmp(str,"Limit")){
					char * sc, * nm;
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if((sc=strchr(arg,':'))){
						*sc='\0';sc++;
						nm=(char*)malloc(strlen(arg)+1+strlen("Limit-"));
						sprintf(nm,"Limit-%s",arg);
						UdmAddStrVar(vars,nm,sc,UDM_VARSRC_GLOBAL);
						free(nm);
					}
				}else
				if(!strcasecmp(str,"CrossWords")){
					Env->use_crossword=0;
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg){
						if(!strcasecmp(arg,"yes")){
							Env->use_crossword=1;
						}
					}
				}else
				if(!strcasecmp(str,"Alias")&&(arg)){
					char * arg1;
					
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					arg1=UdmGetToken(NULL," \t\r\n",&lt);
					if(arg1){
						UDM_ALIAS Alias;
						Alias.match.pattern=arg;
						Alias.replace=arg1;
						Alias.match.match_type=0;
						Alias.match.case_sense=0;
						UdmAliasListAdd(&Env->AliasList,&Alias,Env->errstr,sizeof(Env->errstr));
					}
				}else{
					arg=UdmGetToken(NULL," \t\r\n",&lt);
					UdmReplaceStrVar(vars,tok,arg,UDM_VARSRC_GLOBAL);
				}
			}
		}else{
			char * t;
			t=tmpl[where].template[tmpl[where].nformats];
			if(!t){
				t=strdup(str);
			}else{
				t=(char*)realloc(t,strlen(t)+strlen(str)+1);
				strcat(t,str);
			}
			tmpl[where].template[tmpl[where].nformats]=t;
		}
	}
	fclose(file);
	if(Env->nspell) {
		UdmSortDictionary(Env);
		UdmSortAffixes(Env);
	}
	UdmSynonymListSort(&Env->SynList);
	return 0;
}

void UdmTemplateFree(TMPL * tmpl){
	int t;
	for(t=0;t<T_MAX;t++){
		int f;
		for(f=0;f<tmpl[t].nformats;f++){
			UDM_FREE(tmpl[t].template[f]);
		}
	}
}
