#include "udm_config.h"

#include <stdio.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_url.h"


int UdmParseURL(UDM_URL *url,const char *s){
	char *schema,*anchor,*file,*query;
	char query_string[UDM_URLSIZE]="";

	if(strlen(s)>=UDM_URLSIZE)
		return(UDM_URL_LONG);
	*(url->schema)=0;
	*(url->specific)=0;
	*(url->hostinfo)=0;
	*(url->hostname)=0;
	*(url->anchor)=0;
	*(url->auth)=0;
	url->port=0;
	url->default_port=0;
	*(url->path)=0;
	*(url->filename)=0;

	/* Find possible schema end than   */	
	/* Check that it is really schema  */
	/* It must consist of alphas only  */
	/* We will take in account digits  */
	/* also for oracle8:// for example */
	/* We must check it because        */
	/* It might be anchor also         */
	/* For example:                    */
	/* "mod/index.html#a:1"            */

	if((schema=strchr(s,':'))){
		const char * ch;
		for(ch=s;ch<schema;ch++){
			if(!isalnum(*ch)){
				/* Bad character       */
				/* so it is not schema */
				schema=0;break;
			}
		}
	}

	if(schema){
		/* Have scheme - absolute path */
		*schema=0;
		strcpy(url->schema,s);
		strcpy(url->specific,schema+1);
		*schema=':';
		if(!strcasecmp(url->schema,"http"))url->default_port=80;
		else
		if(!strcasecmp(url->schema,"https"))url->default_port=443;
		else
		if(!strcasecmp(url->schema,"nntp"))url->default_port=119;
		else
		if(!strcasecmp(url->schema,"news"))url->default_port=119;
		else
		if(!strcasecmp(url->schema,"ftp"))url->default_port=21;

		if(!strncmp(url->specific,"//",2)){
			char * ss;
			char * hostname;
			/* Have hostinfo */
			if((ss=strchr(url->specific+2,'/'))){
				/* Have hostname with path */
				*ss=0;
				strcpy(url->hostinfo,url->specific+2);
				*ss='/';
				strcpy(url->path,ss);
			}else{
				/* Hostname without path */
				strcpy(url->hostinfo,url->specific+2);
				strcpy(url->path,"/");
			}
			if((hostname=strchr(url->hostinfo,'@'))){
				/* Username and password is given  */
				/* Store auth string user:password */
				*hostname=0;
				strcpy(url->auth,url->hostinfo);
				*hostname='@';
				hostname++;
			}else{
				hostname=url->hostinfo;
			}
			if((ss=strchr(hostname,':'))){
				*ss=0;
				strcpy(url->hostname,hostname);
				*ss=':';
				url->port=atoi(ss+1);
			}else{
				strcpy(url->hostname,hostname);
				url->port=0;
			}
		}else{
			/* Have not host but have schema */
			/* This is possible for:         */
			/* file:  mailto:  htdb: news:   */
			
			/* FIXME This is actually durty hack for now       */
			/* to avoid changes in "stable"    indexer.c       */
			/* As far as we do not need mailto: just ignore it */
			
			if(!strcasecmp(url->schema,"mailto"))
				return(UDM_URL_BAD);
			else
			if(!strcasecmp(url->schema,"file"))
				
				strcpy(url->path,url->specific);
			else
			if(!strcasecmp(url->schema,"exec"))
				strcpy(url->path,url->specific);
			else
			if(!strcasecmp(url->schema,"cgi"))
				strcpy(url->path,url->specific);
			else
			if(!strcasecmp(url->schema,"htdb"))
				strcpy(url->path,url->specific);
			else
			if(!strcasecmp(url->schema,"news")){
				/* Now we will use localhost as NNTP    */
				/* server as it is not indicated in URL */
				strcpy(url->hostname,"localhost");
				sprintf(url->path,"/%s",url->specific);
				url->default_port=119;
			}else{
				/* Unknown strange schema */
				return(UDM_URL_BAD);
			}
		}
	}else{
		strcpy(url->path,s);
	}

	/* Cat an anchor if exist */
	if((anchor=strstr(url->path,"#")))*anchor=0;


	/* If path is not full just copy it to filename    */
	/* i.e. neither  /usr/local/ nor  c:/windows/temp/ */

	if((url->path[0]!='/')&&(url->path[1]!=':')){ 
		/* Relative path */
		if(!strncmp(url->path,"./",2))
			strcpy(url->filename,url->path+2);
		else
			strcpy(url->filename,url->path);
		url->path[0] = 0;
	}

	/* truncate path to query_string */
	/* and store query_string        */

	if((query=strchr(url->path,'?'))){
		strcpy(query_string,query);
		*(query) = 0;
	}

	UdmURLNormalizePath(url->path);

	/* Now find right '/' sign and copy the rest to filename */

	if((file=strrchr(url->path,'/'))&&(strcmp(file,"/"))){
		strcpy(url->filename,file+1);
		*(file+1)=0;
	}

	/* Restore query_string to filename*/
	strcat(url->filename,query_string);
	
	return(0);
}



char * UdmURLNormalizePath(char * str){
	char * s=str;
	char * q;
	char * d;

	/* Hide query string */

	if((q=strchr(s,'?'))){
		*q++='\0';
		if(!*q)q=NULL;
	}

	/* Remove all "/../" entries */

	while((d=strstr(str,"/../"))){
		char * p;
		
		if(d>str){
			/* Skip non-slashes */
			for(p=d-1;(*p!='/')&&(p>str);p--);
			
			/* Skip slashes */
			while((p>(str+1))&&(*(p-1)=='/'))p--;
		}else{
			/* We are at the top level and have ../  */
			/* Remove it too to avoid loops          */
			p=str;
		}
		memmove(p,d+3,strlen(d)-2);
	}

	/* Remove remove trailig "/.." */

	d=str+strlen(str);
	if((d-str>2)&&(!strcmp(d-3,"/.."))){
		d-=4;
		while((d>str)&&(*d!='/'))d--;
		if(*d=='/')d[1]='\0';
		else	strcpy(str,"/");
	}

	/* Remove all "/./" entries */
	
	while((d=strstr(str,"/./"))){
		memmove(d,d+2,strlen(d)-1);
	}

	/* Remove the trailing "/."  */
	
	if((d=str+strlen(str))>(str+2))
		if(!strcmp(d-2,"/."))
			*(d-1)='\0';

	/* Remove all "//" entries */
	while((d=strstr(str,"//"))){
		memmove(d,d+1,strlen(d));
	}

	
	/* Replace "%7E" with "~"         */
	/* Actually it is to be done      */
	/* for all valid characters       */
	/* which do not require escaping  */
	/* However I'm lazy, do it for 7E */
	/* as the most often "abused"     */

	while((d=strstr(str,"%7E"))){
		*d='~';
		memmove(d+1,d+3,strlen(d+3)+1);
	}

	/* Restore query string */

	if(q){
		char * e=str+strlen(str);
		*e='?';
		memmove(e+1,q,strlen(q)+1);
	}

	return str;
}

