#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_vars.h"



UDM_VAR * UdmFindVar(UDM_VARLIST * vars,const char * name){
	int i;
	
	for(i=0;i<vars->nvars;i++)
		if(!strcmp(name,vars->var[i].name))
			return(&vars->var[i]);
	return(NULL);
}

const char * UdmFindStrVar(UDM_VARLIST * vars,const char * name,const char * defval){
	UDM_VAR * var;
	if((var=UdmFindVar(vars,name)) != NULL)
		return((var->val != NULL)?var->val:defval);
	else
		return(defval);
}


UDM_VARLIST * UdmAllocVarList(void){
	UDM_VARLIST * vars;
	
	vars=(UDM_VARLIST*)malloc(sizeof(*vars));
	bzero(vars,sizeof(*vars));
	return(vars);
}

void UdmFreeVarList(UDM_VARLIST * vars){
	int i;
	for(i=0;i<vars->nvars;i++){
		if(vars->var[i].name)free(vars->var[i].name);
		if(vars->var[i].val)free(vars->var[i].val);
	}
	free(vars);
}

int UdmAddStrVar(UDM_VARLIST * vars,const char * name, const char * val, int src){
	vars->var=(UDM_VAR *)realloc(vars->var,(vars->nvars+1)*sizeof(UDM_VAR));
	vars->var[vars->nvars].scope=src;
	vars->var[vars->nvars].name=name?strdup(name):NULL;
	vars->var[vars->nvars].val=val?strdup(val):NULL;
	vars->nvars++;
	return vars->nvars;
}

int UdmAddIntVar(UDM_VARLIST * vars,const char * name, int val, int src){
	char num[64];
	sprintf(num,"%d",val);
	return UdmAddStrVar(vars,name,num,src);
}

int UdmReplaceStrVar(UDM_VARLIST * vars,const char * name,const char * val,int src){
	UDM_VAR * var;
	if((var=UdmFindVar(vars,name)) != NULL) {
		if(var->val != NULL) free(var->val);
		var->val=(val != NULL)?strdup(val):NULL;
		var->scope=src;
	}else{
		UdmAddStrVar(vars,name,val,src);
	}
	return vars->nvars;
}

int UdmReplaceIntVar(UDM_VARLIST * vars,const char * name,int val,int src){
	UDM_VAR * var;
	char num[64];
	sprintf(num,"%d",val);
	if((var=UdmFindVar(vars,name)) != NULL){
		var->val=strdup(num);
	}else{
		UdmAddIntVar(vars,name,val,src);
	}
	return vars->nvars;
}


UDM_VAR * UdmFindVarWithValue(UDM_VARLIST * vars,const char * name,const char * val){
	int i;
	
	for(i=0;i<vars->nvars;i++)
		if(!strcmp(name,vars->var[i].name)&&!strcmp(val,vars->var[i].val))
			return(&vars->var[i]);
	return(NULL);
}

int UdmFindIntVar(UDM_VARLIST * vars,const char * name,int defval){
	UDM_VAR * var;
	if((var=UdmFindVar(vars,name)) != NULL)
		return((var->val != NULL)?atoi(var->val):defval);
	else
		return(defval);
}

