#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>

#include "udm_config.h"
#include "udm_common.h"
#include "udm_db.h"
#include "udm_log.h"
#include "udm_spell.h"
#include "udm_xmalloc.h"
#include "udm_agent.h"
#include "udm_utils.h"
#include "udm_unicode.h"
#include "udm_mutex.h"
#include "udm_crc32.h"

#include "udm_cache.h"
#include "udm_worddb.h"


#include "udm_word.h"


#define WSIZE		1024
#define BSIZE		10

static int AddOneWord(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc,UDM_SERVER *Server,char *word,int weight){
	/* Realloc memory when required  */
	if(Doc->WordList.nwords>=Doc->WordList.mwords){
		if(Doc->WordList.mwords){
			Doc->WordList.mwords+=WSIZE;
			Doc->WordList.Word=(UDM_WORD *)UdmXrealloc(Doc->WordList.Word,Doc->WordList.mwords*sizeof(UDM_WORD));
		}else{
			Doc->WordList.mwords=WSIZE;
			Doc->WordList.Word=(UDM_WORD *)UdmXmalloc(Doc->WordList.mwords*sizeof(UDM_WORD));
		}
	}

	/* Add new word */
	Doc->WordList.Word[Doc->WordList.nwords].word=strdup(word);
	Doc->WordList.Word[Doc->WordList.nwords].coord=UDM_WRDCOORD(Doc->wordpos,weight);
	Doc->WordList.nwords++;

	if(Indexer->Conf->DBMode==UDM_DBMODE_BLOB){
		int crc=UdmStrCRC32(word);
		int count=0;
		int tabnum;
		int tabpos;

		tabnum=UDM_WRD_TABNUM(crc);
		tabpos=UDM_WRD_TABPOS(crc);

		UDM_DO_LOCK(Indexer->Conf,UDM_LOCK_WORDS);

		for(count=0;count<UDM_WRD_TABLEN;count++){
			if(!Indexer->Conf->WrdList[tabnum][tabpos].nwords)break;
			
			if(!strcmp(Indexer->Conf->WrdList[tabnum][tabpos].word,word)){
				break;
			}else{
				/*fprintf(stderr,"Collision: '%s' '%s'\n",Indexer->Conf->WrdList[crc].word,word);*/
			}
			tabpos++;
			tabpos=UDM_WRD_TABPOS(tabpos);
		}

		if(count>=(UDM_WRD_TABLEN*0.1)){
			fprintf(stderr,"Ooops. Table %0X is full count=%d\n",tabnum,count);
			UDM_DO_UNLOCK(Indexer->Conf,UDM_LOCK_WORDS);
			return IND_FLUSH;
		}

		if(Indexer->Conf->WrdList[tabnum][tabpos].nwords>=Indexer->Conf->WrdList[tabnum][tabpos].mwords){
			Indexer->Conf->WrdList[tabnum][tabpos].mwords+=BSIZE;
			Indexer->Conf->WrdList[tabnum][tabpos].Coords=(UDM_URL_CRD*)realloc(Indexer->Conf->WrdList[tabnum][tabpos].Coords,Indexer->Conf->WrdList[tabnum][tabpos].mwords*sizeof(UDM_URL_CRD));
		}

		if(!Indexer->Conf->WrdList[tabnum][tabpos].word)
			Indexer->Conf->WrdList[tabnum][tabpos].word=strdup(word);
		Indexer->Conf->WrdList[tabnum][tabpos].Coords[Indexer->Conf->WrdList[tabnum][tabpos].nwords].url_id=Doc->url_id;
		Indexer->Conf->WrdList[tabnum][tabpos].Coords[Indexer->Conf->WrdList[tabnum][tabpos].nwords].coord=UDM_WRDCOORD(Doc->wordpos,weight);
		Indexer->Conf->WrdList[tabnum][tabpos].nwords++;

		UDM_DO_UNLOCK(Indexer->Conf,UDM_LOCK_WORDS);
	}
	return(IND_OK);
}

/* This function adds a normalized word form(s) into list using Ispell */
int UdmAddWord(UDM_AGENT *Indexer,UDM_DOCUMENT * Doc,UDM_SERVER *Server,char *word,int weight){
	int	have_digit=0;
	int	have_alpha=0;
	int	res=IND_OK;

	Doc->wordpos++;

#ifdef DEBUG_ADDWORD
	fprintf(stderr,"AddWord '%s' pos=%d\n",word,Doc->wordpos);
#endif

	/* FIXME UNI: move ctype checking into UNICODE form */
	if(Server->number_factor==0||Server->alnum_factor==0){
		char *s;
		s=word;
		while(*s){
			if(isdigit(*s))
				have_digit=1;
			else	
				have_alpha=1;
			if(have_digit&&have_alpha)break;
			s++;
		}
		if(have_digit){
			if(have_alpha){
				if(!Server->alnum_factor)return(0);
			}else{
				if(!Server->number_factor)return(0);
			}
		}
	}

	res=AddOneWord(Indexer,Doc,Server,word,weight);

	switch(res){
		case IND_FLUSH:
			if(Indexer->Conf->DBMode==UDM_DBMODE_BLOB){
				Indexer->Conf->flashmode=UDM_FLASH_ALL;
				res=UdmStoreWordsBlob(Indexer,Doc);
				Indexer->Conf->flashmode=UDM_FLASH_BEST;
				res=AddOneWord(Indexer,Doc,Server,word,weight);
			}else{
				/* Is it possible? */
				res=IND_OK;
			}
		default:
			break;
	}
	return(res);
}

int UdmFreeWordList(UDM_WORDLIST * List){
	size_t i;
	for(i=0;i<List->nwords;i++)
		free(List->Word[i].word);
	List->nwords=0;
	List->swords=0;
	UDM_FREE(List->Word);
	return(0);
}

UDM_WORDLIST * UdmInitWordList(UDM_WORDLIST * List){
	bzero(List,sizeof(*List));
	return(List);
}


UDM_WIDEWORDLIST * UdmWideWordListInit(UDM_WIDEWORDLIST * List){
	bzero(List,sizeof(*List));
	return(List);
}

size_t UdmWideWordListAdd(UDM_WIDEWORDLIST * List,UDM_WIDEWORD * Word){
	/* Realloc memory */
	List->Words=(UDM_WIDEWORD*)realloc(List->Words,sizeof(*(List->Words))*(List->nwords+1));
	bzero(&List->Words[List->nwords],sizeof(*(List->Words)));
	
	/* Copy data */
	List->Words[List->nwords].len=Word->len;
	List->Words[List->nwords].order=Word->order;
	List->Words[List->nwords].count=Word->count;
	List->Words[List->nwords].crcword=Word->crcword;
	List->Words[List->nwords].word=Word->word?strdup(Word->word):NULL;
	List->Words[List->nwords].uword=Word->uword?UdmUniDup(Word->uword):NULL;
	List->nwords++;
	return(List->nwords);
}

void UdmWideWordListFree(UDM_WIDEWORDLIST * List){
	size_t i;
	for(i=0;i<List->nwords;i++){
		UDM_FREE(List->Words[i].word);
		UDM_FREE(List->Words[i].uword);
	}
	UDM_FREE(List->Words);
	UdmWideWordListInit(List);
}
