#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#if (WIN32|WINNT)
#include <time.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_log.h"
#include "udm_searchtool.h"
#include "udm_crc32.h"
#include "udm_cache.h"
#include "udm_agent.h"
#include "udm_worddb.h"

#define DEBUG_SEARCH


/* Word index item */
typedef struct {
	time_t  stamp;
	size_t  pos;
	size_t  cnt;
	char wrd[UDM_MAXDISCWORDSIZE];
} UDM_WIDX;


/* Word index header */
typedef struct {
	char id[4];
	size_t recsize;
	size_t nsorted;
	size_t nunsorted;
} UDM_WIDX_HDR;


/* Word data header */
typedef struct {
	char id[4];
	size_t ncoords;
} UDM_WDAT_HDR;

typedef struct {
	int url_id;
	time_t stamp;
} UDM_DELLOG;


/* File indetifiers */
#define UDM_WIDX_ID "WIDX"
#define UDM_WDAT_ID "WDAT"

static int cmpcoordurl(const UDM_URL_CRD * c1, const UDM_URL_CRD * c2){
	if(c1->url_id<c2->url_id)return -1;
	if(c1->url_id>c2->url_id)return 1;
	
	if(c1->coord<c2->coord)return -1;
	if(c1->coord>c2->coord)return 1;
	
	return 0;
}

int UdmBlobWriteStamp(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	char fname[1024];
	int f,res=IND_OK;
	const char * se;

	sprintf(fname,"%s/raw/del.d",Indexer->Conf->vardir);
	f=open(fname,O_CREAT|O_RDWR|UDM_BINARY,UDM_IWRITE);
	if(f<0){
		se=strerror(errno);
		sprintf(UdmAgentErrorMsg(Indexer),"Can't open dellog: errno=%d strerror='%s'",errno,se?se:"NULL");
		res=IND_ERROR;
	}else{
		off_t ofs;
		UDM_DELLOG dat;
		size_t nwritten;
		
		ofs=0;
		ofs=lseek(f,ofs,SEEK_END);
		dat.url_id=Doc->url_id;
		dat.stamp=time(NULL);
		nwritten=write(f,&dat,sizeof(dat));
		if(nwritten!=sizeof(dat)){
			se=strerror(errno);
			sprintf(UdmAgentErrorMsg(Indexer),"Can't write to dellog: errno=%d strerror='%s'",errno,se?se:"NULL");
			res=IND_ERROR;
		}
		close(f);
	}
	return res;
}


int UdmStoreWordsBlob(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	int max=10000;
	char fname[1024];
	UDM_WIDX idx;
	int tabnum;
	int tabpos;
	int widxsz=sizeof(UDM_WIDX)-UDM_MAXDISCWORDSIZE;
	time_t stamp;
	size_t lennum;
	int res=IND_OK;
	int fdd[UDM_WRD_MAXLENNO];
	int fid[UDM_WRD_MAXLENNO];

	bzero(fdd,sizeof(fdd));
	bzero(fid,sizeof(fid));

	if(Indexer->Conf->flashmode==UDM_FLASH_ALL)max=0;
	Indexer->Conf->ndoc++;

	stamp=time(NULL);
	for(tabnum=0;tabnum<=UDM_WRD_MASK;tabnum++){
		UDM_WIDX_HDR ihdr[UDM_WRD_MAXLENNO];
		UDM_WDAT_HDR dhdr[UDM_WRD_MAXLENNO];
		size_t wnum=0;
		size_t cmem=0;
		size_t nbytes=0;
		size_t nwritten=0;
		size_t nread;
		off_t ofs=0,ofs1;
		float percents;

		for(tabpos=0;tabpos<=UDM_WRD_TABLEN;tabpos++){
			size_t wrdsize,wrdlen;
			
			if(Indexer->Conf->WrdList[tabnum][tabpos].nwords>max){
				
				wrdlen=strlen(Indexer->Conf->WrdList[tabnum][tabpos].word);
				wrdsize=UDM_WRD_LENSIZE(wrdlen);
				lennum=UDM_WRD_LENNO(wrdlen);
				if(!fdd[lennum]){
					/* Open data file */
					sprintf(fname,"%s/raw/%02X-%02d.d",Indexer->Conf->vardir,tabnum,wrdsize);
					fdd[lennum]=open(fname,O_CREAT|O_RDWR|UDM_BINARY,UDM_IWRITE);
					UdmLog(Indexer,UDM_LOG_INFO,"Flushing %s",fname);
					if(fdd[lennum]<0){
						sprintf(UdmAgentErrorMsg(Indexer),"cant open '%s': %s\n",fname,strerror(errno));
						res=IND_ERROR;
						goto err;
					}
					
					ofs=lseek(fdd[lennum],(off_t)0,SEEK_END);
					if(ofs==0){
						/* File just created  */
						/* Let's write header */
						/* FIXME: check that index is empty */
						strncpy(dhdr[lennum].id,UDM_WDAT_ID,sizeof(dhdr[lennum].id));
						dhdr[lennum].ncoords=0;
						nwritten=write(fdd[lennum],&dhdr[lennum],sizeof(UDM_WDAT_HDR));
						if(nwritten!=sizeof(UDM_WDAT_HDR)){
							sprintf(UdmAgentErrorMsg(Indexer),"write index header error: %d %d\n",sizeof(UDM_WDAT_HDR),nwritten);
							res=IND_ERROR;
							goto err;
						}
					}else{
						/* File already exists */
						/* Read header        */
						/* FIXME: Check that index is not too big for small data file */
						nbytes=sizeof(UDM_WDAT_HDR);
						lseek(fdd[lennum],(off_t)0,SEEK_SET);
						nread=read(fdd[lennum],&dhdr[lennum],nbytes);
						if(nread!=nbytes){
							sprintf(UdmAgentErrorMsg(Indexer),"read data header error %d %d\n",nbytes,nread);
							res=IND_ERROR;
							goto err;
						}
						
						/* Move to the end of file */
						ofs=ofs1=dhdr[lennum].ncoords*sizeof(UDM_URL_CRD)+sizeof(UDM_WDAT_HDR);
						if(ofs1!=lseek(fdd[lennum],ofs,SEEK_SET)){
							sprintf(UdmAgentErrorMsg(Indexer),"lseek to the end of data file failed: %d %d\n",(int)ofs,(int)ofs1);
							res=IND_ERROR;
							goto err;
						}
					}

					/* Open index file */
					sprintf(fname,"%s/raw/%02X-%02d.i",Indexer->Conf->vardir,tabnum,wrdsize);
					fid[lennum]=open(fname,O_CREAT|O_RDWR|UDM_BINARY,UDM_IWRITE);
					if(fid[lennum]<0){
						sprintf(UdmAgentErrorMsg(Indexer),"can't open index '%s': %s\n",fname,strerror(errno));
						res=IND_ERROR;
						goto err;
					}
					
					ofs=lseek(fid[lennum],(off_t)0,SEEK_END);
					if(ofs==0){
						/* File just created */
						/* Write header      */
						strncpy(ihdr[lennum].id,UDM_WIDX_ID,sizeof(ihdr[lennum].id));
						ihdr[lennum].recsize=wrdsize+widxsz;
						ihdr[lennum].nsorted=0;
						ihdr[lennum].nunsorted=0;
						nwritten=write(fid[lennum],&ihdr[lennum],sizeof(UDM_WIDX_HDR));
						if(nwritten!=sizeof(UDM_WIDX_HDR)){
							sprintf(UdmAgentErrorMsg(Indexer),"can't write ihdr: %d %d\n",sizeof(UDM_WIDX_HDR),nwritten);
							res=IND_ERROR;
							goto err;
						}
					}else{
						/* File already exists */
						/* Load old header     */
						lseek(fid[lennum],(off_t)0,SEEK_SET);
						nbytes=sizeof(UDM_WIDX_HDR);
						nread=read(fid[lennum],&ihdr[lennum],nbytes);
						if(nread!=nbytes){
							sprintf(UdmAgentErrorMsg(Indexer),"read idx header error: %d %d\n",nbytes,nread);
							res=IND_ERROR;
							goto err;
						}
						
						/* Move to the end of data */
						ofs1=ofs=(ihdr[lennum].nsorted+ihdr[lennum].nunsorted)*ihdr[lennum].recsize+sizeof(UDM_WIDX_HDR);
						if(ofs1!=lseek(fid[lennum],ofs,SEEK_SET)){
							sprintf(UdmAgentErrorMsg(Indexer),"lseek idx error: %d %d\n",(int)ofs,(int)ofs1);
							res=IND_ERROR;
							goto err;
						}
					}
				}
				
				nbytes=Indexer->Conf->WrdList[tabnum][tabpos].nwords*sizeof(UDM_URL_CRD);
				qsort(Indexer->Conf->WrdList[tabnum][tabpos].Coords,Indexer->Conf->WrdList[tabnum][tabpos].nwords,sizeof(UDM_URL_CRD),(qsort_cmp)(cmpcoordurl));
				nwritten=write(fdd[lennum],Indexer->Conf->WrdList[tabnum][tabpos].Coords,nbytes);
				if(nwritten!=nbytes){
					sprintf(UdmAgentErrorMsg(Indexer),"can't write data: %d %d\n",nbytes,nwritten);
					res=IND_ERROR;
					goto err;
				}
				
				bzero(&idx,sizeof(idx));
				strncpy(idx.wrd,Indexer->Conf->WrdList[tabnum][tabpos].word,sizeof(idx.wrd));
				idx.pos=dhdr[lennum].ncoords;
				idx.cnt=Indexer->Conf->WrdList[tabnum][tabpos].nwords;
				idx.stamp=stamp;
				nbytes=ihdr[lennum].recsize;
				nwritten=write(fid[lennum],&idx,nbytes);
				if(nwritten!=nbytes){
					sprintf(UdmAgentErrorMsg(Indexer),"can't write idx: %d %d\n",nbytes,nwritten);
					res=IND_ERROR;
					goto err;
				}
				dhdr[lennum].ncoords+=Indexer->Conf->WrdList[tabnum][tabpos].nwords;
				ihdr[lennum].nunsorted++;

				/* Clear hash table item */
				Indexer->Conf->WrdList[tabnum][tabpos].nwords=0;
				Indexer->Conf->WrdList[tabnum][tabpos].mwords=0;
				Indexer->Conf->WrdList[tabnum][tabpos].hasfile=1;
				UDM_FREE(Indexer->Conf->WrdList[tabnum][tabpos].Coords);
				if(Indexer->Conf->flashmode==UDM_FLASH_ALL)
					UDM_FREE(Indexer->Conf->WrdList[tabnum][tabpos].word);
			}

			cmem+=Indexer->Conf->WrdList[tabnum][tabpos].nwords*sizeof(UDM_URL_CRD);
			if(Indexer->Conf->WrdList[tabnum][tabpos].word){
				wnum++;
			}
		}
		
		/* Now close files and store headers */
		for(lennum=0;lennum<UDM_WRD_MAXLENNO;lennum++){
			
			if(fdd[lennum]){
				/* Store DAT header  */
				ofs=lseek(fdd[lennum],(off_t)0,SEEK_SET);
				nbytes=sizeof(UDM_WDAT_HDR);
				nwritten=write(fdd[lennum],&dhdr[lennum],nbytes);
				if(nwritten!=sizeof(UDM_WDAT_HDR)){
					sprintf(UdmAgentErrorMsg(Indexer),"can't overwrite data header: %d %d (%s)\n",nbytes,nwritten,strerror(errno));
					res=IND_ERROR;
					goto err;
				}
				close(fdd[lennum]);
				fdd[lennum]=0;
			}
			
			if(fid[lennum]){
				/* Now store IDX header */
				ofs=lseek(fid[lennum],(off_t)0,SEEK_SET);
				nbytes=sizeof(UDM_WIDX_HDR);
				nwritten=write(fid[lennum],&ihdr[lennum],nbytes);
				if(nwritten!=nbytes){
					sprintf(UdmAgentErrorMsg(Indexer),"can't overwrite index header: %d %d\n",nbytes,nwritten);
					res=IND_ERROR;
					goto err;
				}
				close(fid[lennum]);
				fid[lennum]=0;
			}
		}
		percents=wnum;
		percents=percents*100/(UDM_WRD_TABLEN+1);
		/*fprintf(stderr,"tab=%0X docs: %d  %dK (%.2f%%) coords %d %dK\n",tabnum,Indexer->Conf->ndoc,wmem/1024,percents,cmem/sizeof(UDM_URL_CRD),cmem/1024);*/
	}
err:
	for(lennum=0;lennum<UDM_WRD_MAXLENNO;lennum++){
		if(fid[lennum]>0)close(fid[lennum]);
		if(fdd[lennum]>0)close(fdd[lennum]);
	}
	if(res!=IND_OK){
		UdmLog(Indexer,UDM_LOG_ERROR,"BLOB error: '%s'",UdmAgentErrorMsg(Indexer));
		exit(0);
	}
	return(res);
}


/*
static int cmpdel(const void * w1, const void * w2){
	int n1=((const UDM_LOGDEL *)(w1))->url_id;
	int n2=((const UDM_LOGDEL *)(w2))->url_id;

	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return 0;
}

static time_t UdmURLAge(UDM_LOGDEL *buf, size_t buf_count, const int url_id){
	register int m;
	register int l;
	register int r;

	l=0;
	r=buf_count;
	while(l<r){
		m=(l+r)/2;
		if(buf[m].url_id<url_id) l=m+1;
		else r=m;
	}
	if(r==buf_count) return(0);
	else
		if(buf[r].url_id==url_id) return(buf[r].stamp);
		else return(0);
}
*/

/* Function to sort DELLOG list in URL_ID order */
static int cmpurldellog(const void *s1,const void *s2){
	unsigned int n1,n2;
	
	n1=((const UDM_DELLOG*)s1)->url_id;
	n2=((const UDM_DELLOG*)s2)->url_id;
	if(n1==n2){
		n1=((const UDM_DELLOG*)s2)->stamp;
		n2=((const UDM_DELLOG*)s1)->stamp;
	}
	if(n1<n2)return(-1);
	if(n1>n2)return(1);
	return(0);
}


static time_t PresentInDelLog(UDM_DELLOG *buf, size_t buf_count, size_t *start, const int url_id){
	size_t m;
	size_t l;
	size_t r;

	if(start)l=*start;
	else l=0;
	r=buf_count;
	while(l<r){
		m=(l+r)/2;
		if(buf[m].url_id<url_id) l=m+1;
		else r=m;
	}
	if(start)*start=r;
	if(r==buf_count) return(0);
	else
		if(buf[r].url_id==url_id) return(buf[r].stamp);
		else return(0);
}

/* 
This function removes duplicate
(wrd_id,url_id) pairs using timestamps.
Most new pairs will remane.
*/

static int RemoveURLDelDups(UDM_DELLOG *words, size_t n){
	int i,j;
	
	j=0;
	for(i=1;i<n;i++){
		if(words[j].url_id!=words[i].url_id){
			j++;
			if(i!=j)words[j]=words[i];
		}
	}
	return(j+1);
}

static int LoadDelLog(UDM_AGENT * Agent,UDM_DELLOG ** del,size_t * ndel){
	char fname[1000];
	int ddd;
	off_t ofs;
	size_t nbytes,nread,nurls;
	
	snprintf(fname,sizeof(fname),"%s/raw/del.d",Agent->Conf->vardir);
	fname[sizeof(fname)-1]='\0';

#ifdef DEBUG_SEARCH
	fprintf(stderr,"Open dellog '%s'\n",fname);
#endif
	ddd=open(fname,O_RDONLY);
	if(ddd<0){
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Cant open '%s' for reading\n",fname);
#endif
		return(1);
	}

	ofs=lseek(ddd,(off_t)0,SEEK_END);
	nbytes=ofs;
	nurls=nbytes/sizeof(UDM_DELLOG);
	*del=(UDM_DELLOG*)malloc(nbytes);
	lseek(ddd,(off_t)0,SEEK_SET);
	nread=read(ddd,*del,nbytes);

	if(nread!=nbytes){
#ifdef DEBUG_SEARCH
		fprintf(stderr,"can't read dellog: %d %d\n",nbytes,nread);
#endif
	}

	*ndel=nread/sizeof(UDM_DELLOG);
	close(ddd);

#ifdef DEBUG_SEARCH
	fprintf(stderr,"ndel=%d\n",*ndel);
	fprintf(stderr,"remove dups\n");
#endif

	qsort(*del,*ndel,sizeof(UDM_DELLOG),cmpurldellog);
	*ndel=RemoveURLDelDups(*del, *ndel);

#ifdef DEBUG_SEARCH
	fprintf(stderr,"ndel=%d\n",*ndel);
#endif
	return 0;
}


typedef struct m {
	int last;
	int cur;
	int num;
} UDM_MERG;

typedef struct pm {
	UDM_URL_CRD * plast;
	UDM_URL_CRD * pcur;
	int num;
} UDM_PMERG;

#define MAXMERGE 1024	/* FIXME!!! */



typedef struct {
	size_t stamp;
	size_t pos;
	size_t cnt;
} UDM_LIDX;

static int cmplidx(const UDM_LIDX * c1, const UDM_LIDX * c2){
	if(c1->pos<c2->pos)return -1;
	if(c1->pos>c2->pos)return 1;
	return 0;
}


typedef struct {
	UDM_DELLOG * del;
	size_t ndel;
	UDM_URL_CRD * coords;
	size_t ncoords;
	UDM_LIDX * lidx;
	size_t nidx;
	size_t midx;
	UDM_MERG merg[MAXMERGE];
	size_t nmerg;
} UDM_WPARAM;


static int LoadCoords(UDM_AGENT * Agent,UDM_WPARAM * prm, const char * word,int wordnum){
	int fid=0,fdd=0;
	UDM_WIDX_HDR ihdr;
	UDM_WDAT_HDR dhdr;
	UDM_WIDX lw;
	UDM_WIDX rw;
	UDM_WIDX mw;
	size_t nibuf=1024*16;
	size_t ribuf=0;
	char * ibuf;
	off_t l,m,r,ofs,ofs1;
	size_t nread,i,nbytes;
	char fname[1000];
	int crc=UdmStrCRC32(word);
	size_t wordlen=strlen(word);
	size_t mcoords=0;
	size_t sncoords=prm->ncoords;
	size_t snidx=prm->nidx;

	sprintf(fname,"%s/raw/%02X-%02d.i",Agent->Conf->vardir,UDM_WRD_TABNUM(crc),UDM_WRD_LENSIZE(wordlen));
#ifdef DEBUG_SEARCH
	fprintf(stderr,"Open '%s'\n",fname);
#endif

	fid=open(fname,O_RDONLY);
	if(fid<0){
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Cant open '%s' for reading\n",fname);
#endif
		goto err;
	}

	/* Load IDX header */
	nread=read(fid,&ihdr,sizeof(ihdr));
	if(nread!=sizeof(ihdr)){
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Can't read: %d %d\n",sizeof(ihdr),nread);
#endif
		goto err;
	}

	wordlen=ihdr.recsize - (sizeof(UDM_WIDX)-64);
#ifdef DEBUG_SEARCH
	fprintf(stderr,"recsize=%d wordlen=%d\n",ihdr.recsize,wordlen);
#endif

	sprintf(fname,"%s/raw/%02X-%02d.d",Agent->Conf->vardir,UDM_WRD_TABNUM(crc),UDM_WRD_LENSIZE(wordlen));
#ifdef DEBUG_SEARCH
	fprintf(stderr,"Open data file '%s'\n",fname);
#endif
	fdd=open(fname,O_RDONLY);
	if(fdd<0){
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Cant open '%s' for reading\n",fname);
#endif
		goto err;
	}


	/* Load DAT header */
	nbytes=sizeof(dhdr);
	nread=read(fdd,&dhdr,nbytes);
	if(nread!=nbytes){
#ifdef DEBUG_SEARCH
		fprintf(stderr,"Can't read: %d %d\n",nbytes,nread);
#endif
		goto err;
	}
#ifdef DEBUG_SEARCH
	fprintf(stderr,"data ncoords=%d\n",dhdr.ncoords);
#endif

	l=0;
	r=ihdr.nsorted;
	lw.wrd[0]='\0';
	rw.wrd[0]='\0';

	while(l<r){
		int res;

		m=(r+l)/2;

		ofs=m*ihdr.recsize+sizeof(ihdr);
		ofs1=lseek(fid,ofs,SEEK_SET);
		if(ofs!=ofs1){
#ifdef DEBUG_SEARCH
			fprintf(stderr,"sorted part: can't lseek middle: %d %d\n",(int)ofs,(int)ofs1);
#endif
			goto err;
		}
		nbytes=ihdr.recsize;
		nread=read(fid,&mw,nbytes);
		if(nread!=nbytes){
#ifdef DEBUG_SEARCH
			fprintf(stderr,"sorted part: can't read middle: %d %d",nbytes,nread);
#endif
			exit(1);
		}
		mw.wrd[wordlen]='\0';
		
		res=strcmp(word,mw.wrd);
		if(res>0){
			l=m+1;
			ofs=l*ihdr.recsize+sizeof(ihdr);
			ofs1=lseek(fid,ofs,SEEK_SET);
			if(ofs!=ofs1){
#ifdef DEBUG_SEARCH
				fprintf(stderr,"sorted part: can't lseek left: %d %d\n",(int)ofs,(int)ofs1);
#endif
				goto err;
			}
			nbytes=ihdr.recsize;
			nread=read(fid,&lw,nbytes);
			if(nread!=nbytes){
#ifdef DEBUG_SEARCH
				fprintf(stderr,"sorted part: can't read left: %d %d",nbytes,nread);
#endif
				goto err;
			}
			lw.wrd[wordlen]='\0';
		}else{
			r=m;
			ofs=r*ihdr.recsize+sizeof(ihdr);
			ofs1=lseek(fid,ofs,SEEK_SET);
			if(ofs!=ofs1){
#ifdef DEBUG_SEARCH
				fprintf(stderr,"sorted part: can't lseek right: %d %d\n",(int)ofs,(int)ofs1);
#endif
				goto err;
			}
			nbytes=ihdr.recsize;
			nread=read(fid,&rw,nbytes);
			if(nread!=nbytes){
#ifdef DEBUG_SEARCH
				fprintf(stderr,"sorted part: can't read left: %d %d",nbytes,nread);
#endif
				goto err;
			}
			rw.wrd[wordlen]='\0';
		}
	}
	

	if(r!=ihdr.nsorted){ 

		/* Found in sorted part */

		ofs=r*ihdr.recsize+sizeof(ihdr);
		ofs1=lseek(fid,ofs,SEEK_SET);
		if(ofs!=ofs1){
#ifdef DEBUG_SEARCH		
			fprintf(stderr,"sorted part: can't lseek first: %d %d\n",(int)ofs,(int)ofs1);
#endif			
			goto err;
		}
		while((nread=read(fid,&mw,ihdr.recsize))){
			if(nread!=ihdr.recsize){
#ifdef DEBUG_SEARCH			
				fprintf(stderr,"unsorted part: can't read mw: %d %d\n",ihdr.recsize,nread);
#endif				
				goto err;
			}
			mw.wrd[wordlen]='\0';
			if(strcmp(mw.wrd,word))
				break;

			if(prm->nidx>=prm->midx){
				printf("realloc cidx\n");
				prm->midx+=1024;
				prm->lidx=(UDM_LIDX*)realloc(prm->lidx,sizeof(UDM_LIDX)*prm->midx);
			}
			prm->lidx[prm->nidx].stamp=mw.stamp;
			prm->lidx[prm->nidx].cnt=mw.cnt;
			prm->lidx[prm->nidx].pos=mw.pos;
			prm->nidx++;
			prm->ncoords+=mw.cnt;
		}
	}

	ofs=ihdr.nsorted*ihdr.recsize+sizeof(ihdr);
	ofs1=lseek(fid,ofs,SEEK_SET);
	if(ofs!=ofs1){
#ifdef DEBUG_SEARCH	
		fprintf(stderr,"unsorted part: can't lseek first: %d %d\n",(int)ofs,(int)ofs1);
#endif		
		goto err;
	}

	/* Allocate temporary buffer */
	/* for unsorted index        */

	nbytes=nibuf*ihdr.recsize;
	ibuf=(char*)malloc(nbytes);
	ribuf=ihdr.nunsorted;


	/* Now read unsorted part */
	if(ribuf>0)do{
		size_t nidx;

		if(ribuf<nibuf){
			nbytes=ribuf*ihdr.recsize;
		}else{
			nbytes=nibuf*ihdr.recsize;
		}

		nread=read(fid,ibuf,nbytes);
		if(nbytes!=nread){
#ifdef DEBUG_SEARCH
			fprintf(stderr,"unsorted part: can't read buf: %d %d\n",nbytes,nread);
#endif
			UDM_FREE(ibuf);
			goto err;
			ribuf=0;
		}
		nidx=nread/ihdr.recsize;

		ribuf-=nidx;

		for(i=0;i<nidx;i++){
			memcpy(&mw,ibuf+i*ihdr.recsize,ihdr.recsize);
			mw.wrd[wordlen]='\0';
			
			if(!strcmp(mw.wrd,word)){
				if(prm->nidx>=prm->midx){
					prm->midx+=1024;
					prm->lidx=(UDM_LIDX*)realloc(prm->lidx,sizeof(UDM_LIDX)*prm->midx);
				}
				prm->lidx[prm->nidx].stamp=mw.stamp;
				prm->lidx[prm->nidx].cnt=mw.cnt;
				prm->lidx[prm->nidx].pos=mw.pos;
				prm->nidx++;
				prm->ncoords+=mw.cnt;
			}
		}
	}while(ribuf>0);
	UDM_FREE(ibuf);
	

	/* Sort index in file offset order */
	qsort(prm->lidx+snidx,prm->nidx-snidx,sizeof(UDM_LIDX),(qsort_cmp)(cmplidx));

	/* Reallocate buffer for words */
	nbytes=prm->ncoords*sizeof(UDM_URL_CRD);
	prm->coords=(UDM_URL_CRD*)realloc(prm->coords,nbytes);
	mcoords=0;

	/* Load data */
	for(i=snidx;i<prm->nidx;i++){
		ofs=prm->lidx[i].pos*sizeof(UDM_URL_CRD)+sizeof(UDM_WDAT_HDR);
		ofs1=lseek(fdd,ofs,SEEK_SET);
		if(ofs!=ofs1){
#ifdef DEBUG_SEARCH
			fprintf(stderr,"can't lseek: %d %d\n",(int)ofs,(int)ofs1);
#endif			
			goto err;
		}
		nbytes=sizeof(UDM_URL_CRD)*prm->lidx[i].cnt;
		nread=read(fdd,prm->coords+sncoords+mcoords,nbytes);
		if(nbytes!=nread){
#ifdef DEBUG_SEARCH		
			fprintf(stderr,"can't read: %d %d\n",(int)nbytes,(int)nread);
#endif			
			goto err;
		}
		prm->lidx[i].pos=sncoords+mcoords;
		mcoords+=prm->lidx[i].cnt;
	}
	

#ifdef DEBUG_SEARCH
	fprintf(stderr,"applying dellog\n");
#endif
	for(i=snidx;i<prm->nidx;i++){
		int j;
		time_t stamp;
		size_t start=0;

		for(j=0;j<prm->lidx[i].cnt;j++){
			stamp=PresentInDelLog(prm->del,prm->ndel,&start,prm->coords[prm->lidx[i].pos+j].url_id);

			if(stamp>prm->lidx[i].stamp){
				prm->coords[prm->lidx[i].pos+j].coord=0;
			}
		}
	}
#ifdef DEBUG_SEARCH
	fprintf(stderr,"done nidx=%d ncoords=%d\n",prm->nidx,prm->ncoords);
#endif

	/* Now sort last added part */
#ifdef DEBUG_SEARCH
	fprintf(stderr,"sort %d %d items\n",(int)time(0),prm->ncoords-sncoords);
#endif
	qsort(prm->coords+sncoords,prm->ncoords-sncoords,sizeof(UDM_URL_CRD),(qsort_cmp)(cmpcoordurl));

#ifdef DEBUG_SEARCH
	fprintf(stderr,"sort done %d\n",(int)time(0));
#endif
	/* FIXME: remove nulls after dellog*/

	/* Add word to merge list */
	if(prm->ncoords>sncoords){
		prm->merg[prm->nmerg].cur=sncoords;
		prm->merg[prm->nmerg].last=prm->ncoords-1;
		prm->merg[prm->nmerg].num=wordnum;
#ifdef DEBUG_SEARCH
		fprintf(stderr,"add merge table cur=%d last=%d\n",sncoords,prm->merg[prm->nmerg].last);
#endif
		prm->nmerg++;
	}
err:
	if(fid)close(fid);
	if(fdd)close(fdd);
	return(prm->ncoords-sncoords);
}


UDM_URL_CRD *UdmFindBlob(UDM_AGENT * Indexer,const char *q){
	int i;
	int nwrd1=0;
	UDM_URL_CRD *wrd1=NULL;
	UDM_PMERG pmerg[MAXMERGE];
	UDM_WPARAM prm;
	int ind_fd=0,dat_fd=0;
/*	
	uint4 *lim_data[MAX_SEARCH_LIMIT];
	uint4 lim_size[MAX_SEARCH_LIMIT];
	uint4 lim_start[MAX_SEARCH_LIMIT];
	uint4 nlims=0;
*/
#ifdef DEBUG_SEARCH
	unsigned long ticks;
	unsigned long total_ticks=UdmStartTimer();
	
	fprintf(stderr,"Start UdmFindCache('%s')\n",q);
#endif

	UdmPrepare(Indexer,q);

	bzero(&prm,sizeof(prm));
	LoadDelLog(Indexer,&prm.del,&prm.ndel);

	/*
	bzero(lim_start,MAX_SEARCH_LIMIT*sizeof(uint4));
	*/

#ifdef DEBUG_SEARCH
	fprintf(stderr,"    Reading limits... ");
	ticks=UdmStartTimer();
#endif

/*
	for(i=0;i<Indexer->Conf->nlimits;i++){
		switch(Indexer->Conf->limits[i].type){
			case 0:	if((lim_data[nlims]=LoadNestedLimit(Indexer,Indexer->Conf->limits[i].file_name,
								Indexer->Conf->limits[i].hi,
								Indexer->Conf->limits[i].lo,
								&lim_size[nlims]))) nlims++;
				else return(NULL);
				break;
			case 1: if((lim_data[nlims]=LoadTimeLimit(Indexer,Indexer->Conf->limits[i].file_name,
								Indexer->Conf->limits[i].hi,
								Indexer->Conf->limits[i].lo,
								&lim_size[nlims]))) nlims++;
				else return(NULL);
				break;
			case 2: if((lim_data[nlims]=LoadLinearLimit(Indexer,Indexer->Conf->limits[i].file_name,
								Indexer->Conf->limits[i].hi,
								&lim_size[nlims]))) nlims++;
				else return(NULL);
				break;
		}
	}
*/

#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"\t\t\tDone (%.2f)\n",(float)ticks/1000);
	fprintf(stderr,"    Reading .wrd files (%d words)... ",Indexer->WWList.nwords);
	ticks=UdmStartTimer();
#endif

	for(i=0;i<Indexer->WWList.nwords;i++){
		size_t num;
#ifdef DEBUG_SEARCH
		fprintf(stderr,"LoadCoords for '%s'\n",Indexer->WWList.Words[i].word);
#endif
		num=LoadCoords(Indexer,&prm,Indexer->WWList.Words[i].word,i);
#ifdef DEBUG_SEARCH  
		fprintf(stderr,"done LoadCoords\n");
#endif
		sprintf(UDM_STREND(Indexer->wordinfo)," %s: %d",Indexer->WWList.Words[i].word,num);
	}

#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"\tDone (%.2f)\n",(float)ticks/1000);
	fprintf(stderr,"    Merging (%d groups, %d urls)... ",prm.nmerg,prm.ncoords);
	ticks=UdmStartTimer();
#endif

	if(prm.ncoords){
		/* Allocate memory for merged words */
		wrd1=(UDM_URL_CRD*)malloc(prm.ncoords*sizeof(*wrd1));
		for(i=0;i<prm.nmerg;i++){
			pmerg[i].pcur=prm.coords+prm.merg[i].cur;
			pmerg[i].plast=prm.coords+prm.merg[i].last;
			pmerg[i].num=prm.merg[i].num;
		}
	}

	while(prm.nmerg&&prm.ncoords){
		UDM_PMERG *p;

		if(prm.nmerg==1){

			for(p=&pmerg[0];(p->pcur)<=(p->plast);(p->pcur)++){
				int add=1;
				/*
				int l;
				for(l=0;l<nlims;l++){
					if(!PresentInLimit(lim_data[l],lim_size[l],&lim_start[l],p->pcur->url_id)){
						add=0;
						break;
					}
				}
				*/
				if(add){
					//memcpy(&wrd1[nwrd1],p->pcur,sizeof(*wrd1));
					wrd1[nwrd1]=*(p->pcur);
					wrd1[nwrd1].coord+=p->num;
					nwrd1++;
				}
			}
			
			break;
		}else{
			size_t k;
			uint4 min_url_id=(pmerg[0].pcur)->url_id;
			uint4 min_pos=UDM_WRDPOS((pmerg[0].pcur)->coord);
			uint4 min_num=0;
			/*int l,*/
			int add=1;

			for(k=1;k<prm.nmerg;k++){ 
				/* note ">=" to compare pos in the same url */
				if( (min_url_id>(pmerg[k].pcur)->url_id) ||
				    ((min_url_id==(pmerg[k].pcur)->url_id)&&
				     (min_pos>UDM_WRDPOS((pmerg[k].pcur)->coord))) ){
					min_url_id=(pmerg[k].pcur)->url_id;
					min_pos=UDM_WRDPOS((pmerg[k].pcur)->coord);
					min_num=k;
				}
			}
			p=&pmerg[min_num];
			/*
			for(l=0;l<nlims;l++){
				if(!PresentInLimit(lim_data[l],lim_size[l],&lim_start[l],p->pcur->url_id)){
					add=0;
					break;
				}
			}
			*/
			if(add){
				//memcpy(&wrd1[nwrd1],p->pcur,sizeof(*wrd1));
				wrd1[nwrd1]=*(p->pcur);
				wrd1[nwrd1].coord+=p->num;
				nwrd1++;
			}
			
			if(p->pcur==p->plast){
				prm.nmerg--;
				if(min_num<prm.nmerg){
					memmove(p,p+1,sizeof(UDM_PMERG)*(prm.nmerg-min_num));
				}
			}else{
				(p->pcur)++;
			}
		}
	}
	UDM_FREE(prm.coords);

	/*
	for(i=0;i<nlims;i++) UDM_FREE(lim_data[i]);
	*/
#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"\tDone (%.2f) nwrd=%d nwrd1=%d\n",(float)ticks/1000,prm.ncoords,nwrd1);
	fprintf(stderr,"    Grouping by url_id... ");
	ticks=UdmStartTimer();
#endif
	if((Indexer->total_found=nwrd1)) UdmGroupByURL(Indexer,wrd1);

#ifdef DEBUG_SEARCH1
	for(i=0;i<Indexer->total_found;i++){
		fprintf(stderr,"res: %d\t%d\n",wrd1[i].url_id,wrd1[i].coord);
	}
#endif

#ifdef DEBUG_SEARCH
	ticks=UdmStartTimer() - ticks;
	fprintf(stderr,"\t\tDone (%.2f) %d found\n",(float)ticks/1000,Indexer->total_found);
	total_ticks=UdmStartTimer() - total_ticks;
	fprintf(stderr,"Stop UdmFindCache() - %.2f\n",(float)total_ticks/1000);
#endif
	return(wrd1);
	
	if(ind_fd)close(ind_fd);
	if(dat_fd)close(dat_fd);
	return(NULL);
}

